/* Handles console commands via serial (usb) or telnet connection

   Armin Diehl <ad@ardiehl.de> December 2017

   Based on the esp32 console example by expressif

   11 June 2018 AD: close telnet connections before restart
   22 Oct  2018 AD:	show cpu version in ver command
   28 Oct  2018 AD: added ping
   04 Dec  2018 AD: separated commands.cpp from evse project + menuconfig

   Unless required by applicable law or agreed to in writing, this
   software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
   CONDITIONS OF ANY KIND, either express or implied.

   This code is in the Public Domain (or CC0 licensed, at your option.)
*/


#include <ctype.h>
#include "commands.h"
#include <stdint.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <time.h>
#include <sys/ioctl.h>

#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/queue.h"
#include "freertos/event_groups.h"
#include "freertos/timers.h"
#include "cons.h"

#include "argtable3_ad.h"
#include "linenoise_ad.h"

#include "telnetDefs.h"

#ifdef ESP32
#include "esp_vfs_dev.h"
#include "driver/uart.h"

#include "esp_vfs_fat.h"
//#include "nvs_flash.h"
//#include "nvs.h"

#ifdef CONFIG_COMMANDS_WLAN
#include "esp_wifi.h"
#endif
#include "tcpip_adapter.h"
#include "esp_event_loop.h"
#ifdef CONFIG_COMMANDS_ETH
//#include "eth_phy/phy_lan8720.h"
#include "esp_eth.h"
#endif

#ifdef CONFIG_COMMANDS_I2C
#include <driver/i2c.h>
#endif

#include "esp_system.h"
#include "esp_spi_flash.h"
//#include "ping/ping.h"
#ifdef CONFIG_COMMANDS_PING
#include "ping.h"
#include "esp_ping.h"
#include <netdb.h>
#endif

#include "esp_err.h"
#include "esp_log.h"
#else  // ESP32
#include "esp_compat.h"
#endif // ESP32


//#include "version.h"
#include <time.h>
#include "update.h"
#include "network.h"

#include "globalLock.h"

#ifdef CONFIG_COMMANDS_I2C_MCP23017
#include "mcp23017_portio.h"
#endif

#include "configfile.h"

#ifdef CONFIG_COMMANDS_ADCREAD
#include "driver/adc.h"
#include "esp_adc_cal.h"
#endif

#define RETCODE_EXIT 4242


static const char* TAG = "commands";

// *** Ping for ESP32 ********************************************************************************************

#if defined (ESP32) && defined (CONFIG_COMMANDS_PING)

typedef struct _pingParam {
	FILE *pingConOut;
	EventGroupHandle_t eventGroup;
	char* pingHost;
	uint32_t maxTime;
	uint32_t minTime;
	uint32_t totalTime;
	uint32_t count;      //how many pings per report
    uint32_t timeout;    //mS till we consider it timed out
    uint32_t interval;   //mS between pings
    uint32_t datalen;
    uint32_t okCount = 0;
    uint32_t failCount = 0;
    FILE * conout;

} pingParam_t;


 pingParam_t * pingParam;

//#define PING_HAS_DATA_LEN

extern "C" esp_err_t pingResults(ping_target_id_t msgType, esp_ping_found * pf){
	if (pf->ping_err == PING_RES_FINISH) {
		xEventGroupSetBits(pingParam->eventGroup, 1);  // signal end of command
	} else if (pf->ping_err == PING_RES_TIMEOUT) {
		fprintf(pingParam->conout,CONS_COLOR_STR_E_N("\ricmp_seq=%d host unreachable / timeout (%d)"),pf->send_count,pf->ping_err);
		pingParam->failCount++;
	} else {
		//fprintf(pingParam->conout,"\rAvgTime:%.1fmS Sent:%d Rec:%d Err:%d min(mS):%d max(mS):%d ", (float)pf->total_time/pf->recv_count, pf->send_count, pf->recv_count, pf->err_count, pf->min_time, pf->max_time );
		//fprintf(pingParam->conout,"Resp(mS):%d Timeouts:%d Total Time:%d\r\n",pf->resp_time, pf->timeout_count, pf->total_time);
		fprintf(pingParam->conout,"\r%d Bytes from %s: icmp_seq: %d time=%d\r\n",pingParam->datalen+8,pingParam->pingHost, pf->send_count, pf->resp_time);
		pingParam->okCount++;
		if (pf->resp_time > pingParam->maxTime) pingParam->maxTime = pf->resp_time;
		if (pf->resp_time < pingParam->minTime) pingParam->minTime = pf->resp_time;
		pingParam->totalTime += pf->resp_time;
	}
	fflush(pingParam->conout);
	return ESP_OK;
}


static struct {
	struct arg_int *count       = arg_int0	("c", "count"    , NULL, "ping count, default=5");
	struct arg_int *interval    = arg_int0	("i", "interval" , NULL, "interval in ms, default=500, min=250");
#ifdef PING_HAS_DATA_LEN
	struct arg_int *datalen     = arg_int0	("s", "size"     , NULL, "packetsize in bytes, defaults to 56");
#endif
    struct arg_str *dest 		= arg_str0	(NULL, NULL, "<destination>", "destination hostname or ip");
    struct arg_end *end 		= arg_end(2);
} ping_commandArgs;


static int ping_command(FILE *fd,  FILE *,int, char**)
{
	int err;
	struct addrinfo hints;
    struct addrinfo *res;
    EventBits_t currBits;
    pingParam_t pp;

    pp.count = 5;  //how many pings per report
    pp.interval = 500; //mS between pings
    pp.datalen = 56;

    pp.maxTime = 0;
    pp.minTime = 999999;
    pp.totalTime = 0;
    pp.okCount = 0;
    pp.failCount = 0;

	if (!(ping_commandArgs.dest->count)) {
		arg_print_glossary(fd, (void**) &ping_commandArgs, NULL);
    	return 0;
	}

	if (ping_commandArgs.count->count) {
		pp.count = ping_commandArgs.count->ival[0];
		if (pp.count < 1) pp.count = 1;
	}

	if (ping_commandArgs.interval->count) {
		pp.interval = ping_commandArgs.interval->ival[0];
		if (pp.interval < 250) {
			pp.interval = 1;
			fprintf(fd,CONS_COLOR_STR_W_N("Ping interval to small, using %d ms"),pp.interval);
		}
	}
#ifdef PING_HAS_DATA_LEN
	if (ping_commandArgs.datalen->count) {
		pp.datalen = ping_commandArgs.datalen->ival[0];
		if (pp.datalen < 56) pp.datalen = 56;
	}
#endif
	pp.timeout = pp.interval *  2;
	pp.conout = fd;

    memset(&hints, 0, sizeof(hints));
    //hints.ai_family=AF_UNSPEC;
    hints.ai_family = AF_INET;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;

    pp.pingHost = (char *)ping_commandArgs.dest->sval[0];
	err = getaddrinfo(pp.pingHost, NULL, &hints, &res);

    if (err != 0 || res == NULL) {
        fprintf(fd,CONS_COLOR_STR_E_N("ping: %s: unknown (%d)"), pp.pingHost, err);
        return 0;
    }

    pp.eventGroup = xEventGroupCreate();
    pingParam = &pp;

    struct sockaddr_in *addr;
    addr = (struct sockaddr_in *)res->ai_addr;

	esp_ping_set_target(PING_TARGET_IP_ADDRESS_COUNT, &pp.count, sizeof(uint32_t));
    esp_ping_set_target(PING_TARGET_RCV_TIMEO, &pp.timeout, sizeof(uint32_t));
    esp_ping_set_target(PING_TARGET_DELAY_TIME, &pp.interval, sizeof(uint32_t));
    esp_ping_set_target(PING_TARGET_IP_ADDRESS, &addr->sin_addr, sizeof(uint32_t));
    esp_ping_set_target(PING_TARGET_RES_FN, (void *)&pingResults, sizeof(void*));
#ifdef PING_HAS_DATA_LEN
    esp_ping_set_target(PING_TARGET_DATA_LEN,&pp.datalen,sizeof(uint32_t));
#endif

    fprintf(fd,CONS_COLOR_STR_I_N("PING %s (%s) %d bytes of data"),pp.pingHost,inet_ntoa(addr->sin_addr),pp.datalen);
    ping_init();

    do {
    	currBits = xEventGroupWaitBits(pp.eventGroup,1,pdFALSE,pdFALSE, pdMS_TO_TICKS(pp.count * pp.timeout * 2));
    	if (currBits == 0)
    		fprintf(fd,CONS_COLOR_STR_E_N("Timeout, PING_RES_FINISH not yet signaled"));

    } while (currBits == 0);
    ping_deinit();

    float loss;
    if (pp.failCount < 1)
    	loss = 0;
    else
    	loss = (1 - ((float)pp.okCount / (float)pp.failCount))*100;

    fprintf(fd,CONS_COLOR_STR_T_N("\r--- %s ping statistics ---"),pp.pingHost);
    fprintf(fd,"\r%d packets transmitted, %d received, %3.1f%% packet loss, time %dms\r\n",pp.count,pp.okCount,loss,pp.totalTime);
    if (pp.okCount)
    	fprintf(fd,"\rrtt min/avg/max = %d/%d/%d ms\r\n",pp.maxTime,pp.totalTime / pp.okCount,pp.minTime);

    vEventGroupDelete(pp.eventGroup);
    pingParam = NULL;
    return 0;
}

#endif /* CONFIG_COMMANDS_PING  */


#ifdef CONFIG_COMMANDS_WLAN
static struct {
    struct arg_int *timeout = arg_int0	("t", "timeout", "<t>", "Connection timeout, ms, default=10000");
    struct arg_lit *def = arg_lit0		("d", "def", "set wlan as default connection");
    struct arg_lit *dis = arg_lit0		("i", "disconnect", "disconnect wlan");
    struct arg_lit *con = arg_lit0		("c", "connect", "connect using saved SSID/PASS");
    struct arg_lit *nodef = arg_lit0	("n", "nodef", "unset wlan default connection");
    struct arg_str *ssid = arg_str0		(NULL, NULL, "<ssid>", "SSID of AP");
    struct arg_str *password = arg_str0	(NULL, NULL, "<pass>", "PSK of AP");
    struct arg_end *end = arg_end(2);
} wlan_commandArgs;


static int wlan_command(FILE *fd,  FILE *,int, char**)
{
    esp_err_t err;

    if (wlan_commandArgs.dis->count) {
    	err = esp_wifi_disconnect();
    	if (err != 0) fprintf(fd,"esp_wifi_disconnect returned %d\r\n",err);
    	else fprintf(fd,"wifi disconnected\r\n");
    	return 0;
    }
    
    if (wlan_commandArgs.nodef->count) {
    	network_WifiSaveSettings (0);
    	fprintf(fd,"wlan not set as default\r\n");
    	return 0;
    }

    if (wlan_commandArgs.con->count) {
    	network_connectWifi(fd);
    	return 0;
    }

    if (wlan_commandArgs.ssid->count < 1) {
    	arg_print_glossary(fd, (void**) &wlan_commandArgs, NULL);
    	return 0;
    }
	
    fprintf(fd, "Connecting to '%s'", wlan_commandArgs.ssid->sval[0]);
    
    if (wlan_commandArgs.timeout->count <= 0) wlan_commandArgs.timeout->ival[0] = DEFAULT_WIFI_CONNECT_TIMEOUT_MS; // set default value
    

    bool connected = wifi_connect(fd,
    		wlan_commandArgs.ssid->sval[0],
			wlan_commandArgs.password->sval[0],
			wlan_commandArgs.timeout->ival[0]);

	network_WifiSaveSettings (
    		wlan_commandArgs.def->count,
			wlan_commandArgs.timeout->ival[0],
			wlan_commandArgs.ssid->sval[0],
			wlan_commandArgs.password->sval[0]);


    if (!connected) {
        fprintf(fd, CONS_COLOR_STR_E_N("Connection timed out"));
        return 1;
    }
    fprintf(fd, CONS_COLOR_STR_I_N("Connected to '%s'"),wlan_commandArgs.ssid->sval[0]);
    return 0;
}

#endif /* CONFIG_COMMANDS_WLAN */


#ifdef CONFIG_COMMANDS_ETH

static struct {
	struct arg_lit *con = arg_lit0		("c", "connect", "connect ethernet");
    struct arg_lit *dis = arg_lit0		("i", "disconnect", "disconnect ethernet");
    struct arg_int *timeout = arg_int0	("t", "timeout", "<t>", "Connection timeout, ms, default=5000");
    struct arg_lit *def = arg_lit0		("d", "def", "set eth as default connection");
    struct arg_lit *nodef = arg_lit0	("n", "nodef", "unset eth default connection");
    struct arg_end *end = arg_end(2);
} eth_commandArgs;


static int eth_command(FILE *fd, FILE *, int, char**)
{

    if (eth_commandArgs.dis->count) {
    	esp_eth_disable();
    	fprintf(fd, CONS_COLOR_STR_I_N("Eth disconnected"));
    	return 0;
    }

    int timeout = network_EthGetTimeout();
    if (eth_commandArgs.timeout->count) timeout = eth_commandArgs.timeout->ival[0];

    if (eth_commandArgs.nodef->count) {
    	network_EthSaveSettings (0,timeout);		// clear default
    	return 0;
    }
    if (eth_commandArgs.def->count) {
    	network_EthSaveSettings (1,timeout);		// set as default
    	return 0;
    }


    if (eth_commandArgs.con->count) {
    	fprintf(fd, "Connecting eth\r\n");

    	bool connected = eth_connect (fd,timeout);

    	if (!connected) {
    		fprintf(fd, CONS_COLOR_STR_E_N("Connection timed out"));
    		return 1;
    	}
    	fprintf(fd, CONS_COLOR_STR_I_N("Eth connected"));
    	network_EthSaveSettings (eth_commandArgs.def->count,timeout);
    } else {
    	arg_print_glossary(fd, (void**) &eth_commandArgs, NULL);
    }
    return 0;
}

#endif   /* CONFIG_COMMANDS_ETH */



/** 'restart_command' command restart_commands the program */
#ifdef CONFIG_COMMANDS_RESTART
static int restart_command(FILE *fp, FILE *, int, char**)
{
    fprintf(fp, "%s : restarting\n", __func__);

#ifdef CONFIG_COMMANDS_TELNET_SERVER
    telnetCloseAll();
#endif

#ifdef ESP32
    esp_restart();
#endif

    return RETCODE_EXIT;
}
#endif   /* CONFIG_COMMANDS_RESTART */


/** 'free' command prints available heap memory */
#ifdef CONFIG_COMMANDS_FREE
static int free_command(FILE *fp, FILE *, int, char**)
{
#ifdef ESP32
    fprintf(fp, "%d\r\n", esp_get_free_heap_size());
#else
    fprintf(fp, "enough\n");
#endif
    return 0;
}
#endif   /* CONFIG_COMMANDS_FREE */


#ifdef ESP32
char currDir[MAXDIRLEN] = "/data/";
#else
char currDir[MAXDIRLEN] = "./data/";
#endif


static int ls_command(FILE *fp, FILE *, int, char**)
{
	DIR *dp;
	struct dirent *ep;
	struct stat s;
	char fullFileName [MAXDIRLEN];

	dp = opendir (currDir);
	if (dp != NULL)
	{
		while ((ep = readdir (dp))) {
			strncpy(fullFileName, currDir, MAXDIRLEN);
			strncat(fullFileName, ep->d_name, MAXDIRLEN);
			fullFileName[MAXDIRLEN-1] = '\0';
			if (stat(fullFileName, &s) != 0) s.st_size = 0;
			if ((s.st_mode & S_IFMT) == S_IFDIR) {
				if ((strcmp(ep->d_name,".") != 0) && (strcmp(ep->d_name,"..") != 0)) {
					fprintf(fp, CONS_COLOR_STR_I_N("%-20s  <DIR>"),ep->d_name);
				}
			} else
				fprintf(fp, "%-20s %6ld\r\n",ep->d_name, s.st_size);
		}
		(void) closedir (dp);
		fprintf (fp,"\r\n");
	} else
		fprintf (fp, CONS_COLOR_STR_E_N("Unable to open the directory '%s'\r\n"),currDir);

    return 0;
}



static int pwd_command(FILE *fp, FILE *, int, char**)
{
/*	char cwd[256];
	if (getcwd(cwd, sizeof(cwd)) != NULL)
	     printf("%s\n", cwd);
	else
	   perror("getcwd() error");
*/
	fprintf(fp, "%s\r\n", currDir);
	return 0;
}

#define MAX_LINE_LEN 255

static struct {
    struct arg_str *fileName = arg_str1(NULL, NULL, "filename", "filename of file to show");
    struct arg_end *end = arg_end(2);
} cat_commandArgs;


int catFile (FILE *fd, char*fileName) {
	FILE *f;

	f = fopen(fileName,"r");
	if (f == NULL) {
		fprintf(fd,CONS_COLOR_STR_E_N("cat: Unable to open %s (%d)"),fileName,errno);
		return 1;
	}
	char buf[MAX_LINE_LEN+1];
    buf[MAX_LINE_LEN] = '0';

    while (fgets(buf,MAX_LINE_LEN, f) != NULL) {
        char *p;

        p = strchr(buf,'\r');
        if (!p) p = strchr(buf,'\n');
        if (p) *p = '\0';
        fprintf(fd,buf); fprintf(fd,"\r\n");
    }
    fclose(f);
    return 0;
}


static int cat_command(FILE *fd, FILE *, int, char**) {
	char fileName[255];
	strcpy(fileName,currDir); strcat(fileName,cat_commandArgs.fileName->sval[0]);
	catFile(fd,fileName);

	return 0;
}

//****************************************************************************************
#ifdef CONFIG_COMMANDS_I2C_MCP

static struct {
    struct arg_int *mcpidx = arg_int0("x", "idx" , NULL, "index of mcp23017, default=0");

    struct arg_lit *read   = arg_lit0("r", "read" ,"read io port register");
    struct arg_int *set    = arg_int0("s", "set"  ,NULL, "set given bits");
    struct arg_int *clear  = arg_int0("c", "clr"  ,NULL, "clear given bits");

    struct arg_int *inp    = arg_int0("n", "in"   ,NULL, "set given bits as input");
    struct arg_int *out    = arg_int0("o", "out"  ,NULL, "set given bits as output");
    struct arg_lit *pullup = arg_lit0("p", "pullup" ,"if setting pins as input, enable pullup");



    // for init i2c only
    struct arg_lit *init   = arg_lit0("i", "init" ,"init i2c and install driver");
    struct arg_int *addr   = arg_int0("a", "addr" , NULL, "i2c address      (for init only)");
    struct arg_int *sda    = arg_int0("d", "sda"  , NULL, "sda port number  (for init only)");
    struct arg_int *scl    = arg_int0("l", "scl"  , NULL, "scl port number  (for init only)");
    struct arg_int *port   = arg_int0("t", "port" , NULL, "i2c port, 0 or 1 (for init only)");
    struct arg_int *speed  = arg_int0("e", "speed", NULL, "i2c speed hz (for init only)");


    struct arg_end *end = arg_end(5);
} mcp_commandArgs;

static int mcp_command(FILE *fd, FILE *, int, char**) {
	mcp23017_t mcp = {
			.i2c_addr = 0x20,
			.port = I2C_NUM_0,
			.sda_pin = 0,
			.scl_pin = 0,
			.sda_pullup_en = GPIO_PULLUP_DISABLE,
			.scl_pullup_en = GPIO_PULLUP_DISABLE,
			.i2cClockSpeed = 100000 //EVSE_I2C_MASTER_FREQ_HZ
	};
	int err;
	int idx = 0;
	int numCmds = 0;
	uint8_t lo;
	uint8_t hi;

	if (mcp_commandArgs.mcpidx->count) idx = mcp_commandArgs.mcpidx->ival[0];

	if (mcp_commandArgs.init->count) {
		if (mcp_commandArgs.sda->count) mcp.sda_pin = mcp_commandArgs.sda->ival[0];
		if (mcp.sda_pin < 1) {
			fprintf(fd,CONS_COLOR_STR_E_N("init requires SDA pin number"));
			return 1;
		}
		if (mcp_commandArgs.scl->count) mcp.scl_pin = mcp_commandArgs.scl->ival[0];
		if (mcp.scl_pin < 1) {
			fprintf(fd,CONS_COLOR_STR_E_N("init requires SCL pin number"));
			return 1;
		}
		if (mcp_commandArgs.speed->count) mcp.i2cClockSpeed = mcp_commandArgs.speed->ival[0];
		if (mcp_commandArgs.addr->count) mcp.i2c_addr = mcp_commandArgs.addr->ival[0];
		if (mcp_commandArgs.port->count) {
			switch (mcp_commandArgs.port->ival[0]) {
				case 0:	mcp.port = I2C_NUM_0; break;
				case 1: mcp.port = I2C_NUM_1; break;
				default: fprintf(fd,CONS_COLOR_STR_E_N("init requires port 0 or 1")); return 1;
			}
		}
		fprintf(fd,"pip_init, i2c_port: %d, i2c_addr: 0x%02x SCL: %d SDA: %d Speed:%d\r\n",(int)mcp.port,(int)mcp.i2c_addr,mcp.scl_pin,mcp.sda_pin,mcp.i2cClockSpeed);
#ifdef ESP32
		err = pio_init(idx, &mcp);
		if (err != 0) {
			fprintf(fd,CONS_COLOR_STR_E("failed ") " with %d\r\n", err);
			return 1;
		}
#endif
		fprintf(fd,CONS_COLOR_STR_I_N("init OK"));
		numCmds++;
	}

	if (mcp_commandArgs.read->count) {
		err = pio_getByte (idx, 0, &lo);
		if (err == 0) err = pio_getByte (idx, 1, &hi);
		if (err == 0) {
			fprintf(fd, "lo: 0x%02x hi:0x%02x\r\n",lo,hi);
		} else fprintf(fd,CONS_COLOR_STR_E("read failed ") " with %d\r\n", err);
		numCmds++;
	}

	if (mcp_commandArgs.set->count) {
		err = pio_setPins (idx, mcp_commandArgs.set->ival[0]);
		fprintf(fd, "pio_setPins (%d, 0x%04x): %d\r\n",idx,mcp_commandArgs.set->ival[0],err);
		if (err == 0) fprintf(fd,"read OK\r\n"); else fprintf(fd,CONS_COLOR_STR_E("read failed ") " with %d\r\n", err);
		numCmds++;
	}

	if (mcp_commandArgs.clear->count) {
		err = pio_clearPins (idx, mcp_commandArgs.clear->ival[0]);
		fprintf(fd, "pio_clearPins (%d, 0x%04x): %d\r\n",idx,mcp_commandArgs.clear->ival[0],err);
		if (err == 0) fprintf(fd,"read OK\r\n"); else fprintf(fd,CONS_COLOR_STR_E("read failed ") " with %d\r\n", err);
		numCmds++;
	}

	if (mcp_commandArgs.inp->count) {
		err = pio_setPinsInput (idx, mcp_commandArgs.inp->ival[0], mcp_commandArgs.pullup->count);
		fprintf(fd, "pio_setPinsInput (%d, 0x%04x, %d): %d\r\n",idx,mcp_commandArgs.inp->ival[0],mcp_commandArgs.pullup->count,err);
		if (err == 0) fprintf(fd,"pio_setPinsInput OK\r\n"); else fprintf(fd,CONS_COLOR_STR_E("pio_setPinsInput failed ") " with %d\r\n", err);
		numCmds++;
	}

	if (mcp_commandArgs.out->count) {
		err = pio_setPinsOutput (idx, mcp_commandArgs.out->ival[0]);
		fprintf(fd, "pio_setPinsOutput (%d, 0x%04x): %d\r\n",idx,mcp_commandArgs.out->ival[0],err);
		if (err == 0) fprintf(fd,"pio_setPinsOutput OK\r\n"); else fprintf(fd,CONS_COLOR_STR_E("pio_setPinsOutput failed ") " with %d\r\n", err);
		numCmds++;
	}

	if (numCmds == 0) arg_print_glossary(fd, (void**) &mcp_commandArgs, NULL);
	return 0;
}
#endif /* CONFIG_COMMANDS_I2C_MCP */
//****************************************************************************************
#ifdef CONFIG_COMMANDS_INOUT
static struct {
    struct arg_int *gpio = arg_int0(NULL, NULL, NULL, "gpio port number");
    struct arg_lit *pullup = arg_lit0("u", "pullup", "enable pullup");
    struct arg_lit *pulldn = arg_lit0("d", "pulldown", "enable pulldown");
    struct arg_end *end = arg_end(2);
} in_commandArgs;


static int in_command(FILE *fd, FILE *, int, char**) {
	/*
	esp_err_t err = gpio_set_direction((gpio_num_t)in_commandArgs.gpio->ival[0], GPIO_MODE_INPUT);
	if (err) ESP_LOGE("in_command","gpio_set_direction returned %d",err); */
	gpio_config_t c;
	gpio_num_t gpioNum = (gpio_num_t)in_commandArgs.gpio->ival[0];

	c.pin_bit_mask = 1 << gpioNum;			// GPIO pin: set with bit mask, each bit maps to a GPIO
	c.mode = GPIO_MODE_INPUT;				// GPIO mode: set input/output mode
	if (in_commandArgs.pullup->count)
		c.pull_up_en = GPIO_PULLUP_ENABLE;
	else
		c.pull_up_en = GPIO_PULLUP_DISABLE;

	if (in_commandArgs.pulldn->count)
		c.pull_down_en = GPIO_PULLDOWN_ENABLE;	// GPIO pull-down
	else
		c.pull_down_en = GPIO_PULLDOWN_DISABLE;
	c.intr_type = GPIO_INTR_DISABLE;
	esp_err_t err = gpio_config(&c);
	if (err) fprintf(fd, CONS_COLOR_STR_E_N("gpio_config returned %d"),err);

	int state = gpio_get_level((gpio_num_t)in_commandArgs.gpio->ival[0]);
	fprintf(fd,"gpio(%d): %d (0x%02x)\r\n",in_commandArgs.gpio->ival[0],state,state);
	return 0;
}

//****************************************************************************************

static struct {
    struct arg_int *gpio = arg_int0(NULL, NULL, NULL, "gpio port number");
    struct arg_int *val = arg_int0(NULL, NULL, NULL, "value");
    struct arg_end *end = arg_end(2);
} out_commandArgs;


static int out_command(FILE *fd, FILE *, int, char**) {
	int val;
/*
	esp_err_t err = gpio_set_direction((gpio_num_t)out_commandArgs.gpio->ival[0], GPIO_MODE_OUTPUT);
	if (err) ESP_LOGE("in_command","gpio_set_direction returned %d",err);
*/

	gpio_config_t c;
	gpio_num_t gpioNum = (gpio_num_t)out_commandArgs.gpio->ival[0];

	c.pin_bit_mask = (uint64_t)1 << (uint64_t) gpioNum;			// GPIO pin: set with bit mask, each bit maps to a GPIO
	fprintf(fd,"pin_bit_mask 0x%llx¸\r\n",c.pin_bit_mask);
	c.mode = GPIO_MODE_OUTPUT;				// GPIO mode: set input/output mode
	c.pull_up_en = GPIO_PULLUP_DISABLE;		//GPIO pull-up
	c.pull_down_en = GPIO_PULLDOWN_DISABLE;	// GPIO pull-down
	c.intr_type = GPIO_INTR_DISABLE;
	esp_err_t err = gpio_config(&c);
	/*if (err)*/ fprintf(fd, CONS_COLOR_STR_E_N("gpio_config returned %d"),err);

	if (out_commandArgs.val->ival[0]) val = 1; else val = 0;

	int state = gpio_set_level(gpioNum, val);
	fprintf(fd,CONS_COLOR_STR_E_N("gpio(%d) set to: %d (0x%02x), rc=%d"),gpioNum,val,val,state);
	return 0;
}
#endif  // CONFIG_COMMANDS_INOUT
//****************************************************************************************

static struct {
	struct arg_int *msec  = arg_int0 (NULL,NULL,"<ms>","milliseconds to delay or 1000 if not specified");
    struct arg_end *end   = arg_end(2);
} delay_commandArgs;


static int delay_command(FILE *fd, FILE *, int, char**) {
	int time = 1000;
	if (delay_commandArgs.msec->count > 0) time=delay_commandArgs.msec->ival[0];
	fprintf(fd,"pausing for %d ms",time); fflush(fd);
	vTaskDelay (pdMS_TO_TICKS(time));
	fprintf(fd,"\r\n");
	return 0;
}


//****************************************************************************************

#ifdef CONFIG_COMMANDS_PS
#if ( configUSE_TRACE_FACILITY == 1 )

static int ps_command(FILE *fd, FILE *, int, char**) {
	TaskStatus_t *pxTaskStatusArray;
	volatile UBaseType_t uxArraySize, x;
	float statsAsPercentage;
	uint32_t ulTotalRunTime, ulTotalRunTime2;
	unsigned long stackHighWaterMark = 0,runTimeCounter;
	int currentPriority;
#ifdef ESP32
	int pinnedToCore = 0;
#endif
	char cpuStr[10];

	cpuStr[0] = '\0';

   /* Make sure the write buffer does not contain a string. */
   //*pcWriteBuffer = 0x00;

   /* Take a snapshot of the number of tasks in case it changes while this
   function is executing. */
   uxArraySize = uxTaskGetNumberOfTasks();
   if (uxArraySize <= 0) {
	   fprintf(fd,"uxTaskGetNumberOfTasks() returned %d\r\n", (int)uxArraySize);
	   return 0;
   }

   /* Allocate a TaskStatus_t structure for each task.  An array could be
   allocated statically at compile time. */
   pxTaskStatusArray = ( TaskStatus_t *)malloc( uxArraySize * sizeof( TaskStatus_t ) );

   if( pxTaskStatusArray != NULL )
   {
      /* Generate raw status information about each task. */
      uxArraySize = uxTaskGetSystemState( pxTaskStatusArray,
                                 uxArraySize,
                                 &ulTotalRunTime2);

   	 fprintf( fd, "%-15s %3s %5s %15s %16s\r\n","task name","cpu","prio","min stack free","run time count");
   	 fprintf( fd, "-----------------------------------------------------------------\r\n");

   	 ulTotalRunTime = 0;
   	 for(x=0; x<uxArraySize; x++)
   		 ulTotalRunTime += pxTaskStatusArray[x].ulRunTimeCounter;

   	 //fprintf(fd,"%d\r\n%d\r\n",ulTotalRunTime,ulTotalRunTime2);
   	 /* For each populated position in the pxTaskStatusArray array,
		format the raw data as human readable ASCII data. */
	for(x=0; x<uxArraySize; x++)
	{
		/* What percentage of the total run time has the task used?
			This will always be rounded down to the nearest integer. */
		if (ulTotalRunTime>0) {
			statsAsPercentage = (float)pxTaskStatusArray[x].ulRunTimeCounter *100 / ulTotalRunTime;
		} else
			statsAsPercentage = 0;

#if ( INCLUDE_uxTaskGetStackHighWaterMark == 1 )
            // not available on FreeRTOS posix port
            stackHighWaterMark = uxTaskGetStackHighWaterMark(pxTaskStatusArray[x].xHandle);
#endif
            currentPriority = (int)pxTaskStatusArray[x].uxCurrentPriority;
            runTimeCounter = pxTaskStatusArray[x].ulRunTimeCounter;

#ifdef ESP32
            pinnedToCore = xTaskGetAffinity(pxTaskStatusArray[x].xHandle);	/* Get the current core affinity of a task if running on ESP32 */
            switch (pinnedToCore) {
            	case 0:	strcpy(cpuStr,"0");
            			break;
            	case 1:	strcpy(cpuStr,"1");
            			break;
            	case tskNO_AFFINITY:
            			strcpy(cpuStr,"any");
            			break;
            	default:
            			strcpy(cpuStr,"?");
            }
#endif

            fprintf( fd, "%-15s %3s %5d %15lu %16lu %02.2f%%\r\n",
            			pxTaskStatusArray[ x ].pcTaskName,
						cpuStr,
						currentPriority,
						stackHighWaterMark,
						runTimeCounter,
						statsAsPercentage);
         }

      /* The array is no longer needed, free the memory it consumes. */
      free(pxTaskStatusArray);
   } else
	   fprintf( fd, CONS_COLOR_STR_E_N("malloc failed"));
   fprintf( fd, "\r\n");
   return 0;
}
#endif

#endif /* CONFIG_COMMANDS_PS */

//****************************************************************************************

extern "C" char* ver_getVersionStrLong();
extern "C" char* ver_getAuthor();

static int ver_command(FILE *fd, FILE *, int, char**) {
	fprintf(fd,CONS_COLOR_STR_I("%s"),ver_getVersionStrLong());
	fprintf(fd,CONS_COLOR_STR_W_N("%s"),ver_getAuthor());
#ifdef ESP32
	// Print chip information
    esp_chip_info_t chip_info;
    esp_chip_info(&chip_info);
    fprintf(fd,CONS_COLOR_I "ESP32 with %d CPU cores, WiFi%s%s, ",
            chip_info.cores,
            (chip_info.features & CHIP_FEATURE_BT) ? "/BT" : "",
            (chip_info.features & CHIP_FEATURE_BLE) ? "/BLE" : "");

    fprintf(fd,"revision %d, ", chip_info.revision);

    fprintf(fd,"%dMB %s flash, idf-version %s\r\n" CONS_RESET_COLOR,
    		spi_flash_get_chip_size() / (1024 * 1024), (chip_info.features & CHIP_FEATURE_EMB_FLASH) ? "embedded" : "external",
    		esp_get_idf_version());
#endif
	return 0;
}

//****************************************************************************************

static int date_command(FILE *fd, FILE *, int, char**) {
	time_t t = time(NULL);
	struct tm tm = *localtime(&t);

	fprintf(fd,CONS_COLOR_STR_I("%0d.%02d.%04d %d:%02d:%02d\r\n"), tm.tm_mday, tm.tm_mon + 1, tm.tm_year + 1900, tm.tm_hour, tm.tm_min, tm.tm_sec);
	return 0;
}

//****************************************************************************************
#ifdef CONFIG_COMMANDS_UPDATE
static struct {
	struct arg_str *server = arg_str0("s", "server", NULL, "name or ip of server hosting the firmware binary");
	struct arg_str *path   = arg_str0("f", "filename", NULL, "filename/filepath of firmware binary, default /evse/evse.bin");
	struct arg_int *port   = arg_int0("p", "port", NULL, "portnumber, defaults to 80");
	struct arg_lit *test   = arg_lit0("t", "test", "dryrun, test only, don't flash");
	struct arg_lit *restart= arg_lit0("r", "restart", "restart after successful update");
	struct arg_lit *swap   = arg_lit0("w", "swap", "swap active partitions without changing content");
    struct arg_end *end = arg_end(5);
} update_commandArgs;


static int update_command(FILE *fd, FILE *, int, char**) {
	char *c_server    = configfile.get_str("UPD_SERVER");
	char *c_filename  = configfile.get_str("UPD_FILENAME");
	int port = configfile.get_u32("UPD_PORT");
	if (port == 0) port = 80;

	if (update_commandArgs.swap->count) {
		return toggleActivePartition (fd);
	}

	if (update_commandArgs.server->count) {
		free(c_server); c_server = strdup(update_commandArgs.server->sval[0]);
	}

	if (update_commandArgs.path->count) {
		free(c_filename); c_filename = strdup(update_commandArgs.path->sval[0]);
	}

	if (c_filename == NULL) { c_filename = strdup("/evse/evse.bin"); }

	if ((c_filename == NULL) || (c_server == NULL)) {
		fprintf(fd,CONS_COLOR_STR_E_N("Error: server and filename needs to be specified"));
	} else {
		configfile.set_str("UPD_SERVER", c_server);
		configfile.set_str("UPD_FILENAME", c_filename);
		configfile.set_u32("UPD_PORT", port);
		configfile.save();
		if (update_firmware (fd, c_server, c_filename, port, update_commandArgs.test->count) == true) {
			fprintf(fd,CONS_COLOR_STR_I_N("update: success"));
#ifdef ESP32
			if (update_commandArgs.restart->count) {
				esp_restart();
			}
#endif
			fprintf(fd,"type restart to run the new firmware\r\n");
		} else
			fprintf(fd,CONS_COLOR_STR_E_N("Error: update failed"));

	}

	free(c_server); free(c_filename);
	return 0;
}

#endif /* CONFIG_COMMANDS_UPDATE */
//****************************************************************************************

static struct {
	struct arg_str *tag = arg_str0("t", "tag", NULL, "set for a specific tag, e.g. wifi, lan8720 or u8g2_hal otherwise set for all");
	struct arg_int *lev = arg_int0(NULL,NULL,NULL,"0=none, 1=err, 2=warn, 3=info, 4=debug, 5=verbose");
    struct arg_end *end = arg_end(5);
} log_commandArgs;


#define TAG_MAXLEN 30
static int log_command(FILE *fd, FILE *, int, char**) {
    char tag[TAG_MAXLEN+1];
    esp_log_level_t level = ESP_LOG_NONE;

    if (log_commandArgs.lev->count < 1) {
    	arg_print_glossary(fd, (void**) &log_commandArgs, NULL);
    	return 0;
    }

    strcpy(tag,"*");
    if (log_commandArgs.tag->count) {
    	strncpy(&tag[0], (const char *)log_commandArgs.tag->sval[0], TAG_MAXLEN);
    	tag[TAG_MAXLEN] = '\0';
    }

    switch (log_commandArgs.lev->ival[0]) {
    	case 0:	level = ESP_LOG_NONE;
    			break;
    	case 1:	level = ESP_LOG_ERROR;
    			break;
    	case 2:	level = ESP_LOG_WARN;
    			break;
    	case 3:	level = ESP_LOG_INFO;
    			break;
    	case 4:	level = ESP_LOG_DEBUG;
    			break;
    	case 5:	level = ESP_LOG_VERBOSE;
    			break;
    	default:
    		fprintf(fd, CONS_COLOR_STR_E_N("Invalid loglevel %d"),log_commandArgs.lev->ival[0]);
    		return 0;
    }

    esp_log_level_set(&tag[0], level);

    configfile.set_u32("LOGLEVEL", 1000 + (int)level);
    configfile.save();
    fprintf(fd, CONS_COLOR_STR_I_N("loglevel set to %d"),log_commandArgs.lev->ival[0]);
    return 0;
}

//****************************************************************************************
#ifdef CONFIG_COMMANDS_ADCREF
static struct {
    struct arg_int *gpio_Pin = arg_int0(NULL, NULL, NULL, "gpio pin");
    struct arg_end *end = arg_end(5);
} adcref_commandArgs;


int adcref_command(FILE *fp, FILE *, int, char**)
{
	esp_err_t status;

    gpio_num_t pin = (gpio_num_t)adcref_commandArgs.gpio_Pin->ival[0];
    fprintf(fp,"%s: setting adc refvoltage to pin %d", TAG, pin);
#ifdef ESP32
    status = adc2_vref_to_gpio(pin);
#else
    status = 1;
#endif
    if (status == ESP_OK){
        fprintf(fp, "v_ref routed to GPIO\n");
    }else{
        fprintf(fp, "failed to route v_ref\n");
    }
    return 0;
}
#endif /* CONFIG_COMMANDS_ADCREF */

//****************************************************************************************
#ifdef CONFIG_COMMANDS_ADCREAD
static struct {
    struct arg_int *width			= arg_int0("w","width", "", "adc width 9,10,11 or 12 Bit");
    struct arg_int *channel			= arg_int0("c","channel", "", "adc1 channel 0..7");
    struct arg_int *atten			= arg_int0("a","atten", "", "adc1 attenuation 0,2,6 or 11 where 2 will set 2.5db");
    struct arg_int *count			= arg_int0("n","count", "", "run n times (min=2) and show min, max, avg, attenuation needs to be specified");
    struct arg_end *end = arg_end(5);
} adcreadCommandArgs;

#define DEFAULT_VREF   1100

int adcread_command(FILE * fd, FILE *, int, char**) {
	adc_bits_width_t w;
    adc_atten_t a_start, a_end;
    adc1_channel_t ch;
    int argErr = 0;
    int n=2;
    int i;
#ifdef ESP32
    char at[5];

#endif

    if (adcreadCommandArgs.channel == 0) argErr++;


    i = CONFIG_COMMANDS_ADCREAD_WIDTH; // ad set via menuconfig
    if (adcreadCommandArgs.width->count > 0) i = adcreadCommandArgs.width->ival[0];
     w = ADC_WIDTH_BIT_12; /* suppress warning */
    if (adcreadCommandArgs.width->count > 0) {
    	switch (adcreadCommandArgs.width->ival[0]) {
			case (9):	w = ADC_WIDTH_BIT_9; break;
			case (10):	w = ADC_WIDTH_BIT_10; break;
			case (11):	w = ADC_WIDTH_BIT_11; break;
			case (12):	w = ADC_WIDTH_BIT_12; break;
			default:
				fprintf(fd,CONS_COLOR_STR_E_N("invalid width specified"));
				return 0;
    	}
	}
#ifdef ESP32
    adc1_config_width(w);
#else
    fprintf(fd, "adc1_config_width(%d);\r\n",(int) w);
#endif
    a_start = ADC_ATTEN_DB_0;
    a_end = ADC_ATTEN_DB_11;
    if (adcreadCommandArgs.atten->count > 0) {
    	switch (adcreadCommandArgs.atten->ival[0]) {
    		case (0):	a_start = ADC_ATTEN_DB_0; break;
			case (2):	a_start = ADC_ATTEN_DB_2_5; break;
			case (6):	a_start = ADC_ATTEN_DB_6; break;
			case (11):	a_start = ADC_ATTEN_DB_11; break;
			default: argErr++;
    	}
    	a_end = a_start;
    }
	ch = (adc1_channel_t)adcreadCommandArgs.channel->ival[0];
	if (ch > 7) argErr++;

	if (adcreadCommandArgs.count->count) {
		if (adcreadCommandArgs.atten->count < 1) argErr++;
		if (adcreadCommandArgs.atten->ival[0] < 2) argErr++;
		n = adcreadCommandArgs.count->ival[0];
	}

	if (argErr) {
		arg_print_glossary(fd, (void**) &adcreadCommandArgs, NULL);
    	return 0;
    }

#ifdef ESP32
	gpio_num_t gnum;
	esp_adc_cal_characteristics_t *adc_chars;

	for (int a = (int)a_start; a <= (int)a_end; a++) {

		//Characterize ADC
		adc_chars = (esp_adc_cal_characteristics_t *)calloc(1, sizeof(esp_adc_cal_characteristics_t));
		esp_adc_cal_characterize(ADC_UNIT_1, (adc_atten_t)a, w, DEFAULT_VREF, adc_chars);

		adc1_config_channel_atten(ch, (adc_atten_t)a);

		//esp_adc_cal_get_characteristics(V_REF, (adc_atten_t)a, w, &characteristics);

		uint32_t voltageAvg, voltage, voltageMin, voltageMax;
		voltage = 0; voltageMin = 9999; voltageMax = 0; voltageAvg = 0;
		if (adcreadCommandArgs.atten->count) {
			for (i=0; i<n; i++) {
				voltage = adc1_get_raw(ch);
				voltageAvg += voltage;
				if (voltage < voltageMin) voltageMin = voltage;
				if (voltage > voltageMax) voltageMax = voltage;
			}
			voltage = voltageAvg / adcreadCommandArgs.count->ival[0];
		} else
			voltage = adc1_get_raw(ch);

		voltage = esp_adc_cal_raw_to_voltage(voltage, adc_chars);
		free(adc_chars);
		adc1_pad_get_io_num(ch, &gnum);

		switch ((adc_atten_t)a) {
			case (ADC_ATTEN_DB_0)  : strcpy(at,"  0"); break;
			case (ADC_ATTEN_DB_2_5): strcpy(at,"2.5"); break;
			case (ADC_ATTEN_DB_6)  : strcpy(at,"  6"); break;
			case (ADC_ATTEN_DB_11) : strcpy(at," 11"); break;
			default: strcpy(at,"0");
		}
		fprintf(fd, "%d mV attenuation %s (adc1, channel %d, gpio-port: %d, count: %d)\r\n",voltage, at, ch, gnum, n);
		if (adcreadCommandArgs.atten->count) fprintf(fd," min voltage: %d, max voltage: %d\r\n",voltageMin, voltageMax);
	}
#else
    //fprintf(fd, "adc1_config_channel_atten(%d, %d);\r\n",(int) ch, (int) a);
#endif
    return 0;
}

#endif /* CONFIG_COMMANDS_ADCREAD */

//****************************************************************************************

#ifdef CONFIG_COMMANDS_I2C_SCAN

static struct {
#if defined (CONFIG_COMMANDS_I2C_0) && defined (CONFIG_COMMANDS_I2C_1)
	struct arg_int *port = arg_int0("p","port", "", "i2c num (0 or 1)");
#endif
    struct arg_end *end = arg_end(5);
} i2cscan_commandArgs;


int i2cscan_command(FILE *fd, FILE *, int, char**) {

#if defined (CONFIG_COMMANDS_I2C_0) && defined (CONFIG_COMMANDS_I2C_1)
	i2c_port_t I2C_MASTER_NUM = I2C_NUM_0;
	if (i2cscan_commandArgs.port->count) {
		I2C_MASTER_NUM = (i2c_port_t)i2cscan_commandArgs.port->ival[0];
		if ((int)I2C_MASTER_NUM > 1) {
			fprintf(fd,CONS_COLOR_STR_E_N("i2c scanner: i2c port can only be 0 or 1"));
			return 0;
		}
	}
#else
#if defined (CONFIG_COMMANDS_I2C_0)
#define I2C_MASTER_NUM I2C_NUM_0
#endif
#if defined (CONFIG_COMMANDS_I2C_1)
#define I2C_MASTER_NUM I2C_NUM_1
#endif
#endif

	fprintf(fd,CONS_COLOR_STR_I_N("i2c scanner: getting global lock"));
	globalLockTake(I2C_MASTER_NUM==I2C_NUM_1 ? GLOCK_I2C_1 : GLOCK_I2C_0);
	fprintf(fd,"ok\r\n");

#ifdef ESP32
	int i;
	esp_err_t espRc;
	fprintf(fd,"     0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f\r\n");
	fprintf(fd,"00:         ");
	for (i=3; i< 0x78; i++) {
		i2c_cmd_handle_t cmd = i2c_cmd_link_create();
		i2c_master_start(cmd);
		i2c_master_write_byte(cmd, (i << 1) | I2C_MASTER_WRITE, 1 /* expect ack */);
		i2c_master_stop(cmd);

		espRc = i2c_master_cmd_begin(I2C_MASTER_NUM, cmd, 10/portTICK_PERIOD_MS);
		if (i%16 == 0) {
			fprintf(fd,"\r\n%.2x:", i);
		}
		if (espRc == 0) {
			fprintf(fd," %.2x", i);
		} else {
			fprintf(fd," --");
		}
		//ESP_LOGD(tag, "i=%d, rc=%d (0x%x)", i, espRc, espRc);
		i2c_cmd_link_delete(cmd);
	}
#else
	fprintf(fd,"Sorry, ESP only\r\n");
#endif
	globalLockGive(I2C_MASTER_NUM==I2C_NUM_1 ? GLOCK_I2C_1 : GLOCK_I2C_0);
	fprintf(fd,"\r\n");
	return 0;
}
#endif /* CONFIG_COMMANDS_I2C_SCAN */
//****************************************************************************************

//const char* prompt = "evse> ";
const char * prompt = CONFIG_COMMANDS_PROMPT "> ";

void commandLoop(FILE *fd_stdin, FILE *fd_stdout) {
	linenoise_t *l;
	int ret;
	char* line;
	int fno_stdout;
	int quit = 0;

	l = new linenoise_t();
	l->setMultiLine(0);
	fno_stdout = fileno(fd_stdout);
	l->setFileNo(fileno(fd_stdin),fno_stdout);  /* linennoise uses fileno instead of *FILE */
	l->historySetMaxLen(20);

	//l->printKeyCodes();

	do {
		line = l->run(prompt, &ret);
		fprintf(fd_stdout,"\r\n");
		if (ret == -2) {	// ^D
			quit++;
			if (line != NULL) l->freeBuffer(line);
		} else {
			if (line != NULL) {
				l->historyAdd(line);
				console_run(fd_stdout, fd_stdin, line, &ret);
				if (ret == RETCODE_EXIT) quit++;
				l->freeBuffer(line);
			}
		}
		fflush(fd_stdout);
	} while (!quit);
	delete l;
}


static int exit_command(FILE *, FILE *, int, char**) {
	return RETCODE_EXIT;
}

//*************************************************************************************************************
#ifdef CONFIG_COMMANDS_TELNET_SERVER

xTaskHandle *pvTelnetListenerTask = NULL;

struct telnetClients_t {
	TaskHandle_t taskHandle;
	int          socketHandle;
};

telnetClients_t telnetClients[CONFIG_COMMANDS_TELNET_SERVER_CLIENTS];


void telnetCloseAll() {
	int i;

	for (i=0;i<CONFIG_COMMANDS_TELNET_SERVER_CLIENTS;i++) {
		if (telnetClients[i].socketHandle != 0) {
			close(telnetClients[i].socketHandle);
			telnetClients[i].socketHandle = 0;
		}
	}
	vTaskDelay(pdMS_TO_TICKS(250));
}


void telnetClientTaskDelete() {   /* delete this task and mark the slot as free */
	TaskHandle_t myself;
	int i;
	int slot = -1;

	myself = xTaskGetCurrentTaskHandle();
	if (myself == NULL) {
		ESP_LOGE(TAG,"%s: xTaskGetCurrentTaskHandle returned NULL ????", __func__);
		vTaskDelete(NULL);
	}
	for (i=0;i<CONFIG_COMMANDS_TELNET_SERVER_CLIENTS;i++) {
		if (telnetClients[i].taskHandle == myself) slot=i;
	}
	if (slot < 0) {
		ESP_LOGE(TAG,"%s: unable to find myself ????", __func__);
		vTaskDelete(NULL);
	}
	telnetClients[slot].taskHandle = NULL;
	close(telnetClients[slot].socketHandle);
	telnetClients[slot].socketHandle = 0;
	ESP_LOGI(TAG,"telnet client %d terminated",slot);
	//for (i=0;i<CONFIG_COMMANDS_TELNET_SERVER_CLIENTS;i++) {
	//	printf("telnetClients[%d]=%p\n",i,telnetClients[i]);
	//}
	vTaskDelete(NULL);
}

const char * noFreeSlotsMSG = "no free slots for starting a new session\r\n";
const char * noTaskMSG = "failed to start client task\r\n";



/* keep it simple: initialize telnet session, most important: disable remote echo

sample from DD-WRT and linux telnet:
SERVER CONNECTION ESTABLISHED
SERVER IAC DO 1 (ECHO)
SERVER IAC DO 31 (NAWS)
SERVER IAC WILL 1 (ECHO)
SERVER IAC WILL 3 (SGA)
CLIENT IAC WONT 1 (ECHO)
CLIENT IAC WILL 31 (NAWS)
CLIENT SUB 31 (NAWS) [4 bytes]:  [<0x00><0x50><0x00><0x18>]
CLIENT IAC DO 1 (ECHO)
CLIENT IAC DO 3 (SGA)
SERVER DATA:
 [<0x0D><0x0D><0x0A>
]
SERVER DATA: DD-WR
 *
 * */
void telnetClientInit(int fd) {
	uint8_t myOptions [] = {
			TELNET_IAC, TELNET_DO, TELNET_TELOPT_ECHO,		// client should no echo
			TELNET_IAC, TELNET_DONT, TELNET_TELOPT_NAWS,	// window size, we should disable this
			TELNET_IAC, TELNET_WILL,TELNET_TELOPT_ECHO,
			TELNET_IAC, TELNET_WILL,TELNET_TELOPT_SGA
	};
	uint8_t reply[255];

	write(fd, (void *)&myOptions, sizeof(myOptions));
	YIELDS();

	// we should get at least one packet from the client with answers
	recv(fd, (void *)&reply, sizeof(reply), 0);

/*
	ESP_LOGD(TAG, "telnetClientInit response for fd %d, size %d (IGNORED, ASSUMED TO BE OK)",fd,rc);
	uint8_t *c;

	if (rc > 0) {
		c = &reply[0];
		for (int i=0; i<rc; i++) {
			printf("%02x ",*c); c++;
		}
		printf("\n");
	}
*/
	YIELDS();
}


extern "C" void telnetClientTask(void * slot) {
	FILE *in;
	FILE *out;

	int fd = telnetClients[(int)slot].socketHandle;

	telnetClientInit(fd);

	ESP_LOGD(TAG, "telnetClientTask started with fd %d",fd);

// Set the socket I/O mode: In this case FIONBIO
    // enables or disables the blocking mode for the
    // socket based on the numerical value of iMode.
    // If iMode = 0, blocking is enabled;
    // If iMode != 0, non-blocking mode is enabled.
	//iMode = 0;
    //ioctl(fd, FIONBIO, &iMode);


	out = fdopen(fd, "w");
	if (out == NULL) {
		ESP_LOGE(TAG,"%s: fdopen(out) failed, errno: %d",__func__,errno);
		telnetClientTaskDelete();
	}
	showWelcomeMsg(out);
	//write(fd,greetingMSG,strlen(greetingMSG)); YIELD();

	in = fdopen(fd, "r");  // dup it first ? -> esp32 has no dup !
	if (in == NULL) {
		ESP_LOGE(TAG,"%s: fdopen(in) failed, errno: %d",__func__,errno);
		telnetClientTaskDelete();
	}

	//fprintf(out,"starting commandLoop\r\n");
	commandLoop(in,out); YIELDS();
	//fprintf(out, "telnetClientTask terminating\r\n"); fflush(out); YIELD();
	fclose(out); fclose(in);
	telnetClientTaskDelete();
}


// returns listenfd or -1 on error
int setupListenSocket() {
	int listenfd ;
    int res=0;
    int on;
    struct sockaddr_in6 serv_addr;
    char sendBuff [255];

    memset(&serv_addr, '0', sizeof(serv_addr));
    memset(sendBuff, '0', sizeof(sendBuff));
    //listenfd = socket(AF_INET, SOCK_STREAM, 0);
	listenfd = socket(AF_INET6, SOCK_STREAM, 0);
    if (listenfd < 0)
    {
  	 	 ESP_LOGD(TAG,"socket failed with code %d, errno: %d",res, errno);
  	 	 return -1;
	}
    YIELDS();

#ifndef ESP32
    /****************************************************************************/
	/* The setsockopt() function is used to allow the local address to          */
	/* be reused when the server is restart_commanded before the required wait  */
	/* time expires.                                                            */
	/****************************************************************************/
    on=1;
    if (setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR,
                      (char *)&on,sizeof(on)) < 0)
    {
         ESP_LOGE(TAG,"setsockopt(SO_REUSEADDR) failed, errno: %d",errno);
    }
    YIELD();
#endif

    /********************************************************************/
    // on=1 would allow ipv6 only. Otherwise both, ipv4 and ipv6 clients
    // can connect
	/********************************************************************/
    on=0;
    res = setsockopt(listenfd, IPPROTO_IPV6, IPV6_V6ONLY,(char *)&on,sizeof(on));
    if (res < 0)
    {
  	 	 ESP_LOGE(TAG,"setsockopt(IPV6_V6ONLY = %d) failed with code %d, errno: %d",on,res,errno);
	}
    YIELDS();

       /*
    serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	serv_addr.sin_port = htons(5555);  */
	serv_addr.sin6_family = AF_INET6;
	serv_addr.sin6_port   = htons(LISTEN_PORT);
	/********************************************************************/
	/* Note: applications use in6addr_any similarly to the way they use */
	/* INADDR_ANY in IPv4.  A symbolic constant IN6ADDR_ANY_INIT also   */
	/* exists but can only be used to initialize an in6_addr structure  */
	/* at declaration time (not during an assignment).                  */
	/********************************************************************/
    serv_addr.sin6_addr   = in6addr_any;

    res = bind(listenfd, (struct sockaddr*)&serv_addr, sizeof(serv_addr));
    if (res < 0)
    {
  	 	 ESP_LOGE(TAG,"bind failed with code %d, errno: %d",res, errno);
  	 	 close(listenfd);
  	 	 return -1;
	}
    YIELDS();

    res = listen(listenfd, 2);
    if (res < 0)
    {
  	 	 ESP_LOGE(TAG,"listen failed with code %d, errno: %d",res, errno);
  	 	 close(listenfd);
  	 	 return -1;
	}
    YIELDS();
    return listenfd;
}


// TODO: check if this works if we reconnect the network without restart_commanding this task or if this task is started before network is active
extern "C" void telnetListenerTask(void * UNUSED(param)) {
	int listenfd = -1, connfd = 0;
    //char sendBuff [255];
    int i, slot, res=0;
    char name[16];

    do {
    	if (listenfd < 0) listenfd = setupListenSocket();
    	if (listenfd < 0)
    	{
  	 	 	 ESP_LOGE(TAG,"socket failed with code %d, errno: %d",listenfd, errno);
  	 	 	 vTaskDelay(pdMS_TO_TICKS(3000));
		} else {
			YIELDS();
			connfd = accept(listenfd, (struct sockaddr*)NULL, NULL);
			if (connfd < 0) {
				res = errno;
				if (res != EINTR) {
					ESP_LOGE(TAG,"accept failed with code %d, errno: %d",connfd, res);
					close(listenfd); listenfd = -1;
				} else YIELD();
			} else {
				// new client, lets find a free slot
				slot = -1;
				for (i=0;i<CONFIG_COMMANDS_TELNET_SERVER_CLIENTS;i++) {
					if (telnetClients[i].taskHandle == NULL) slot=i;
				}
				if (slot < 0) {
					write(connfd,noFreeSlotsMSG,strlen(noFreeSlotsMSG));
					close(connfd);
					ESP_LOGE(TAG,"rejected client connection due to no slot available");
				} else {
					//ESP_LOGI(TAG,"starting telnet client task using fno %d",connfd);
					sprintf(name,"telnetClient %d",slot);
					telnetClients[slot].socketHandle = connfd;
					xTaskCreatePinnedToCore(telnetClientTask, name, 8192, (void *)slot, STD_TASK_PRIO, &telnetClients[slot].taskHandle,APP_CPU);

					//printf("telnetClients[%d]:%p\n",slot,telnetClients[slot]);
					if (telnetClients[slot].taskHandle == NULL) {
						ESP_LOGE(TAG,"failed to start telnet client task");
						write(connfd,noTaskMSG,strlen(noTaskMSG));
						close(connfd);
					}
				}
			}
		}
    } while (1);
}

#endif /* CONFIG_COMMANDS_TELNET_SERVER */
//*************************************************************************************************************


void esp32_initialize_console()
{
#ifdef ESP32
    /* Disable buffering on stdin and stdout */
    setvbuf(stdin, NULL, _IONBF, 0);
    setvbuf(stdout, NULL, _IONBF, 0);

    /* Minicom, screen, idf_monitor send CR when ENTER key is pressed */
    esp_vfs_dev_uart_set_rx_line_endings(ESP_LINE_ENDINGS_CR);
    /* Move the caret to the beginning of the next line on '\n' */
    esp_vfs_dev_uart_set_tx_line_endings(ESP_LINE_ENDINGS_CRLF);

    /* Install UART driver for interrupt-driven reads and writes */
    ESP_ERROR_CHECK( uart_driver_install((uart_port_t)CONFIG_CONSOLE_UART_NUM, 256, 0, 0, NULL, 0) );

    /* Tell VFS to use UART driver */
    esp_vfs_dev_uart_use_driver(CONFIG_CONSOLE_UART_NUM);
#endif // ESP32
}

void initialize_filesystem()
{
#ifdef ESP32
    static wl_handle_t wl_handle;
    const esp_vfs_fat_mount_config_t mount_config = {
    		.format_if_mount_failed = true,
            .max_files = 4,
			.allocation_unit_size = 0
    };
    esp_err_t err = esp_vfs_fat_spiflash_mount(MOUNT_PATH, "storage", &mount_config, &wl_handle);
    if (err != ESP_OK) {
        ESP_LOGE(TAG, "Failed to mount FATFS (0x%x)", err);
        return;
    }
    ESP_LOGI(TAG, "FATFS mounted");
#endif // ESP32
}

#ifdef CONFIG_COMMANDS_I2C
#ifdef ESP_PLATFORM
void i2c_init() {
	int err;
	i2c_config_t conf;


#ifdef CONFIG_COMMANDS_I2C_0
	printf("init i2c #0 (sda:%d scl:%d)\n",CONFIG_COMMANDS_I2C_0_SDA,CONFIG_COMMANDS_I2C_0_SCL);
	memset(&conf,0,sizeof(conf));
	conf.mode = I2C_MODE_MASTER;
	conf.sda_io_num = (gpio_num_t)CONFIG_COMMANDS_I2C_0_SDA;
	conf.scl_io_num = (gpio_num_t)CONFIG_COMMANDS_I2C_0_SCL;
#ifdef CONFIG_COMMANDS_I2C_0_SDA_PULLUP
	conf.sda_pullup_en = GPIO_PULLUP_ENABLE;
#endif
#ifdef CONFIG_COMMANDS_I2C_0_SCL_PULLUP
	conf.scl_pullup_en = GPIO_PULLUP_ENABLE;
#endif
	conf.master.clk_speed = CONFIG_COMMANDS_I2C_0_FREQ;

	err = i2c_param_config(I2C_NUM_0, &conf);
	ESP_LOGV(TAG,"i2c_param_config returned %d (sda:%d scl:%d dpeed:%d)",err,conf.sda_io_num,conf.scl_io_num,conf.master.clk_speed);

	if( err != ESP_OK ) {
		ESP_LOGE(TAG,"i2c_param_config returned %d",err);
	} else {
		// install the driver
		err = i2c_driver_install(I2C_NUM_0, I2C_MODE_MASTER, 0,0,0);
		if(err != ESP_OK) {
			ESP_LOGE(TAG,"I2C driver install for i2c 0 failed, err:%d",err);
		}
	}
#endif /* CONFIG_COMMANDS_I2C_0 */

#ifdef CONFIG_COMMANDS_I2C_1
	printf("init i2c #1 (sda:%d scl:%d)\n",CONFIG_COMMANDS_I2C_1_SDA,CONFIG_COMMANDS_I2C_1_SCL);
	memset(&conf,0,sizeof(conf));
	conf.mode = I2C_MODE_MASTER;
	conf.sda_io_num = (gpio_num_t)CONFIG_COMMANDS_I2C_1_SDA;
	conf.scl_io_num = (gpio_num_t)CONFIG_COMMANDS_I2C_1_SCL;
#ifdef CONFIG_COMMANDS_I2C_1_SDA_PULLUP
	conf.sda_pullup_en = GPIO_PULLUP_ENABLE;
#endif
#ifdef CONFIG_COMMANDS_I2C_1_SCL_PULLUP
	conf.scl_pullup_en = GPIO_PULLUP_ENABLE;
#endif
	conf.master.clk_speed = CONFIG_COMMANDS_I2C_1_FREQ;

	err = i2c_param_config(I2C_NUM_1, &conf);
	ESP_LOGV(TAG,"i2c_param_config returned %d (sda:%d scl:%d dpeed:%d)",err,conf.sda_io_num,conf.scl_io_num,conf.master.clk_speed);

	if( err != ESP_OK ) {
		ESP_LOGE(TAG,"i2c_param_config returned %d",err);
	} else {
		// install the driver
		err = i2c_driver_install(I2C_NUM_1, I2C_MODE_MASTER, 0,0,0);
		if(err != ESP_OK) {
			ESP_LOGE(TAG,"I2C driver install for i2c 1 failed, err:%d",err);
		}
	}
#endif /* CONFIG_COMMANDS_I2C_1 */

#ifdef CONFIG_COMMANDS_I2C_MCP23017_0
	pio_init(0, CONFIG_COMMANDS_I2C_MCP23017_0_ADDR, (i2c_port_t)CONFIG_COMMANDS_I2C_MCP23017_0_PORT);
#endif

#ifdef CONFIG_COMMANDS_I2C_MCP23017_1
	pio_init(1, CONFIG_COMMANDS_I2C_MCP23017_1_ADDR, (i2c_port_t)CONFIG_COMMANDS_I2C_MCP23017_1_PORT);
#endif

} /* i2c_init */

#else /* ESP_PLATFORM */

void i2c_init() {

}
#endif /* ESP_PLATFORM */
#endif /* CONFIG_COMMANDS_I2C */


void commands_Initialize(void) {
	esp32_initialize_console();
	printf("commands_Initialize\n");

#ifdef CONFIG_COMMANDS_TELNET_SERVER
	memset((void *)&telnetClients,0,sizeof(telnetClients));
#endif

	console_config_t console_config = {
			  .max_cmdline_length = 256
            , .max_cmdline_args = 8
			, .hint_color = atoi(CONS_COLOR_CYAN)
			, .hint_bold = 1
	};

    console_init(&console_config);
	console_register_help_command();

#ifdef CONFIG_COMMANDS_I2C
	i2c_init();
#endif

#ifdef CONFIG_COMMANDS_ADCREF
    adcref_commandArgs.gpio_Pin->ival[0] = CONFIG_COMMANDS_ADCREF_GPIO;  // default pin
    CMD_REG(adcref,"Set adcref voltage to given pin",NULL,&adcref_command,&adcref_commandArgs);
#endif

#ifdef CONFIG_COMMANDS_ADCREAD
    //adc_read_console_args.width->ival[0] = 12;
    CMD_REG(adcread,"Read ADC",NULL,&adcread_command, &adcreadCommandArgs);
#endif

	CMD_REG(exit		, "close session", NULL, &exit_command, NULL);

#ifdef CONFIG_COMMANDS_RESTART
	CMD_REG(restart		, "restart system", NULL, &restart_command, NULL);
#endif

	CMD_REG(ls			, "list directory of flash fat partition", NULL, &ls_command, NULL);
	CMD_REG(cat			, "type a file", NULL, &cat_command, &cat_commandArgs);
	CMD_REG(pwd			,"print current directory", NULL, &pwd_command, NULL);

#ifdef CONFIG_COMMANDS_FREE
	CMD_REG(free		, "Get the total size of heap memory available", NULL, &free_command, NULL);
#endif

#if defined (ESP32) && defined (CONFIG_COMMANDS_PING)
	CMD_REG2(ping       ,"ping destination",NULL);
#endif



#ifdef CONFIG_COMMANDS_WLAN
	CMD_REG2(wlan		,"connect to wlan access point", NULL);
#endif

#ifdef CONFIG_COMMANDS_ETH
	CMD_REG2(eth 		,"connect ethernet", NULL);
#endif
	CMD_REG(date		,"show date/time", NULL, &date_command, NULL);
	CMD_REG2(delay		,"delay ms", NULL);

#ifdef CONFIG_COMMANDS_INOUT
	CMD_REG(in			,"read gpio port", NULL, &in_command, &in_commandArgs);
	CMD_REG(out			,"write gpio port", NULL, &out_command, &out_commandArgs);
#endif

#ifdef CONFIG_COMMANDS_I2C_MCP
	CMD_REG2(mcp		,"set/read mcp23017", NULL);
#endif

#ifdef CONFIG_COMMANDS_I2C_SCAN
	CMD_REG2(i2cscan	,"scan for i2c devices", NULL);
#endif

#ifdef CONFIG_COMMANDS_UPDATE
	CMD_REG2(update		,"update firmware from a http-server", NULL);
#endif

	CMD_REG2(log        ,"set esp log level", NULL);

#ifdef CONFIG_COMMANDS_PS
#if ( configUSE_TRACE_FACILITY == 1 )
	CMD_REG(ps			, "show running tasks", NULL, &ps_command, NULL);
#endif
#endif

    CMD_REG(version		,"show version", NULL, &ver_command, NULL);


#ifdef CONFIG_COMMANDS_TELNET_SERVER
	xTaskCreatePinnedToCore(telnetListenerTask, "telnetServer", 4096, NULL, STD_TASK_PRIO, (TaskHandle_t * const)&pvTelnetListenerTask,APP_CPU);
	if (pvTelnetListenerTask == NULL) {
		ESP_LOGE(TAG,"unable to start telnetListenerTask");
	}
#endif
}  /* commands_Initialize */


extern "C" char* ver_getVersionStrLong();

void showWelcomeMsg(FILE*fp) {
	fprintf(fp,CONS_COLOR_I "%s" CONS_RESET_COLOR "\r\n" \
		"type " CONS_BOLD(CONS_COLOR_WHITE) "help" CONS_RESET_COLOR " for a list of available commands\r\n\n",ver_getVersionStrLong()); fflush(fp);
}
