/*
 * network.cpp
 * Network (wlan and eth) init and connect funtions for esp32
 *
 *  Created on: Feb 9, 2018
 *      Author: Armin Diehl <ad@ardiehl.de>
 */

#include "network.h"

#include <ctype.h>
#include <stdbool.h>
#include <stdint.h>
//#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>


#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/queue.h"
#include "freertos/event_groups.h"
#include "freertos/timers.h"

#include "configfile.h"

#ifdef ESP32
#include "esp_err.h"
#include "esp_system.h"
#include "esp_log.h"
//#include "sdkconfig.h"
#ifdef CONFIG_COMMANDS_WLAN
#include "esp_wifi.h"
#endif
#include "tcpip_adapter.h"
#include "esp_event_loop.h"
#include "apps/sntp/sntp.h"

extern "C" esp_err_t esp_event_handler(void *, system_event_t *event);
#else
#include "esp_compat.h"
#endif

#include "esp_event.h"

#ifdef CONFIG_COMMANDS_NETWORK

#ifndef APP_CPU
#define APP_CPU 1
#endif

static const char* TAG = "network";

EventGroupHandle_t network_event_group;
int CONNECTED_BIT = BIT0;

extern "C" void networkConnectEvent(int connected);

/**** ETHERNET **************************************************************************************/

#ifdef CONFIG_COMMANDS_ETH

#ifdef CONFIG_PHY_LAN8720
#include "eth_phy/phy_lan8720.h"
#define DEFAULT_ETHERNET_PHY_CONFIG phy_lan8720_default_ethernet_config
#endif

#ifdef CONFIG_PHY_TLK110
#include "eth_phy/phy_tlk110.h"
#define DEFAULT_ETHERNET_PHY_CONFIG phy_tlk110_default_ethernet_config
#endif

#ifndef DEFAULT_ETHERNET_PHY_CONFIG
#error no PHY defined for ethernet
#endif

#define PIN_PHY_POWER CONFIG_PHY_POWER_PIN
#define PIN_SMI_MDC   CONFIG_PHY_SMI_MDC_PIN
#define PIN_SMI_MDIO  CONFIG_PHY_SMI_MDIO_PIN

#ifdef CONFIG_PHY_USE_POWER_PIN
/* This replaces the default PHY power on/off function with one that
   also uses a GPIO for power on/off.

   If this GPIO is not connected on your device (and PHY is always powered), you can use the default PHY-specific power
   on/off function rather than overriding with this one.
*/
extern "C" void phy_device_power_enable_via_gpio(bool enable)
{
    assert(DEFAULT_ETHERNET_PHY_CONFIG.phy_power_enable);

    if (!enable) {
        /* Do the PHY-specific power_enable(false) function before powering down */
        DEFAULT_ETHERNET_PHY_CONFIG.phy_power_enable(false);
    }

    gpio_pad_select_gpio(PIN_PHY_POWER);
    gpio_set_direction((gpio_num_t)PIN_PHY_POWER,GPIO_MODE_OUTPUT);
    if(enable == true) {
        gpio_set_level((gpio_num_t)PIN_PHY_POWER, 1);
        //ESP_LOGD(TAG, "phy_device_power_enable(TRUE)");
    } else {
        gpio_set_level((gpio_num_t)PIN_PHY_POWER, 0);
        //ESP_LOGD(TAG, "power_enable(FALSE)");
    }

    // Allow the power up/down to take effect, min 300us
    vTaskDelay(1);

    if (enable) {
        /* Run the PHY-specific power on operations now the PHY has power */
        DEFAULT_ETHERNET_PHY_CONFIG.phy_power_enable(true);
    }
}
#endif

extern "C" void eth_gpio_config_rmii(void)
{
    // RMII data pins are fixed:
    // TXD0 = GPIO19
    // TXD1 = GPIO22
    // TX_EN = GPIO21
    // RXD0 = GPIO25
    // RXD1 = GPIO26
    // CLK == GPIO0
    phy_rmii_configure_data_interface_pins();
    // MDC is GPIO 23, MDIO is GPIO 18
    phy_rmii_smi_configure_pins(PIN_SMI_MDC, PIN_SMI_MDIO);
}

#endif
/****************************************************************************************************/


char * getGateway (void) {
#ifdef ESP32
	tcpip_adapter_ip_info_t ip;
	char *res;

	if (tcpip_adapter_get_ip_info(TCPIP_ADAPTER_IF_ETH, &ip) == 0) {
		if (ip.gw.addr != 0) { asprintf(&res,IPSTR,IP2STR(&ip.gw)); return res; }
	}
	if (tcpip_adapter_get_ip_info(TCPIP_ADAPTER_IF_STA, &ip) == 0) {
		if (ip.gw.addr != 0) { asprintf(&res,IPSTR,IP2STR(&ip.gw)); return res; }
	}
#endif
	return NULL;
}


void initialize_sntp(void)
{
#ifdef ESP32
	//ESP_LOGE(TAG,"initialize_sntp");
	sntp_stop();
	char * serverName;
	serverName = configfile.get_str ("NTP_SERVER");

	if (serverName == NULL) {
		// lets try to use the default gateway. If this is a router, there may be an ntpd or forwarder running, todo: IPV6
		serverName = getGateway();
		if (serverName == NULL) {
			ESP_LOGE(TAG,"got no gateway address and we have no NTP server configured");
			//return;
		}
	}

    sntp_setoperatingmode(SNTP_OPMODE_POLL);
    if (serverName != NULL) {
    	ESP_LOGI(TAG, "Initializing SNTP (%s)",serverName);
    	//sntp_servermode_dhcp(0);
    	sntp_setservername(0, serverName);
    	sntp_init();
    }
    free(serverName);
#endif  // ESP32
}




int connected = 0;

int network_isConnected() {
	return connected;
}

extern "C" esp_err_t esp_event_handler(void *, system_event_t *event)
{
    switch(event->event_id) {
    case SYSTEM_EVENT_STA_GOT_IP:
    case SYSTEM_EVENT_ETH_GOT_IP:
    case SYSTEM_EVENT_GOT_IP6:
        xEventGroupSetBits(network_event_group, CONNECTED_BIT);
        if (connected == 0) {
        	//ev_send (EV_TASK_NET,EV_NET_CONNECTED,EV_DATA_NONE);
        	networkConnectEvent(1);
        	initialize_sntp();
        	connected++;
        }
        break;
    case SYSTEM_EVENT_STA_DISCONNECTED:
    case SYSTEM_EVENT_STA_LOST_IP:
    case SYSTEM_EVENT_ETH_DISCONNECTED:
#ifdef ESP32
        //esp_wifi_connect();  // TODO, reconnect ?
#endif
        xEventGroupClearBits(network_event_group, CONNECTED_BIT);
        if (connected) {
    //    	ev_send (EV_TASK_NET,EV_NET_DISCONNECTED,EV_DATA_NONE);
        	connected = 0;
        	networkConnectEvent(0);
        }
        break;
    default:
        break;
    }
    return ESP_OK;
}

#ifndef ESP32
// for testing from linux ui
void sendEspNetworkEvent (int connected) {
	system_event_t event;

	memset((void *)&event,0, sizeof(event));

	if (connected) {
		event.event_id = SYSTEM_EVENT_STA_GOT_IP;
	} else {
		event.event_id = SYSTEM_EVENT_STA_LOST_IP;
	}
	esp_event_handler(NULL,&event);
	vTaskDelay(pdMS_TO_TICKS(250));

}
#endif  // ESP32



static bool network_initialized = false;

void network_initialize (FILE *) {
	if (network_initialized == false) {
		network_event_group = xEventGroupCreate();

#ifdef ESP32
		tcpip_adapter_init();				// needed for wlan and eth
		ESP_ERROR_CHECK (esp_event_loop_init(esp_event_handler, NULL));
#endif

		network_initialized = true;
	}
}

#ifdef CONFIG_COMMANDS_WLAN

static bool wifi_initialized = false;

void wifi_initialize (FILE *out) {
	if (wifi_initialized == false) {
		network_initialize (out);
		/***** WLAN *********************************************************************************************/
#ifdef ESP32
		esp_log_level_set("wifi", ESP_LOG_WARN);
		wifi_init_config_t cfg = WIFI_INIT_CONFIG_DEFAULT();
		ESP_ERROR_CHECK( esp_wifi_init(&cfg) );
		ESP_ERROR_CHECK( esp_wifi_set_storage(WIFI_STORAGE_RAM) );
		ESP_ERROR_CHECK( esp_wifi_set_mode(WIFI_MODE_NULL) );   // WIFI_MODE_STA ?
		initialize_sntp();
#endif
		wifi_initialized = true;
	}
}
#endif /* CONFIG_COMMANDS_WLAN */


#ifdef CONFIG_COMMANDS_ETH

static bool eth_initialized = false;

void eth_initialize (FILE *out) {

	if (eth_initialized == false) {
		network_initialize (out);
		/***** ETHERNET *****************************************************************************************/
		eth_config_t config = DEFAULT_ETHERNET_PHY_CONFIG;
		/* Set the PHY address in the example configuration */
		config.phy_addr = (eth_phy_base_t)CONFIG_PHY_ADDRESS;
		config.gpio_config = eth_gpio_config_rmii;
		config.tcpip_input = tcpip_adapter_eth_input;
		config.clock_mode = (eth_clock_mode_t)CONFIG_PHY_CLOCK_MODE;

#ifdef CONFIG_PHY_USE_POWER_PIN
		/* Replace the default 'power enable' function with an example-specific
       	   one that toggles a power GPIO. */
		config.phy_power_enable = phy_device_power_enable_via_gpio;
#endif

		int ret = esp_eth_init(&config);

		if(ret != ESP_OK) {
			ESP_LOGW(TAG, "esp_eth_init failed with rc %d",ret);
		}
		//initialize_sntp();
		eth_initialized = true;
	}
}


bool eth_connect (FILE *out, int timeout_ms) {
//#ifndef ETH_ENABLED
//	//fprintf(fd,"ESP32 would try to connect ethernet using timeout %d\r\n",timeout_ms);
//	return timeout_ms-timeout_ms;
//#else
	ESP_LOGV(TAG,"s eth_connect");
	eth_initialize (out);
	esp_eth_enable();
	if (timeout_ms == 0) timeout_ms = DEFAULT_ETH_CONNECT_TIMEOUT_MS;
    int bits = xEventGroupWaitBits(network_event_group, CONNECTED_BIT, 0, 1, pdMS_TO_TICKS(timeout_ms));
    ESP_LOGV(TAG,"e eth_connect res=%d",bits & CONNECTED_BIT);
    return (bits & CONNECTED_BIT) != 0;
}



#endif /* CONFIG_COMMANDS_ETH */

// ****************************************************************************************************************

#ifdef CONFIG_COMMANDS_WLAN

bool wifi_connect(FILE *fd, const char* ssid, const char* pass, int timeout_ms)
{

	wifi_initialize (fd);
#ifndef ESP32
	fprintf(fd,"ESP32 would try to connect to '%s' using pass '%s' and timeout %d\r\n",ssid,pass,timeout_ms);
	return 0;
#else
    wifi_config_t wifi_config;
    memset(&wifi_config,0,sizeof(wifi_config));
    strncpy((char*) wifi_config.sta.ssid, ssid, sizeof(wifi_config.sta.ssid));
    if (pass) {
        strncpy((char*) wifi_config.sta.password, pass, sizeof(wifi_config.sta.password));
    }

    ESP_ERROR_CHECK( esp_wifi_set_mode(WIFI_MODE_STA) );
    ESP_ERROR_CHECK( esp_wifi_set_config(ESP_IF_WIFI_STA, &wifi_config) );
    ESP_ERROR_CHECK( esp_wifi_start() );
    ESP_ERROR_CHECK( esp_wifi_connect() );

    int bits = xEventGroupWaitBits(network_event_group, CONNECTED_BIT, 0, 1, pdMS_TO_TICKS(timeout_ms));
    return (bits & CONNECTED_BIT) != 0;
#endif  // ESP32
}


// connect to saved SSID using saved password
int network_connectWifi(FILE *fp) {
	char *ssid;
	char *pw;
	uint32_t timeout;
	int res = 0;

	network_initialize(fp);

	ssid = configfile.get_str ("WLAN_SSID");
	if (ssid != NULL) {
       	if (*ssid != 0) {
       		pw = configfile.get_str ("WLAN_PWD");
       		timeout = configfile.get_u32 ("WLAN_TIMEOUT");
       		if (timeout <= 0) timeout = DEFAULT_WIFI_CONNECT_TIMEOUT_MS;
       		ESP_LOGV(TAG,"connecting to '%s' using PW '%s' and timeout %d",ssid,pw ? pw : "<NULL>",timeout);
       		res = wifi_connect(fp, ssid, pw, timeout);
       		ESP_LOGV("network_connect","res:%d",res);
       		free(pw);
       	}
       	free(ssid);
    }
    if (res == 0) {
       	esp_wifi_stop();
	}
	return res;
}

#endif /* CONFIG_COMMANDS_WLAN */


/* if set as default, try to bring up wlan else bring up thernet
 * if wlan fails, try ethernet
 * returns 1 if connected
 */

int network_connect(FILE *fp) {
#ifdef CONFIG_COMMANDS_ETH
	uint32_t timeout;
#endif
	int res = 0;
#ifdef CONFIG_COMMANDS_WLAN
	int useWlan = configfile.get_u32 ("WLAN_DEFAULT");
	if (useWlan) {
		res = network_connectWifi(fp);
		if (res == 0)
			useWlan = 0;	// try ethernet if enabled
	}
#endif

#ifdef CONFIG_COMMANDS_ETH
	int useEth  = configfile.get_u32 ("ETH_ENABLED");
	if (useEth) {
#ifdef CONFIG_COMMANDS_WLAN
		if (useWlan == 0) {				// try to connect via ethernet
#endif
			timeout = configfile.get_u32 ("ETH_TIMEOUT");
			res = eth_connect(fp, timeout); //(fp, timeout);
			if (res == 0) {
				esp_eth_disable();
			}
#ifdef CONFIG_COMMANDS_WLAN
		}
#endif
	}
#endif

	return res;
}



void setTimezone() {
#ifdef ESP32
	char *tz;
	char *tz_cfg;
	char c[20];

	strcpy(c,TZ_DEFAULT);

	tz = &c[0];
	tz_cfg = configfile.get_str("TZ");
	if (tz_cfg) tz=tz_cfg;

	setenv("TZ", tz, 1);
	tzset();
	free(tz_cfg);
#endif  // ESP32
}





/* try to connect every 15 seconds */
extern "C" void network_connector(void *) {
	int sntpRestartCounter = 0;

	vTaskDelay(pdMS_TO_TICKS(2500));
	network_connect(stdout);
	while(1) {
		int bits = xEventGroupWaitBits(network_event_group, CONNECTED_BIT, 0, 1, pdMS_TO_TICKS(15000));
		if ((bits & CONNECTED_BIT) == 0) {
			connected = network_connect(stdout);
			sntpRestartCounter = SNTP_REINIT_WAIT;
		}
		vTaskDelay(NETCON_RETRY_WAIT);	// 10 seconds
		if (sntpRestartCounter) {		// reinitialize sntp every x hours, the auto update in lwip does not seem to work
			sntpRestartCounter--;
			if (sntpRestartCounter==0) {
				initialize_sntp();
				sntpRestartCounter = SNTP_REINIT_WAIT;
			}
		}
	}
}


void networkTask_start(FILE*fp) {  // start a task that will try to connect to the network every x seconds
	setTimezone();
	network_initialize(fp);
	xTaskCreatePinnedToCore(network_connector, "net-connect", 4096, NULL, 5, NULL, APP_CPU);
	vTaskDelay(pdMS_TO_TICKS(250));
}

#ifdef CONFIG_COMMANDS_WLAN
void network_WifiSaveSettings (int wifiDefault, int wifiTimeoutMS, const char * ssid, const char *pwd) {
	configfile.set_str("WLAN_SSID", ssid);
    configfile.set_str("WLAN_PWD", pwd);
    if (wifiTimeoutMS >= 0)
    	configfile.set_u32("WLAN_TIMEOUT", wifiTimeoutMS);
    configfile.set_u32("WLAN_DEFAULT", wifiDefault);
    if (wifiDefault) configfile.set_u32("ETH_ENABLED",0);
    configfile.save();
}

void network_WifiSaveSettings (int wifiDefault) {
	configfile.set_u32("WLAN_DEFAULT", wifiDefault);
	configfile.save();
}

#endif /* CONFIG_COMMANDS_ETH */



#ifdef CONFIG_COMMANDS_ETH

void network_EthSaveSettings (int ethDefault, int ethTimeoutMS) {
	if (ethTimeoutMS >= 0)
		configfile.set_u32("WLAN_TIMEOUT", ethTimeoutMS);
    configfile.set_u32("ETH_ENABLED", ethDefault);
    if (ethDefault) configfile.set_u32("WLAN_DEFAULT",0);
    configfile.save();
}


int network_EthGetTimeout(void) {
	int timeout = configfile.get_u32("WLAN_TIMEOUT");
	if (timeout < 1) timeout = DEFAULT_ETH_CONNECT_TIMEOUT_MS;
	return timeout;
}

#endif /* CONFIG_COMMANDS_ETH */


#else /* CONFIG_COMMANDS_NETWORK */

int network_isConnected() {
	return 0;
}


#endif  /* CONFIG_COMMANDS_NETWORK */



