/*
 * configfile.cpp
 *
 *  Created on: Jan 2, 2018
 *      Author: Armin Diehl <ad@ardiehl.de>
 */


#include "configfile.h"
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>

configfile_t configfile;

char * configfile_t::filename() {					// return current file name, do not touch or free returned string
	return (lastFilename);
}

configfile_t::configfile_t() {
	memset(&lines,0,sizeof(lines));
	lastFilename = NULL;
	changed = 0;
	nLines = 0;
}


int configfile_t::save() {							// save if something was changed
	if (changed) return save (lastFilename);
	return 0;
}

int configfile_t::set_u32 (const char *name, uint32_t value) {
	char buf[20];
	if (value == 0) {
		buf[0] = '\0';
	} else
		sprintf(buf,"%d",value);
	return set_str(name,&buf[0]);
}

int configfile_t::set_str (const char *name, const char *value) {
	int i = findNameIdx(name);

	if (i >= 0) {
		free (lines[i]); lines[i]=NULL;
	}
	changed++;
	// dont save empty strings
	if (value == NULL) return 0;
	if (*value == '\0') return 0;
	add (name,value);

	return 0;
}


int configfile_t::get_u32 (const char *name) {				// 0 if not set
	char *s = findName(name);

	if (!s) return 0;
	return strtol(s,NULL,10);
}

char * configfile_t::get_str (const char *name) {
	char *s = findName(name);
	if (s) return strdup(s);
	return NULL;
}

int configfile_t::add (const char *name, const char *value) {

	for (int i=0; i<CONFIGFILE_MAX_LINES; i++) {
		if (lines[i]==NULL) {
			lines[i] = (char *)malloc (strlen(name)+strlen(value)+2);
			if (lines[i] == NULL) return CONFIGFILE_ERR_NOMEM;
			strcpy(lines[i],name); strcat(lines[i],"="); strcat(lines[i],value);
			if (nLines < i+1) nLines = i+1;
			return 0;
		}
	}
	return CONFIGFILE_ERR_FULL;
}

int configfile_t::add (const char *str) {
	if (strlen(str) > CONFIGFILE_LINE_LEN)	return CONFIGFILE_ERR_TOLONG;

	for (int i=0; i<CONFIGFILE_MAX_LINES; i++) {
		if (lines[i]==NULL) {
			lines[i] = strdup(str);
			if (lines[i] == NULL) return CONFIGFILE_ERR_NOMEM;
			if (nLines < i+1) nLines = i+1;
			return 0;
		}
	}
	return CONFIGFILE_ERR_FULL;
}


int strpos(char c, char * str) {
	int i = 0;
	char *p = str;

	while(*p != c) {
		if (*p == 0) return -1;
		p++;
		i++;
	}
	return i;
}


char * configfile_t::findName(const char *name) {
	int i;
	char *foundEqual;

	for (i=0;i<nLines;i++) {
		if (lines[i]) {
			foundEqual = strchr(lines[i],'=');
			if (foundEqual) {
				if (strncmp (name,lines[i], foundEqual - lines[i]) == 0) {
					foundEqual++;
					return foundEqual;
				}
			}
		}
	}
	return NULL;
}


int configfile_t::findNameIdx(const char *name) {
	int i;
	char *foundEqual;

	for (i=0;i<nLines;i++) {
		if (lines[i]) {
			foundEqual = strchr(lines[i],'=');
			if (foundEqual) {
				if (strncmp (name,lines[i], foundEqual - lines[i]) == 0) {
					return i;
				}
			}
		}
	}
	return -1;
}


int configfile_t::load (const char * configFilename) {
//printf("configfile_t::readConfigFile (%s)\r\n",configFilename);
	freeConfigData();
	if (lastFilename) { free(lastFilename); lastFilename = NULL; }

	if (configFilename == NULL) return CONFIGFILE_ERR_FILE_NOT_FOUND;
	if (strlen(configFilename) == 0) return CONFIGFILE_ERR_FILE_NOT_FOUND;

	lastFilename = strdup(configFilename);

	FILE *fd = fopen(configFilename,"r");
    char buf[CONFIGFILE_LINE_LEN+1];
    buf[CONFIGFILE_LINE_LEN] = '0';

    if (fd == NULL) return CONFIGFILE_ERR_FILE_NOT_FOUND;

    while (fgets(buf,CONFIGFILE_LINE_LEN, fd) != NULL) {
        char *p;

        p = strchr(buf,'\r');
        if (!p) p = strchr(buf,'\n');
        if (p) *p = '\0';
        if (strchr(buf,'=')) add(buf);
    }
    fclose(fd);
    return 0;
}

int configfile_t::save (const char * configFilename) {
    //mode_t old_umask = umask(S_IXUSR|S_IRWXG|S_IRWXO);
    FILE *fd;
    int i;
    int fnIsBlank = 0;
    if (configFilename == NULL) fnIsBlank = 1;
    else
    if (strlen(configFilename) == 0) fnIsBlank = 1;

    if (fnIsBlank) {
    	if (lastFilename == NULL) return CONFIGFILE_ERR_FILE_NOT_FOUND;
    	if (strlen(lastFilename) == 0) return CONFIGFILE_ERR_FILE_NOT_FOUND;
    	fd = fopen(lastFilename,"w");
    	if (fd == NULL) return CONFIGFILE_ERR_FILE_CREATE;
    	//chmod(lastFilename,S_IRUSR|S_IWUSR);
    } else {
    	fd = fopen(configFilename,"w");
    	if (fd == NULL) return CONFIGFILE_ERR_FILE_CREATE;
    	//chmod(configFilename,S_IRUSR|S_IWUSR);
    }
    //umask(old_umask);

    for (i = 0; i < nLines; i++) {
    	if (lines[i]) fprintf(fd,"%s\n",lines[i]);
    }
    fclose(fd);
    return 0;
}



void configfile_t::freeConfigData() {
	int i;

	for (i=0;i<CONFIGFILE_MAX_LINES;i++) {
		if(lines[i] != NULL) { free(lines[i]); lines[i]=NULL; }
	}
	nLines = 0;
}
