// Copyright 2016-2017 Espressif Systems (Shanghai) PTE LTD
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// modified for evse
// Armin Diehl <ad@ardiehl.de>
// Dec 2017

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <sys/param.h>
#include "cons.h"
#include "linenoise_ad.h"
#include "argtable3_ad.h"
#include <sys/queue.h>

#include <fcntl.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include <unistd.h>


#define SS_FLAG_ESCAPE 0x8

typedef enum {
    /* parsing the space between arguments */
    SS_SPACE = 0x0,
    /* parsing an argument which isn't quoted */
    SS_ARG = 0x1,
    /* parsing a quoted argument */
    SS_QUOTED_ARG = 0x2,
    /* parsing an escape sequence within unquoted argument */
    SS_ARG_ESCAPED = SS_ARG | SS_FLAG_ESCAPE,
    /* parsing an escape sequence within a quoted argument */
    SS_QUOTED_ARG_ESCAPED = SS_QUOTED_ARG | SS_FLAG_ESCAPE,
} split_state_t;

size_t console_split_argv(char *line, char **argv, size_t argv_size)
{
    const int QUOTE = '"';
    const int ESCAPE = '\\';
    const int SPACE = ' ';
    split_state_t state = SS_SPACE;
    int argc = 0;
    char *next_arg_start = line;
    char *out_ptr = line;
    for (char *in_ptr = line; argc < argv_size - 1; ++in_ptr) {
        int char_in = (unsigned char) *in_ptr;
        if (char_in == 0) {
            break;
        }
        int char_out = -1;

        /* helper function, called when done with an argument */
#define end_arg  char_out = 0; argv[argc++] = next_arg_start; state = SS_SPACE

        switch (state) {
        case SS_SPACE:
            if (char_in == SPACE) {
                /* skip space */
            } else if (char_in == QUOTE) {
                next_arg_start = out_ptr;
                state = SS_QUOTED_ARG;
            } else if (char_in == ESCAPE) {
                next_arg_start = out_ptr;
                state = SS_ARG_ESCAPED;
            } else {
                next_arg_start = out_ptr;
                state = SS_ARG;
                char_out = char_in;
            }
            break;

        case SS_QUOTED_ARG:
            if (char_in == QUOTE) {
                end_arg;
            } else if (char_in == ESCAPE) {
                state = SS_QUOTED_ARG_ESCAPED;
            } else {
                char_out = char_in;
            }
            break;

        case SS_ARG_ESCAPED:
        case SS_QUOTED_ARG_ESCAPED:
            if (char_in == ESCAPE || char_in == QUOTE || char_in == SPACE) {
                char_out = char_in;
            } else {
                /* unrecognized escape character, skip */
            }
            state = (split_state_t) (state & (~SS_FLAG_ESCAPE));
            break;

        case SS_ARG:
            if (char_in == SPACE) {
                end_arg;
            } else if (char_in == ESCAPE) {
                state = SS_ARG_ESCAPED;
            } else {
                char_out = char_in;
            }
            break;
        }
        /* need to output anything? */
        if (char_out >= 0) {
            *out_ptr = char_out;
            ++out_ptr;
        }
    }
    /* make sure the final argument is terminated */
    *out_ptr = 0;
    /* finalize the last argument */
    if (state != SS_SPACE && argc < argv_size - 1) {
        argv[argc++] = next_arg_start;
    }
    /* add a NULL at the end of argv */
    argv[argc] = NULL;

    return argc;
}


#define ANSI_COLOR_DEFAULT      39      /** Default foreground color */

typedef struct cmd_item_ {
    /**
     * Command name (statically allocated by application)
     */
    const char* command;
    /**
     * Help text (statically allocated by application), may be NULL.
     */
    const char* help;
    /**
     * Hint text, usually lists possible arguments, dynamically allocated.
     * May be NULL.
     */
    char* hint;
    console_cmd_func_t func;        //!< pointer to the command handler
    void* argtable;                 //!< optional pointer to arg table
    SLIST_ENTRY(cmd_item_) next;    //!< next command in the list
} cmd_item_t;

/** linked list of command structures */
static SLIST_HEAD(cmd_list_, cmd_item_) s_cmd_list;

/** run-time configuration options */
static console_config_t s_config;

/** temporary buffer used for command line parsing */
static char* s_tmp_line_buf;


int console_init(const console_config_t* config)
{
    if (s_tmp_line_buf) {
        return -1;
    }
    memcpy(&s_config, config, sizeof(s_config));
    if (s_config.hint_color == 0) {
        s_config.hint_color = ANSI_COLOR_DEFAULT;
    }
    s_tmp_line_buf = (char *)calloc(config->max_cmdline_length, 1);
    if (s_tmp_line_buf == NULL) {
        return -1;
    }
    return 0;
}

int console_deinit()
{
    if (!s_tmp_line_buf) {
        return -1;
    }
    free(s_tmp_line_buf);
    cmd_item_t *it, *tmp;
    SLIST_FOREACH_SAFE(it, &s_cmd_list, next, tmp) {
        free(it->hint);
        free(it);
    }
    return 0;
}

int console_cmd_register(const console_cmd_t *cmd)
{
    cmd_item_t *item = (cmd_item_t *) calloc(1, sizeof(*item));
    if (item == NULL) {
        return -1;
    }
    if (cmd->command == NULL) {
        return -1;
    }
    if (strchr(cmd->command, ' ') != NULL) {
        return -1;
    }
    item->command = cmd->command;
    item->help = cmd->help;
    if (cmd->hint) {
        /* Prepend a space before the hint. It separates command name and
         * the hint. arg_print_syntax below adds this space as well.
         */
        asprintf(&item->hint, " %s", cmd->hint);
    } else if (cmd->argtable) {
        /* Generate hint based on cmd->argtable */
        char* buf = NULL;
        size_t buf_size = 0;
        FILE* f = open_memstream(&buf, &buf_size);
        if (f != NULL) {
            arg_print_syntax(f, (void **)cmd->argtable, NULL);
            fclose(f);
        }
        item->hint = buf;
    }
    item->argtable = cmd->argtable;
    item->func = cmd->func;
    cmd_item_t* last = SLIST_FIRST(&s_cmd_list);
    if (last == NULL) {
        SLIST_INSERT_HEAD(&s_cmd_list, item, next);
    } else {
        cmd_item_t* it;
        while ((it = SLIST_NEXT(last, next)) != NULL) {
            last = it;
        }
        SLIST_INSERT_AFTER(last, item, next);
    }
    return 0;
}

void console_get_completion(const char *buf, linenoiseCompletions_t *lc)
{
    size_t len = strlen(buf);
    if (len == 0) {
        return;
    }
    cmd_item_t* it;
    SLIST_FOREACH(it, &s_cmd_list, next) {
        /* Check if command starts with buf */
        if (strncmp(buf, it->command, len) == 0) {
            linenoiseAddCompletion(lc, it->command);
        }
    }
}

const char* console_get_hint(const char *buf, int *color, int *bold)
{
    int len = strlen(buf);
    cmd_item_t* it;
    SLIST_FOREACH(it, &s_cmd_list, next) {
        if (strlen(it->command) == len &&
                strncmp(buf, it->command, len) == 0) {
            *color = s_config.hint_color;
            *bold = s_config.hint_bold;
            return it->hint;
        }
    }
    return NULL;
}

static const cmd_item_t* find_command_by_name(FILE *fp_stdout, const char* name)
{
    const cmd_item_t* cmd = NULL;
    int compareLen;
    int numFound;

    cmd_item_t* it;
    SLIST_FOREACH(it, &s_cmd_list, next) {
        if (strcasecmp(name, it->command) == 0) {
            cmd = it;
            break;
        }
    }
    if (cmd) return cmd;

    compareLen = strlen(name);
    while (compareLen >= strlen(name)) {
    	numFound = 0;
    	SLIST_FOREACH(it, &s_cmd_list, next) {
    		if (strncasecmp(name, it->command, compareLen) == 0) {
    			cmd = it;
    			numFound++;
    		}
    	}
        if (numFound == 1) return cmd;
        if (numFound > 1) {
        	fprintf(fp_stdout,CONS_COLOR_STR_E("Ambiguous command") " '%s'. Candidates: " CONS_BOLD(CONS_COLOR_WHITE), name);
        	SLIST_FOREACH(it, &s_cmd_list, next) {
        		if (strncasecmp(name, it->command, compareLen) == 0) {
        			fprintf(fp_stdout,"%s ",it->command);
        		}
        	}
        	fprintf(fp_stdout,CONS_RESET_COLOR "\r\n");
        	return NULL;
        }
        compareLen--;
    }
    fprintf(fp_stdout,CONS_COLOR_STR_E("Unknown command") " '%s'.\r\n" CONS_BOLD(CONS_COLOR_WHITE), name);
    return NULL;
}

int console_run(FILE *fp_stdout, FILE *fp_stdin, const char* cmdline, int* cmd_ret)
{
    if (s_tmp_line_buf == NULL) {
        return -1;
    }
    char** argv = (char**) calloc(s_config.max_cmdline_args, sizeof(char*));
    if (argv == NULL) {
        return -2;
    }
    // strlcpy(s_tmp_line_buf, cmdline, s_config.max_cmdline_length);
    snprintf(s_tmp_line_buf, s_config.max_cmdline_length, "%s",	cmdline);


    size_t argc = console_split_argv(s_tmp_line_buf, argv, s_config.max_cmdline_args);
    if (argc == 0) {
        return ERR_ARGS;
    }
    const cmd_item_t* cmd = find_command_by_name(fp_stdout, argv[0]);  /* prints message if cmd was not found */
    if (cmd == NULL) {
    	//if (fp_stdout != NULL)	fprintf(fp_stdout, CONS_COLOR_STR_E_N("invalid command '%s'"),argv[0]);
        return ERR_INVALID_COMMAND;
    }

    if (cmd->argtable) {
    	int nerrors = arg_parse(argc, argv, (void**)cmd->argtable);
    	if (nerrors != 0) {
    		if (fp_stdout != NULL) {
    			struct arg_end * end = arg_find_end((void **)cmd->argtable);
    			arg_print_errors(fp_stdout, end, argv[0]);
    		}
        	return 1;
    	}
    }

    *cmd_ret = (*cmd->func)(fp_stdout, fp_stdin, argc, argv);
    free(argv);
    return 0;
}


static struct {
    struct arg_str *cmd = arg_strn(NULL, NULL, "<command [command] ..>", 0,25,"name of command for help");
    struct arg_end *end = arg_end(2);
} helpCommandArgs;


static void print_help(FILE *fp, const cmd_item_t* it, int printDetails) {
	/* First line: command name and hint
     * Pad all the hints to the same column
     */
	const char* hint = (it->hint) ? it->hint : "";
	fprintf(fp,CONS_BOLD(CONS_COLOR_WHITE) "%-s" CONS_RESET_COLOR " %s\r\n", it->command, hint);
	/* Second line: print help.
	 * Argtable has a nice helper function for this which does line
	 * wrapping.
	*/
	fprintf(fp, CONS_COLOR_I "  "); // arg_print_formatted does not indent the first line
	arg_print_formatted(fp, 2, 78, it->help);
	fprintf(fp,CONS_RESET_COLOR);
	if (printDetails) {
		/* Finally, print the list of arguments, but only if help for one commend was requested */
		if (it->argtable) {
			arg_print_glossary(fp, (void**) it->argtable, NULL);
		}
		fprintf(fp, "\r\n");
	}
}


static int help_command(FILE *fp, FILE *, int, char**)
{
    const cmd_item_t* it;
    int i;

    if (helpCommandArgs.cmd->count) {	/* help for one or more commands */
    	for (i=0; i < helpCommandArgs.cmd->count; i++) {
    		it = find_command_by_name(fp, helpCommandArgs.cmd->sval[i]);
    		if (it) print_help(fp, it, true);
    	}
    } else {
    	/* Print summary of each command */
    	SLIST_FOREACH(it, &s_cmd_list, next) {
    		if (it->help == NULL) {
    			continue;
    		}
    		print_help(fp, it, false);
    	}
    	fprintf(fp,CONS_COLOR_STR_T_N("\nCommands can be shortened as long as they are not ambiguous."));
    }
    return 0;
}


int console_register_help_command()
{

    console_cmd_t command = {
            .command = "?",
            .help = NULL,
			.hint = NULL,
            .func = &help_command,
			.argtable = &helpCommandArgs
    };
    console_cmd_register(&command);
    CMD_REG(help, "Print the list of registered commands", NULL, &help_command, &helpCommandArgs);

    return 0;
}


int fd_set_blocking(int fd, int blocking) {
    /* Save the current flags */
    int flags = fcntl(fd, F_GETFL, 0);
    if (flags == -1)
        return 0;

    if (blocking)
        flags &= ~O_NONBLOCK;
    else
        flags |= O_NONBLOCK;
    return fcntl(fd, F_SETFL, flags) != -1;
}

/* read a char from console, returns -1 if no char is available */
int console_getChar (FILE *fp_stdin, int timeout) {
	int res;
	int timeRemaining;
	char c;

	fd_set_blocking (fileno(fp_stdin),0);
	timeRemaining = timeout;

	do {
		//res = fread(&c,1,1,fp_stdin);
		res = read(fileno(fp_stdin),&c,1);
		//if (res >= 0) printf("read returned %d %d \r\n",res,(int)c);
		if (res == 1)
			if (c == '\0') res = -1;  // this only happen on esp32 over telnet ??????

		timeRemaining -= 100;
		if (res <= 0) vTaskDelay (pdMS_TO_TICKS(100));  // 100ms
	} while ((res < 1) & (timeRemaining > 0));

	if (res == 1) res = c;
	else res = -1;

	fd_set_blocking (fileno(fp_stdin),1);

	return res;
}

