/* linenoise.h -- VERSION 1.0
 *
 * Guerrilla line editing library against the idea that a line editing lib
 * needs to be 20,000 lines of C code.
 *
 * See linenoise.c for more information.
 *
 * ------------------------------------------------------------------------
 *
 * Copyright (c) 2010-2014, Salvatore Sanfilippo <antirez at gmail dot com>
 * Copyright (c) 2010-2013, Pieter Noordhuis <pcnoordhuis at gmail dot com>
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *  *  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  *  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __LINENOISE_H
#define __LINENOISE_H

#include <stddef.h>

typedef struct linenoiseCompletions_t {
  size_t len;
  char **cvec;
} Completions_t;

typedef void(linenoiseCompletionCallback_t)(const char *, linenoiseCompletions_t *);
typedef char*(linenoiseHintsCallback_t)(const char *, int *color, int *bold);
typedef void(linenoiseFreeHintsCallback_t)(void *);

/* The linenoiseState structure represents the state during line editing.
 * We pass this state to functions implementing specific editing
 * functionalities. */
struct linenoiseState_t {
	int ifd;            /* Terminal stdin file descriptor. */
    int ofd;            /* Terminal stdout file descriptor. */
    char *buf;          /* Edited line buffer. */
    size_t buflen;      /* Edited line buffer size. */
    const char *prompt; /* Prompt to display. */
    size_t plen;        /* Prompt length. */
    size_t pos;         /* Current cursor position. */
    size_t oldpos;      /* Previous refresh cursor position. */
    size_t len;         /* Current edited line length. */
    size_t cols;        /* Number of columns in terminal. */
    size_t maxrows;     /* Maximum num of rows used so far (multiline mode) */
    int history_index;  /* The history index we are currently editing. */
};

struct abuf_t {
    char *b;
    int len;
};




/* This function is used by the callback function registered by the user
 * in order to add completion options given the input string when the
 * user typed <tab>. See the example.c source code for a very easy to
 * understand example. */
void linenoiseAddCompletion(linenoiseCompletions_t *lc, const char *str);

/* Free a list of completion option populated by linenoiseAddCompletion(). */
void linenoiseFreeCompletions(linenoiseCompletions_t *lc);


class linenoise_t
{
  public:

	linenoise_t();

	int readChar (char *c);

	void setFileNo(int in_fileno, int out_fileno);

	void setCompletionCallback(linenoiseCompletionCallback_t *);
	void setHintsCallback(linenoiseHintsCallback_t *);
	void setFreeHintsCallback(linenoiseFreeHintsCallback_t *);


	int probe(void);
	char *run(const char *prompt, int * result = NULL);
/* This is just a wrapper the user may want to call in order to make sure
 * the linenoise returned buffer is freed with the same allocator it was
 * created with. Useful when the main program is using an alternative
 * allocator. */
	void freeBuffer(void *ptr);
/* This is the API call to add a new entry in the linenoise history.
 * It uses a fixed array of char pointers that are shifted (memmoved)
 * when the history max length is reached in order to remove the older
 * entry and make room for the new one, so it is not exactly suitable for huge
 * histories, but will work well for a few hundred of entries.
 *
 * Using a circular buffer is smarter, but a bit more complex to handle. */
	int historyAdd(const char *line);

	/* Set the maximum length for the history. This function can be called even
 * if there is already some history, the function will make sure to retain
 * just the latest 'len' elements if the new history length value is smaller
 * than the amount of items already inside the history. */
	int historySetMaxLen(int len);

	/* Save the history in the specified file. On success 0 is returned
 * otherwise -1 is returned. */
	int historySave(const char *filename);

/* Load the history from the specified file. If the file does not exist
 * zero is returned and no operation is performed.
 *
 * If the file exists and the operation succeeded 0 is returned, otherwise
 * on error -1 is returned. */
	int historyLoad(const char *filename);
	void freeHistory();
	/* Set if to use or not the multi line mode. */
	void setMultiLine(int ml);
/* Set if terminal does not recognize escape sequences */
	void printKeyCodes(void);

	void beep(void);


  private:
	int my_STDIN_FILENO, my_STDOUT_FILENO;
	linenoiseCompletionCallback_t *completionCallback;
	linenoiseHintsCallback_t      *hintsCallback;
	linenoiseFreeHintsCallback_t  *freeHintsCallback;

	int mlmode;         /* Multi line mode. Default is single line. */
	int dumbmode;       /* Dumb mode where line editing is disabled. Off by default */
	int history_max_len;
	int history_len;
	char **history;

		void refreshLine(struct linenoiseState_t *l);
	int getCursorPosition(int ifd, int ofd);
	int getColumns(int ifd, int ofd); /* Try to get the number of columns in the current terminal, or assume 80 if it fails. */
	void clearScreen(void);  /* Clear the screen. Used to handle ctrl+l */
	int completeLine(struct linenoiseState_t *ls);
	void refreshShowHints(struct abuf_t *ab, struct linenoiseState_t *l, int plen);
	void refreshSingleLine(struct linenoiseState_t *l);
	void refreshMultiLine(struct linenoiseState_t *l);
	int editInsert(struct linenoiseState_t *l, char c);
	void editMoveLeft(struct linenoiseState_t *l);
	void editMoveRight(struct linenoiseState_t *l);
	void editMoveHome(struct linenoiseState_t *l);
	void editMoveEnd(struct linenoiseState_t *l);
	void editHistoryNext(struct linenoiseState_t *l, int dir);
	void editDelete(struct linenoiseState_t *l);
	void editBackspace(struct linenoiseState_t *l);
	void editDeletePrevWord(struct linenoiseState_t *l);
	int edit(int stdin_fd, int stdout_fd, char *buf, size_t buflen, const char *prompt);
	int raw(char *buf, size_t buflen, const char *prompt);
/* This function is called when linenoise() is called with the standard
 * input file descriptor not attached to a TTY. So for example when the
 * program using linenoise is called in pipe or with a file redirected
 * to its standard input. In this case, we want to be able to return the
 * line regardless of its length (by default we are limited to 4k). */
	char *noTTY(void);

	int dumb(char* buf, size_t buflen, const char* prompt);

};

#endif /* __LINENOISE_H */
