#include "mcp23017.h"

#ifdef ESP_PLATFORM
#include <driver/gpio.h>
#include <driver/i2c.h>
#endif

#include "esp_log.h"

static const char* TAG = "MCP23017";

// disable buffers
#ifdef ESP_PLATFORM
static const size_t I2C_MASTER_TX_BUF_DISABLE = 0;
static const size_t I2C_MASTER_RX_BUF_DISABLE = 0;
static const int INTR_FLAGS = 0;
#endif // ESP_PLATFORM

#ifndef ESP_PLATFORM
uint8_t registers[MCP23017_OLATB+1] = {
		0xff,0xff,	// dir
		0,0,0,0,0,0,0,0,0,0
};
#endif  // ESP_PLATFORM

/**
 * Converts generic register and group (A/B) to register address
 * @param reg the generic register index
 * @param group the group (A/B) to compute offset
 * @return The register address specified by the parameters
*/
uint8_t mcp23017_register(mcp23017_reg_t reg, mcp23017_gpio_t group) {
   return (group == GPIOA)?(reg << 1):(reg << 1) | 1;
}

/**
 * Initializes the MCP23017
 * @param mcp the MCP23017 interface structure
 * @return an error code or MCP23017_ERR_OK if no error encountered
*/
mcp23017_err_t mcp23017_init(mcp23017_t *mcp, int initDriver) {
#ifdef ESP_PLATFORM
	esp_err_t ret;

	if (initDriver) {
		// setup i2c controller
		i2c_config_t conf = {
			.mode = I2C_MODE_MASTER,
			.sda_io_num = mcp->sda_pin,
			.scl_io_num = mcp->scl_pin,
			.sda_pullup_en = mcp->sda_pullup_en,
			.scl_pullup_en = mcp->scl_pullup_en,
			.master.clk_speed = 100000
		};
		if (mcp->i2cClockSpeed > 1) conf.master.clk_speed = mcp->i2cClockSpeed;
		ret = i2c_param_config(mcp->port, &conf);
		ESP_LOGV(TAG,"i2c_param_config sda:%d scl:%d dpeed:%d",mcp->sda_pin,mcp->scl_pin,conf.master.clk_speed);

		if( ret != ESP_OK ) {
			ESP_LOGE(TAG,"PARAM CONFIG FAILED");
			return MCP23017_ERR_CONFIG;
		}
		ESP_LOGV(TAG,"PARAM CONFIG done");
	
		// install the driver
		ret = i2c_driver_install(mcp->port, I2C_MODE_MASTER, I2C_MASTER_TX_BUF_DISABLE, I2C_MASTER_RX_BUF_DISABLE, INTR_FLAGS);

		if(ret != ESP_OK) {
			ESP_LOGE(TAG,"I2C driver install failed, err:%d",ret);
			return MCP23017_ERR_INSTALL;
		}
		ESP_LOGV(TAG,"I2C DRIVER INSTALLED");
	}
#endif
	// make all I/O's output
	mcp23017_write_register(mcp, MCP23017_IODIR, GPIOA, 0x00);
	mcp23017_write_register(mcp, MCP23017_IODIR, GPIOB, 0x00);

	return MCP23017_ERR_OK;
}

/**
 * Writes a value to an MCP23017 register
 * @param mcp the MCP23017 interface structure
 * @param reg A generic register index
 * @param group the group (A/B) to compute register address offset
 * @param v the value to write to the register
 * @return an error code or MCP23017_ERR_OK if no error encountered
*/
mcp23017_err_t mcp23017_write_register(mcp23017_t *mcp, mcp23017_reg_t reg, mcp23017_gpio_t group, uint8_t v) {
	uint8_t r = mcp23017_register(reg, group);
#ifdef ESP_PLATFORM

   i2c_cmd_handle_t cmd = i2c_cmd_link_create();
   i2c_master_start(cmd);
   i2c_master_write_byte(cmd, mcp->i2c_addr << 1 | WRITE_BIT, ACK_CHECK_EN);
   i2c_master_write_byte(cmd, r, ACK_CHECK_EN);
   i2c_master_write_byte(cmd, v, ACK_CHECK_EN);
   i2c_master_stop(cmd);
   esp_err_t ret = i2c_master_cmd_begin(mcp->port, cmd, MCP_I2C_TIMEOUT_MS / portTICK_RATE_MS);
   i2c_cmd_link_delete(cmd);
   if (ret != 0) {
      ESP_LOGE(TAG,"ERROR: unable to write to register, err: %d",ret);
      return MCP23017_ERR_FAIL;
   }
#else
   registers[r] = v;
   ESP_LOGD(TAG,"write_register (%d, %d, %d), addr: %d",reg,group,v,mcp->i2c_addr);
#endif
   return MCP23017_ERR_OK;
}

/**
 * Reads a value to an MCP23017 register
 * @param mcp the MCP23017 interface structure
 * @param reg A generic register index
 * @param group the group (A/B) to compute register address offset
 * @param data a pointer to an 8 bit value to be read from the device
 * @return an error code or MCP23017_ERR_OK if no error encountered
*/
mcp23017_err_t mcp23017_read_register(mcp23017_t *mcp, mcp23017_reg_t reg, mcp23017_gpio_t group, uint8_t *data) {
	uint8_t r = mcp23017_register(reg, group);
	esp_err_t ret;
#ifdef ESP_PLATFORM
   // from the generic register and group, derive register address
   
	i2c_cmd_handle_t cmd = i2c_cmd_link_create();
	i2c_master_start(cmd);
	i2c_master_write_byte(cmd, (mcp->i2c_addr << 1) | I2C_MASTER_WRITE, ACK_CHECK_EN);
	i2c_master_write_byte(cmd, r, 1);
	i2c_master_stop(cmd);

	ret =i2c_master_cmd_begin(mcp->port, cmd, MCP_I2C_TIMEOUT_MS / portTICK_RATE_MS);
	i2c_cmd_link_delete(cmd);
	if( ret == ESP_FAIL ) {
	   ESP_LOGE(TAG,"ERROR: unable to write address %02x to read reg %02x",mcp->i2c_addr,r);
	   return MCP23017_ERR_FAIL;
	}

	// ad: it does not work if we are doing the read in one transaction
	cmd = i2c_cmd_link_create();
	i2c_master_start(cmd);
	i2c_master_write_byte(cmd, (mcp->i2c_addr << 1) | I2C_MASTER_READ, ACK_CHECK_EN);
	i2c_master_read_byte(cmd, data, 1);
	i2c_master_stop(cmd);  // AD: this was missing
	ret =i2c_master_cmd_begin(mcp->port, cmd, MCP_I2C_TIMEOUT_MS / portTICK_RATE_MS);
	i2c_cmd_link_delete(cmd);
	if (ret != 0) {
	   ESP_LOGE(TAG,"ERROR: unable to read reg %02x from address %02x, err: %d",r,mcp->i2c_addr,ret);
	   return MCP23017_ERR_FAIL;
	}
#else
	*data = registers[r];
	ESP_LOGD(TAG,"read_register (%d, %d, %d), addr: %d",reg,group,*data,mcp->i2c_addr);
#endif
	return MCP23017_ERR_OK;
}

/**
 * Clears a bit from a current register value
 * @param mcp address of the MCP23017 data structure
 * @param bit The number of the bit to set
 * @param reg A generic register index
 * @param group the group (A/B) to compute register address offset
 * @return an error code or MCP23017_ERR_OK if no error encountered
*/
mcp23017_err_t mcp23017_set_bit(mcp23017_t *mcp, uint8_t bit, mcp23017_reg_t reg, mcp23017_gpio_t group) {   
   uint8_t current_value;
   if( mcp23017_read_register(mcp, reg, group, &current_value) != MCP23017_ERR_OK ) {
      uint8_t r = mcp23017_register(reg, group);
      ESP_LOGE(TAG, "ERROR: unable to read current value of register %02x",r);
      return MCP23017_ERR_FAIL;
   }
   current_value |= 1 << bit;
   if( mcp23017_write_register(mcp, reg, group, current_value) != MCP23017_ERR_OK ) {
      uint8_t r = mcp23017_register(reg, group);
      ESP_LOGE(TAG, "ERROR: unable to write new value %02X to register %02x",current_value, r);
      return MCP23017_ERR_FAIL;
   }
   return MCP23017_ERR_OK;
}

/**
 * Clears a bit from a current register value
 * @param mcp address of the MCP23017 data structure
 * @param bit The number of the bit to set
 * @param reg A generic register index
 * @param group the group (A/B) to compute register address offset
 * @return an error code or MCP23017_ERR_OK if no error encountered
*/
mcp23017_err_t mcp23017_clear_bit(mcp23017_t *mcp, uint8_t bit, mcp23017_reg_t reg, mcp23017_gpio_t group) {   
   uint8_t current_value;
   if( mcp23017_read_register(mcp, reg, group, &current_value) != MCP23017_ERR_OK ) {
      uint8_t r = mcp23017_register(reg, group);
      ESP_LOGE(TAG, "ERROR: unable to read current value of register %02x",r);
      return MCP23017_ERR_FAIL;
   }
   current_value &= ~(1 << bit);
   if( mcp23017_write_register(mcp, reg, group, current_value) != MCP23017_ERR_OK ) {
      uint8_t r = mcp23017_register(reg, group);
      ESP_LOGE(TAG, "ERROR: unable to write new value %02X to register %02x",current_value, r);
      return MCP23017_ERR_FAIL;
   }
   return MCP23017_ERR_OK;
}

