/* EVSE_AD - display
   Handles the display and keyboard

   Armin Diehl <ad@ardiehl.de> December 2017

   Jun 7, 2018 AD:	manual mode
   	   	   	   	   	allow adjust phases
   	   	   	   	   	show car page only if car is connected
   Jun 11, 2018 AD: show energy meter data
   Jun 12, 2018 AD: show energy meter KW/h as well

   Unless required by applicable law or agreed to in writing, this
   software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
   CONDITIONS OF ANY KIND, either express or implied.
*/

#include "evse_global.h"
#include <u8g2.h>
#include <stdarg.h>
#include "energyMeter.h"
#include "j1772pilot.h"

//-----------------------------------------
#ifdef ESP_PLATFORM
#include "u8g2_esp32_hal.h"
#include "tcpip_adapter.h"

// SDA - GPIO21
#define PIN_SDA (gpio_num_t)4

// SCL - GPIO22
#define PIN_SCL (gpio_num_t)16

#endif  // ESP_PLATFORM
//-----------------------------------------

#define SCR_WIDTH   128
#define SCR_HEIGHT   64
#define STAT_HEIGHT  11
#define SCR_APP_HEIGHT (SCR_HEIGHT-STAT_HEIGHT)
#define SCR_APP_OFFSET STAT_HEIGHT

/* time for auto change to next page */
#define AUTO_PAGE_CHANGE_TIME_MS 2000
/* multiplier to wait on a page that was selected via keys */
#define AUTO_PAGE_WAIT_AFTER_KEY 10
/* timeout to exit a page in edit mode */
#define TIMOUT_TIME_MS           120*1000

static const char *TAG = "UI";

u8g2_t u8g2;      /* a structure which will contain all the data for one display */
static int networkConnected = 0;
static int availablePower = 0;
static int currentActivePhases = 0;  // bitmask of active phases
static int cableAmps = 0;

#define D &u8g2   /* this is more easy to type */
int currFontHeight;


// Flags
#define UI_PG_FLAG_CANHANDLE_FOCUS 0x0001
#define UI_PG_FLAG_CANHANDLE_EDIT  0x0002
#define UI_PG_FLAG_NOAUTOSHOW      0x0004

// return values from page handler function
#define UI_PH_OK					0
// page has nothing to show, skip to next one
#define UI_PH_DONTSHOW				1
// page is in edit mode, do not auto change pages
#define UI_PH_HASFOCUS				2

struct ui_page_t;

struct ui_page_t {
	const char *text;
    int flags;
    int ( *fp )(struct ui_page_t *p, ev_event_t event, ev_data_t *data, int hasFocus);      // function to call

};

static int focusNumInput;
static int mqttConnected;

int handle_CARSTAT_page			(struct ui_page_t *p, ev_event_t event, ev_data_t *data, int);
int handle_IPINFO_page			(struct ui_page_t *p, ev_event_t event, ev_data_t *data, int);
int handle_POWER_page			(struct ui_page_t *p, ev_event_t event, ev_data_t *data, int);
int handle_MANUAL_CHARGE_page	(struct ui_page_t *p, ev_event_t event, ev_data_t *data, int hasFocus);
int handle_PHASES_page			(struct ui_page_t *p, ev_event_t event, ev_data_t *data, int hasFocus);
int handle_EM_page			(struct ui_page_t *p, ev_event_t event, ev_data_t *data, int hasFocus);

#ifdef LANG_EN
#define TX_MANEXIT      "Press OK to exit\nmanual charge mode"
#define TX_MANMODE_INFO "Manual charging\nStart: %s\nEnd: %s\n%d Amps"
#define TX_STARTTIME    "Starttime"
#define TX_ENDTIME      "Endtime"
#define TX_MANCONF      "OK to start\nmanual charging"
#define TX_MAXAMPS      "max Amps"
#define TX_PHASES       "Phases"
#define TX_MANCH        "Man. charge"
#define TX_CAR          "Car"
#define TX_AUTO         "automatic"
#define TX_CARINFO      "%5d Watt\n%5d Phases"
#define TX_CONNECTED 	"connected"
#define TX_DISCONNECTED "not connected"
#define TX_EM            "Meter"
#else
#define TX_MANEXIT      "OK zum beenden des\nmanuellen ladens"
#define TX_MANMODE_INFO "Manuelles Laden\nStart: %s\nEnde: %s\n%d Amps"
#define TX_STARTTIME    "Startzeit"
#define TX_ENDTIME      "Endzeit"
#define TX_MANCONF      "OK zum\nsofortigen laden"
#define TX_MAXAMPS      "max Amps"
#define TX_PHASES       "Phasen"
#define TX_MANCH        "Man. Laden"
#define TX_CAR          "Auto"
#define TX_AUTO         "automatisch"
#define TX_CARINFO      "%5d Watt\n%5d Phasen"
#define TX_CONNECTED 	"verbunden"
#define TX_DISCONNECTED "nicht verbunden"
#define TX_EM            "Meter"
#endif

#define UI_PG_CARSTAT 			0		/* Car Status */
#define UI_PG_IPINFO			1		/* Show IP Status and Addresses */
#define UI_PG_POWER				2		/* Show IP Status and Addresses */
#define UI_PG_MANUAL_CHARGE		3		/* enable/disable manual charging */
#define UI_PG_PHASES			4		/* set supported phases */
#define UI_PG_EM                5		/* engery meter data */
#define UI_PG_EOL				6		/* end of list marker */

#define UI_PAGE_FIRST UI_PG_CARSTAT

ui_page_t pages_data [] = {
		{ TX_CAR          , 0                          , handle_CARSTAT_page },
		{ "IP"            , UI_PG_FLAG_NOAUTOSHOW      , handle_IPINFO_page },
		{ "Power"         , 0                          , handle_POWER_page },
		{ TX_MANCH        , UI_PG_FLAG_CANHANDLE_FOCUS , handle_MANUAL_CHARGE_page },
		{ TX_PHASES       , UI_PG_FLAG_CANHANDLE_FOCUS , handle_PHASES_page },
		{ TX_EM           , UI_PG_FLAG_NOAUTOSHOW      , handle_EM_page }
};

/* horizontal space in pixel between lines, may later be depending on font size */
int getHorSpaceBetweenLines() {
	return 1;
}

/* get the pixel Y start for a given text line (x) in the user area */
int getLineY(int lineNo) {
	if (lineNo == -1) {
		// center vertical
		return (((SCR_HEIGHT - STAT_HEIGHT) / 2) + ((currFontHeight+getHorSpaceBetweenLines())/2)) + STAT_HEIGHT;
	}
	return (STAT_HEIGHT+((lineNo+1) * (currFontHeight+getHorSpaceBetweenLines() )));
}


#define MAX_FONT_SIZE 12
#define MIN_FONT_SIZE 6
void setFontSize (int height) {
	if (height < MIN_FONT_SIZE) height = MIN_FONT_SIZE;
	if (height > MAX_FONT_SIZE) height = MAX_FONT_SIZE;

	currFontHeight = height;
	switch (height) {
	/*	case 4:		u8g2_SetFont(D, u8g2_font_u8glib_4_tf);
					break;
		case 5:		u8g2_SetFont(D, 8g2_font_4x6_tf);
					break; */
		case 6:		u8g2_SetFont(D, u8g2_font_5x8_tf);
					break;
		case 7:		u8g2_SetFont(D, u8g2_font_6x10_tf);
					break;
		case 8:		u8g2_SetFont(D, u8g2_font_t0_11_tf);
					break;
		case 9:		u8g2_SetFont(D, u8g2_font_7x13_tf); //u8g2_font_t0_11b_tf);
					break;
		case 10:	u8g2_SetFont(D, u8g2_font_7x14_tf);
					break;
		case 11:	u8g2_SetFont(D, u8g2_font_fur11_tf);
					break;
		case 12:	u8g2_SetFont(D, u8g2_font_crox3hb_tf);
					break;
	}
}


/* Show a message in the user area
 * Can contain multiple lines separated by \n. The maximum font size
 * fitting height and width will be used
 */

#define MSG_MAX_LINES 10
#define MSG_MAX_LEN 255
char showMessageBuf[MSG_MAX_LEN+1];

void showMessage(	int horizontalCenter,
					const char* format, ...) __attribute__ ((format (printf, 2, 3)));

void showMessage(	int horizontalCenter,
					const char* format, ...) {
	int lines;
	char *l[MSG_MAX_LINES];
	char *p;
	int i,minFontSize,fontSize,maxNumLines,startLine,width,x;

	for (lines=0;lines<MSG_MAX_LINES;lines++) {
		l[lines]=NULL;
	}

	va_list list; va_start(list, format);
    vsnprintf(&showMessageBuf[0], MSG_MAX_LEN, format, list);
    va_end(list);

	showMessageBuf[MSG_MAX_LEN] = 0;  /* to make sure ... */

	lines = 1;  /* determine the number of lines to print, replace CR or LF
	 	 	 	   with \0 and store the start of each line in l */
	p = &showMessageBuf[0];
	l[0]=p;
	while (*p) {
		if ((*p == 13) || (*p == 10)) {
			*p = 0;			// replace separator by end of string
			p++;
			l[lines] = p;
			lines++;
		} else {
			p++;
		}
	}
	/*ESP_LOGI(TAG,"'%s' lines: %d",s,lines);
	for (i=0;i<lines;i++) {
		ESP_LOGI(TAG,"%d: '%s'",i,l[i]);
	}*/

	/* determine maximum possible font height to fit number of lines */
	minFontSize = (SCR_APP_HEIGHT-(lines*getHorSpaceBetweenLines() )) / lines;
	fontSize = minFontSize;
	if (minFontSize > MAX_FONT_SIZE) fontSize = MAX_FONT_SIZE;
	if (minFontSize < MIN_FONT_SIZE) fontSize = MIN_FONT_SIZE;
	setFontSize(fontSize);
	//ESP_LOGD(TAG,"minFontSize: %d, fontSize: %d",minFontSize,fontSize);

	/* determine if a smaller font size is needed to fit each line */
	for (i=0;i<lines;i++) {
		while ((fontSize > MIN_FONT_SIZE) && (u8g2_GetStrWidth(D,l[i]) > SCR_WIDTH)) {
			fontSize--;
			setFontSize(fontSize);
		}
	}
	maxNumLines = SCR_APP_HEIGHT / (fontSize + getHorSpaceBetweenLines() );
	startLine = 0;
	if (lines < maxNumLines) {
		startLine = (maxNumLines - lines) / 2;
	}
	if (lines > maxNumLines) lines = maxNumLines;
	//ESP_LOGD(TAG,"fontSize after checking width: %d, maxNumLines: %d, startLine: %d, AppHeight: %d",fontSize,maxNumLines,startLine,SCR_APP_HEIGHT);

	for (i=0;i<lines;i++) {
		x = 0;
		if (horizontalCenter) {
			width = u8g2_GetStrWidth(D,l[i]);
			if (width < SCR_WIDTH-1) {
				x = (SCR_WIDTH - width) / 2;
			}
		}
		u8g2_DrawStr(D, x, getLineY(i+startLine), l[i]);
	}
}


void showStatusLine (struct ui_page_t *p) {
	u8g2_ClearBuffer(D);

	char str[7];
	int width;

	if (cableAmps == 0) {
		strcpy(str,"-- ---");
	} else sprintf(str,"%02d ---",cableAmps);
	if (currentActivePhases & 1) str[3] = '1';
	if (currentActivePhases & 2) str[4] = '1';
	if (currentActivePhases & 4) str[5] = '1';

	u8g2_DrawHLine(D, 0, STAT_HEIGHT-1, SCR_WIDTH-1);
	setFontSize(9);
	if (p == NULL) {
#ifdef LANG_EN
		u8g2_DrawStr(D, 0, STAT_HEIGHT-2, "Initialisiere....");
#else
		u8g2_DrawStr(D, 0, STAT_HEIGHT-2, "Initializing....");
#endif
		return;
	}
	u8g2_DrawStr(D, 0, STAT_HEIGHT-2, p->text);
	width = u8g2_GetStrWidth(D,str);
	u8g2_DrawStr(D, SCR_WIDTH-width, STAT_HEIGHT-2, str);
}

// **********************************************************************************************************


int handle_EM_page (struct ui_page_t *p, ev_event_t UNUSED(event), ev_data_t UNUSED(*data), int) {
	showStatusLine(p);

	em_data_t *d = energyMeter_getCurrentData();

	showMessage (0,
			"%03.1fV %02.1fA %4dW\n" \
			"%03.1fV %02.1fA %4dW\n" \
			"%03.1fV %02.1fA %4dW\n" \
			"    %d KW/h",
			d->voltageL1, d->currL1, d->pwrL1,
			d->voltageL2, d->currL2, d->pwrL2,
			d->voltageL3, d->currL3, d->pwrL3,
			d->kwh);

	u8g2_SendBuffer(D);
	return UI_PH_OK;
}

int handle_POWER_page (struct ui_page_t *p, ev_event_t UNUSED(event), ev_data_t UNUSED(*data), int) {
	showStatusLine(p);
	if (availablePower >= 1000) {
		char s[20];
		snprintf(&s[0],20,"%d",availablePower % 1000);
		showMessage(1,"Solar: %d.%c KW", availablePower / 1000, s[0]);
	} else
		showMessage(1,"Solar: %d W", availablePower);

	u8g2_SendBuffer(D);
	return UI_PH_OK;
}


int handle_CARSTAT_page (struct ui_page_t *p, ev_event_t UNUSED(event), ev_data_t UNUSED(*data), int) {
	if (!j1772_carConnected()) return UI_PH_DONTSHOW;  // don't show car stat page if car is not connected
	int numPhases;
	int powerConsumption;
	int pwmAmps;
	char s[20];

	showStatusLine(p);

	energyMeter_getPower (&numPhases, &powerConsumption);
	showMessage(0,TX_CARINFO, powerConsumption, numPhases);  // %d watt\n%d phases

	pwmAmps = j1772_getPwmAmps ();
	if (pwmAmps > 0) {
		setFontSize(8);
		if (pwmAmps == j1772_PILOT_P12) strcpy(&s[0],"+12V");
		else if (pwmAmps <= 0) strcpy(&s[0],"-12V");
		else sprintf((char *)&s,"%d Amps",pwmAmps);
		u8g2_DrawStr(D, SCR_WIDTH - u8g2_GetStrWidth(D,s), SCR_HEIGHT-1, (char *)s);
	}

	u8g2_SendBuffer(D);
	return UI_PH_OK;
}

int handle_IPINFO_page (struct ui_page_t *p, ev_event_t UNUSED(event), ev_data_t UNUSED(*data), int) {
	char type[9];
	char ip[20];
	char gw[20];
	char nm[20];

	type[0] = 0;
	//ESP_LOGD(TAG,"handle_IPINFO_page");
	showStatusLine(p);

if (networkConnected) {
#ifdef ESP_PLATFORM
		typedef struct {
			uint8_t a0,a1,a2,a3;
		} ipv4addr_t;
		tcpip_adapter_ip_info_t ip_info;
		memset(&ip_info, 0, sizeof(tcpip_adapter_ip_info_t));
		tcpip_adapter_get_ip_info(TCPIP_ADAPTER_IF_ETH, &ip_info);
		if (ip_info.ip.addr != 0) {
			strcpy(type,"Ethernet");
		} else {
			tcpip_adapter_get_ip_info(TCPIP_ADAPTER_IF_STA, &ip_info);
			if (ip_info.ip.addr != 0) strcpy(type,"WLAN");
		}
#define IP2STR(ipaddr) ip4_addr1_16(ipaddr), \
		ip4_addr2_16(ipaddr), \
    	ip4_addr3_16(ipaddr), \
		ip4_addr4_16(ipaddr)

#define IPSTR "%d.%d.%d.%d"
		if (type[0] != 0) {
			sprintf(&ip[0],"%d,%d,%d,%d",((ipv4addr_t *) &ip_info.ip)->a0,((ipv4addr_t *) &ip_info.ip)->a1,((ipv4addr_t *) &ip_info.ip)->a2,((ipv4addr_t *) &ip_info.ip)->a3);
			sprintf(&nm[0],"%d,%d,%d,%d",((ipv4addr_t *) &ip_info.netmask)->a0,((ipv4addr_t *) &ip_info.netmask)->a1,((ipv4addr_t *) &ip_info.netmask)->a2,((ipv4addr_t *) &ip_info.netmask)->a3);
			sprintf(&gw[0],"%d,%d,%d,%d",((ipv4addr_t *) &ip_info.gw)->a0,((ipv4addr_t *) &ip_info.gw)->a1,((ipv4addr_t *) &ip_info.gw)->a2,((ipv4addr_t *) &ip_info.gw)->a3);
		}
#else
		strcpy(type,"Ethernet");
		strcpy(ip,"123.156.189.221");
		strcpy(nm,"255.255.255.0");
		strcpy(gw,"123.156.189.1");
#endif
	}
	if (type[0] == 0) {
#ifdef LANG_EN
		showMessage(1,"no network\nconnection");
#else
		showMessage(1,"keine\nNetzwerk-\nverbindung");
#endif
	} else {
        showMessage(0,	"%s\n" \
        				"IP: %s\n" \
						"GW: %s\n" \
						"NM: %s\n" \
						"MQ: %s",
						type,ip,gw,nm, mqttConnected ? TX_CONNECTED : TX_DISCONNECTED);
	}
	u8g2_SendBuffer(D);
	return UI_PH_OK;
}


// **********************************************************************************************************


void formatDateTime (time_t t, char * s, int bufSize) {
  if (t == -1) {
	  strcpy(s,"-");
	  return;
  }
#ifdef LANG_EN
  strftime(s, bufSize, "%Y-%m-%d %H:%M", localtime((const time_t*)&t));
#else
  strftime(s, bufSize, "%d.%m.%Y %H:%M", localtime((const time_t*)&t));
#endif
}

void decTime(time_t * t, time_t minTime) {
	time_t min;
	if (minTime == -1) min = time(NULL); else min = minTime;
	*t -= (60*30);

	if (*t < min) *t = -1;
}

void incTime(time_t * t, time_t minTime) {
	struct tm temp;
	int minute;
	time_t min;

	if (minTime == -1) min = time(NULL); else min = minTime;
	if (*t < min) {
		*t = min;
		localtime_r(t,&temp);
		temp.tm_sec = 0;
		minute = temp.tm_min; temp.tm_min = 0;
		*t = mktime(&temp);
		if (minute > 30) (*t) += 30*60;
	}
	*t += (60*30);
}


#define INP_RC_NONE	0
#define INP_RC_BACK	1
#define INP_RC_NEXT	2

int handleTimeInput (time_t * t, int keyCode, time_t minTime) {

	switch (keyCode) {
	case EV_KEY_UP:
		decTime(t, minTime);
		break;
	case EV_KEY_DN:
		incTime(t, minTime);
		break;
	case EV_KEY_OK:
		return INP_RC_NEXT;
		break;
	case EV_KEY_BACK:
		return INP_RC_BACK;
		break;
	}
	return INP_RC_NONE;
}


void displayTimeInput (const char * tx, time_t t) {
	char s[20];
	formatDateTime (t, (char *)&s, sizeof(s));
	showMessage(1,"%s\n%s",tx,s);
}

// **********************************************************************************************************

int handleIntInput (int * value, int min, int max, int keyCode) {
	if (*value > max) *value = max;
	if (*value < min) *value = min;
	switch (keyCode) {
	case EV_KEY_UP:
		(*value)++;
		break;
	case EV_KEY_DN:
		(*value)--;
		break;
	case EV_KEY_OK:
		return INP_RC_NEXT;
		break;
	case EV_KEY_BACK:
		return INP_RC_BACK;
		break;
	}
	if (*value > max) *value = max;
	if (*value < min) *value = min;
	return INP_RC_NONE;
}

void displayIntInput (const char * tx, int value) {
	showMessage(1,"%s\n%d",tx,value);
}

// **********************************************************************************************************

int handlePhasesInput (int * value, int keyCode) {
	switch (keyCode) {
	case EV_KEY_UP:
		(*value)++;
		break;
	case EV_KEY_DN:
		(*value)--;
		break;
	case EV_KEY_OK:
		return INP_RC_NEXT;
		break;
	case EV_KEY_BACK:
		return INP_RC_BACK;
		break;
	}
	if (*value > 7) *value = 7;
	if (*value < 0) *value = 0;
	return INP_RC_NONE;
}

void displayPhasesInput (const char * tx, int value) {
	char s[20];

	if (value == 0) strcpy(s,"auto");
	else {
		strcpy(s,"L1:0 L2:0 L3:0");
		if (value & 1) s[3] = '1';
		if (value & 2) s[8] = '1';
		if (value & 4) s[13] = '1';
	}
	showMessage(1,"%s\n%s",tx,s);
}

// **********************************************************************************************************

static j1772_manual_settings_t manSettings;


int handle_MANUAL_CHARGE_page	(struct ui_page_t *p, ev_event_t event, ev_data_t *data, int hasFocus) {

	int rc;

	showStatusLine(p);

	if (hasFocus) {
		if (j1772_manualModeEnabled ()) {
			if (event == EV_KEY_PRESS) {
				switch(*data) {
					case EV_KEY_OK:		// abort manual mode
						j1772_getManualSettings (&manSettings);
						manSettings.enabled = 0;
						j1772_updateManualSettings (&manSettings);
						u8g2_SendBuffer(D);
						return UI_PH_OK;	// exit focus mode
						break;
					case EV_KEY_BACK:
						u8g2_SendBuffer(D);
						return UI_PH_OK;	// exit focus mode without change
						break;
				}
			}

			showMessage(0,TX_MANEXIT);

			u8g2_SendBuffer(D);
			return UI_PH_HASFOCUS;
		}

		// we are not yet in manual mode
		rc = INP_RC_NONE;

		// handle data input
		if (event == EV_KEY_PRESS)
		switch (focusNumInput) {
			case 0:	rc = handleTimeInput (&manSettings.startTime, *data, -1); break;
			case 1: rc = handleTimeInput (&manSettings.endTime, *data, manSettings.startTime); break;
			case 2: rc = handleIntInput (&manSettings.man_amps , j1772_minAmps(),j1772_maxAmps(), *data); break;
			//case 3: rc = handlePhasesInput (&manSettings.man_phases, *data);
		}

		// handle prev/next/exit
		if (rc == INP_RC_BACK) {
			if (focusNumInput == 0) {
				u8g2_SendBuffer(D);
				return UI_PH_OK;		// exit focus mode
			}
			focusNumInput--;			// prev input
		} else
		if (rc == INP_RC_NEXT) {
			focusNumInput++;
			if (focusNumInput > 2) {	// start action
				manSettings.enabled = 1;
				j1772_updateManualSettings (&manSettings);
				u8g2_SendBuffer(D);
				return UI_PH_OK;
			}
		}


		// handle display data
		switch (focusNumInput) {
			case 0:	displayTimeInput (TX_STARTTIME, manSettings.startTime); break;
			case 1: displayTimeInput (TX_ENDTIME, manSettings.endTime); break;
			case 2: displayIntInput  (TX_MAXAMPS, manSettings.man_amps); break;
			//case 3: displayPhasesInput (TX_PHASES, manSettings.man_phases); break;
		}

		u8g2_SendBuffer(D);
		return UI_PH_HASFOCUS;

	} else {
		// we don't have focus yet
		j1772_getManualSettings (&manSettings);


		if ((event == EV_KEY_PRESS) & (*data == EV_KEY_OK)) {
			focusNumInput = 0;
			manSettings.startTime = -1;
			manSettings.endTime = -1;
			if (manSettings.man_amps < j1772_minAmps()) manSettings.man_amps = j1772_minAmps();
//			if (manSettings.man_amps > j1772_AMPS_MAX) manSettings.man_amps = j1772_AMPS_MAX;
			return UI_PH_HASFOCUS;
		}

		if (manSettings.enabled) {
			char s[20];
			char e[20];

			formatDateTime (manSettings.startTime,(char *)&s,sizeof(s));
			formatDateTime (manSettings.endTime,(char *)&e,sizeof(e));
			showMessage(1, TX_MANMODE_INFO,s,e,manSettings.man_amps);
		} else {
			showMessage(1, TX_MANCONF);
		}
	}
	u8g2_SendBuffer(D);
	return UI_PH_OK;
}

// **********************************************************************************************************

int phases;

int handle_PHASES_page	(struct ui_page_t *p, ev_event_t event, ev_data_t *data, int hasFocus) {

	int rc;
	int L1,L2,L3;

	showStatusLine(p);

	if (hasFocus) {
		rc = INP_RC_NONE;

		// handle data input
		if (event == EV_KEY_PRESS)
		switch (focusNumInput) {
			case 0: rc = handlePhasesInput (&phases, *data);
		}

		// handle prev/next/exit
		if (rc == INP_RC_BACK) {
			if (focusNumInput == 0) {
				u8g2_SendBuffer(D);
				return UI_PH_OK;		// exit focus mode
			}
			focusNumInput--;			// prev input
		} else
		if (rc == INP_RC_NEXT) {
			focusNumInput++;
			if (focusNumInput > 0) {	// start action
				L1 = 0; if (phases & 1) L1++;
				L2 = 0; if (phases & 2) L2++;
				L3 = 0; if (phases & 4) L3++;

				configfile.set_u32("PHASE_L1", L1);
				configfile.set_u32("PHASE_L2", L2);
				configfile.set_u32("PHASE_L3", L3);
				configfile.save();
				u8g2_SendBuffer(D);
				return UI_PH_OK;
			}
		}


		// handle display data
		switch (focusNumInput) {
			case 0: displayPhasesInput (TX_PHASES, phases); break;
		}

		u8g2_SendBuffer(D);
		return UI_PH_HASFOCUS;

	} else {
		L1 = configfile.get_u32("PHASE_L1");
		L2 = configfile.get_u32("PHASE_L2");
		L3 = configfile.get_u32("PHASE_L3");
		phases = L1 + (L2 << 1) + (L3 << 2);
		// we don't have focus yet
		if ((event == EV_KEY_PRESS) & (*data == EV_KEY_OK)) {
			focusNumInput = 0;
			return UI_PH_HASFOCUS;
		}

		if (phases == 0) showMessage(1, TX_AUTO);
		else showMessage(1, "L1:%d, L2:%d, L3:%d",L1,L2,L3);
	}
	u8g2_SendBuffer(D);
	return UI_PH_OK;
}

// **********************************************************************************************************

void ui_timer_next_page_callback (TimerHandle_t xTimer) {
	//ESP_LOGD(TAG,"ui_timer_next_page_callback");
	ev_send  (EV_TASK_UI, EV_UI_TIMER_PAGE, 0);		// queue an event for selecting the next page
	xTimerStart(xTimer,pdMS_TO_TICKS(100));			// and restart the timer
}

void ui_timer_timeout_callback (TimerHandle_t xTimer) {
	ESP_LOGI(TAG,"ui_timer_timeout_callback");
	ev_send  (EV_TASK_UI, EV_UI_TIMER_TIMEOUT, 0);	// queue an event for aborting a page in edit mode
	xTimerStart(xTimer,pdMS_TO_TICKS(100));			// and restart the timer
}


TaskHandle_t xDisplayTaskHandle = NULL;

#define DEF_FOCUS_PAGE_TIMEOUT

extern "C" void task_ui(void UNUSED(*ignore)) {
	TimerHandle_t ui_nextPageTimer;
	TimerHandle_t ui_timeoutPageTimer;
	ev_data_t evdata;
	ev_event_t event;
	int currUIPage = UI_PAGE_FIRST;
	int res;
	int currUIPage_hasFocus = 0;
	int autoPageChangeDelay = 0;
	int changeWasViaKeyboard = 0;

	u8g2_InitDisplay(D); // send init sequence to the display, display is in sleep mode after this,
	u8g2_SetPowerSave(D, 0); // wake up display
	u8g2_SetFontDirection(D, 0);
	u8g2_SetFontRefHeightAll(D);
	showStatusLine (NULL);  // show init in status line
	u8g2_SendBuffer(D);


	ui_nextPageTimer = xTimerCreate(	"UI-P",
										pdMS_TO_TICKS(AUTO_PAGE_CHANGE_TIME_MS),
										pdFALSE,       /* auto reload */
										(void *) 0,    /* some id */
										ui_timer_next_page_callback);
	if (ui_nextPageTimer == NULL) ESP_LOGE(TAG,"unable to create UI-P timer");
	if (xTimerStart (ui_nextPageTimer, 0) != pdPASS) ESP_LOGE(TAG,"unable to start UI-P timer");

	ui_timeoutPageTimer = xTimerCreate(	"UI-A",
										pdMS_TO_TICKS(TIMOUT_TIME_MS),
										pdFALSE,       /* auto reload */
										(void *) 0,    /* some id */
										ui_timer_timeout_callback);
	if (ui_timeoutPageTimer == NULL) ESP_LOGE(TAG,"unable to create UI-T timer");
	//ESP_LOGD(TAG, "entering main loop");
	while (1) {
		vTaskDelay(100 / portTICK_PERIOD_MS);
		res = ev_receive (EV_TASK_UI, &event, &evdata, 500 / portTICK_PERIOD_MS);
		changeWasViaKeyboard = 0;
		if (res == EV_RC_OK) {
			//ESP_LOGD(TAG,"got event %d",(int)event);

				switch(event) {
					case EV_END_TASK:
						xTimerDelete(ui_timeoutPageTimer, pdMS_TO_TICKS(1000));
						xTimerDelete(ui_nextPageTimer, pdMS_TO_TICKS(1000));
						ESP_LOGI (TAG,"terminating");
						xDisplayTaskHandle = NULL;
						vTaskDelete(NULL);
						break;
					case EV_KEY_PRESS:
						//ESP_LOGE (TAG,"KEY_PRESS %d",evdata);
						if (currUIPage_hasFocus) {
							xTimerReset(ui_timeoutPageTimer,50);  // reset the focused page timeout counter
						} else {
							switch(evdata) {
								case EV_KEY_BACK:  // back will force the current page to stay for a long time
									autoPageChangeDelay = AUTO_PAGE_WAIT_AFTER_KEY * 2;
									break;
								case EV_KEY_UP:
									currUIPage++;
									if (currUIPage >= UI_PG_EOL) currUIPage = UI_PAGE_FIRST;
									event = EV_UI_REDRAW;
									autoPageChangeDelay = AUTO_PAGE_WAIT_AFTER_KEY;
									changeWasViaKeyboard = 1;
									break;
								case EV_KEY_DN:
									if (currUIPage == UI_PAGE_FIRST) currUIPage = UI_PG_EOL;
									currUIPage--;
									event = EV_UI_REDRAW;
									autoPageChangeDelay = AUTO_PAGE_WAIT_AFTER_KEY;
									changeWasViaKeyboard = 1;
									break;
							}
						}
					break;
					case EV_NET_CONNECTED:
						networkConnected = 1;
						currUIPage = UI_PG_IPINFO;
						event = EV_UI_REDRAW;
						break;
					case EV_NET_DISCONNECTED:
						networkConnected = 0;
						availablePower = 0;
						currUIPage = UI_PG_IPINFO;
						event = EV_UI_REDRAW;
						break;
					case EV_AVAILABLE_PWR_CHANGE:
						availablePower = evdata;
						//currUIPage = UI_PG_POWER;
						//event = EV_UI_REDRAW;
						event = EV_IGNORE;
						break;
					case EV_CAR_POWER_PHASES_CHANGE:
						currentActivePhases = evdata;
						event = EV_UI_REDRAW;
						break;
					case EV_CABLE_CHANGE:
						if (evdata >= 0) {
							cableAmps = evdata;
							event = EV_UI_REDRAW;
						}
						break;

					case EV_MQTT_CONNECTED:
						mqttConnected = 1;
						event = EV_UI_REDRAW;
						break;
					case EV_MQTT_DISCONNECTED:
						mqttConnected = 0;
						event = EV_UI_REDRAW;
						break;
					case EV_EM_DATA_CHANGED:
						if (currUIPage != UI_PG_EM) event = EV_IGNORE;  // dont update on every energy meter event
						break;
					case EV_UI_TIMER_PAGE:
						if (currUIPage_hasFocus == 0) {
							if (autoPageChangeDelay) {
								autoPageChangeDelay--;
								event = EV_IGNORE;
							} else {
								currUIPage++;
								if (currUIPage >= UI_PG_EOL) {
									currUIPage = UI_PAGE_FIRST;
								} else {
									while (pages_data [currUIPage].flags && UI_PG_FLAG_NOAUTOSHOW) {
										currUIPage++;
										if (currUIPage >= UI_PG_EOL) currUIPage = UI_PAGE_FIRST;
									}
								}
								event = EV_UI_REDRAW;
							}
						}
						break;
				}

			if (event) {
				res = pages_data[currUIPage].fp(&pages_data[currUIPage], event, &evdata, currUIPage_hasFocus);
				if (res == UI_PH_HASFOCUS) {
					if (currUIPage_hasFocus == 0) {
						currUIPage_hasFocus = 1;
						pages_data[currUIPage].fp(&pages_data[currUIPage], EV_UI_REDRAW, &evdata, currUIPage_hasFocus);
					}
				} else {
					if (currUIPage_hasFocus) {
						currUIPage_hasFocus = 0;
						pages_data[currUIPage].fp(&pages_data[currUIPage], EV_UI_REDRAW, &evdata, currUIPage_hasFocus);
					}
				}
				if (event == EV_UI_TIMER_TIMEOUT) currUIPage_hasFocus = 0;
				if (res == UI_PH_DONTSHOW) {
					autoPageChangeDelay=0;
					if (changeWasViaKeyboard)
					  ev_send (EV_TASK_UI,EV_KEY_PRESS,evdata);
					else
					  ev_send (EV_TASK_UI,EV_UI_TIMER_PAGE,0);
				}
			}
		}
	}
	//vTaskDelete(NULL);
}


void display_setup(void) {
#ifdef ESP_PLATFORM
	u8g2_esp32_hal_t u8g2_esp32_hal = U8G2_ESP32_HAL_DEFAULT;
	u8g2_esp32_hal.sda   = PIN_SDA;
	u8g2_esp32_hal.scl  = PIN_SCL;
	u8g2_esp32_hal.no_i2c_init = true;    // i2c already initialized in evse_main.cpp
	u8g2_esp32_hal_init(u8g2_esp32_hal);


	u8g2_Setup_sh1106_128x64_noname_f(
		&u8g2,
		U8G2_R0,
		u8g2_esp32_msg_i2c_cb,
		u8g2_esp32_msg_i2c_and_delay_cb);

	u8x8_SetI2CAddress(&u8g2.u8x8,0x78);
#else   // ESP_PLATFORM
	u8g2_SetupBuffer_SDL_128x64(&u8g2, &u8g2_cb_r0);
#endif  // ESP_PLATFORM

	ESP_LOGI (TAG,"creating Task");
	xTaskCreatePinnedToCore(task_ui, "UI", 4096, NULL, EVSE_STD_TASK_PRIO, &xDisplayTaskHandle, APP_CPU);
	if (xDisplayTaskHandle == NULL) {
		ESP_LOGE(TAG,"xTaskCreate did not return a handle for the ui task");
		u8g2_InitDisplay(&u8g2); // send init sequence to the display, display is in sleep mode after this,
		u8g2_SetPowerSave(&u8g2, 0); // wake up display
		u8g2_ClearBuffer(&u8g2);
		u8g2_SetFont(&u8g2, u8g2_font_6x10_tr);
		u8g2_SetFontDirection(&u8g2, 0);
		u8g2_SetFontRefHeightAll(&u8g2);
		u8g2_DrawStr(&u8g2, 20, 10, "UITaskCreateErr");
		u8g2_SendBuffer(&u8g2);
		return;
	}
	ESP_LOGI (TAG,"UI set up and Task created");
}


