/* EVSE_AD - energyMeter
   Handles reading the SDM630 modbus energy meter
   SDM630-V2: https://bg-etech.de/download/manual/SDM630-Modbus-V2.pdf

   Armin Diehl <ad@ardiehl.de> December 2017

   Jun 10, 2018 AD:	added current and voltage readout
   Jun 11, 2018 AD:	added consumption (kwh) readout

   Unless required by applicable law or agreed to in writing, this
   software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
   CONDITIONS OF ANY KIND, either express or implied.
*/


#include "evse_global.h"
#include "energyMeter.h"

#include <lightmodbus/core.h>
#include <lightmodbus/master.h>

#include "cons.h"
#include "argtable3_ad.h"
#include <stdarg.h>

static const char* TAG = "energyMeter";

#define INC_ECHO_TEST

static int ignoreErrCount;

static int debug = 0;
static FILE * debug_stderr = NULL;
static const char * MODULE = "meter ";

static void DMSG(
        const char* tag,
        const char* format, ...) __attribute__ ((format (printf, 2, 3)));

static void DMSG(
        const char* tag,
        const char* format, ...) {
	if (debug) {
		if (debug_stderr != NULL) {
			va_list list;
			va_start(list, format);
			fprintf(debug_stderr,CONS_COLOR_STR_I("\r%s"),MODULE);
			fprintf(debug_stderr,CONS_COLOR_STR_T("%-20s"), tag);
			vfprintf(debug_stderr,format, list);
			va_end(list);
			fprintf(debug_stderr,"\r\n");
			fflush(debug_stderr);
		}
		if (ferror(debug_stderr) != 0) {
			//fclose(debug_stderr);
			debug_stderr = NULL;
			debug = 0;
			ESP_LOGE(TAG,"debug stdout no longer valid, debug mode disabled");
		}
	}
}

void dumpBuffer(char *recsend, char *buf, int len) {
	int addr = 0;
	char lineBuf[60];
	int charsInBuf = 0;
	char hexData[6];
	lineBuf[0]='\0';

	if (debug > 1) {
		while (len) {
			sprintf(&hexData[0],"%02x ",(unsigned char)*buf);
			strcat(&lineBuf[0],&hexData[0]);
			buf++; len--; charsInBuf++;
			if (charsInBuf==16) {
				DMSG(recsend,"%04x %s",addr,lineBuf);
				charsInBuf = 0;
				addr+=16;
				lineBuf[0] = '\0';
			}
		}
	}
	if (charsInBuf>0) {
		DMSG(recsend,"%04x %s",addr,lineBuf);
	}
}


#ifndef ESP_PLATFORM

#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <errno.h>

int serFd;

struct termios oldtio,newtio;

int initSerial () {
	char *port;

	port = configfile.get_str("em_serialPort");
	if (port == NULL) port = strdup(EM_UART_PORT);
	serFd = open(port, O_RDWR | O_NOCTTY );

	if (serFd <0) {fprintf(stderr,"unable to open '%s', errno: %d\n",port,errno); free (port); return 1; }

	DMSG(TAG,"Port %s opened",port);
	free (port);


	tcgetattr(serFd,&oldtio); /* save current port settings */

	bzero(&newtio, sizeof(newtio));
	newtio.c_cflag = EM_BAUD | CRTSCTS | CS8 | CLOCAL | CREAD;
	newtio.c_iflag = IGNPAR;
	newtio.c_oflag = 0;

	/* set input mode (non-canonical, no echo,...) */
	newtio.c_lflag = 0;

	newtio.c_cc[VTIME]    = 10;   /* inter-character timer 1 second */
	newtio.c_cc[VMIN]     = 1;   /* blocking read until 5 chars received */

	tcflush(serFd, TCIFLUSH);
	tcsetattr(serFd,TCSANOW,&newtio);

	return 0;
}

int uart_write_bytes(int, const char* src, size_t size) {
	return write (serFd, src, size);
}

int uart_read_bytes(int, uint8_t* buf, uint32_t maxLen, TickType_t) {
	int bytesReceived = 0;
	int bytesRemaining = maxLen;
	int res;

         //  printf("receive maxLen=%d\n",maxLen);
	while (bytesReceived<maxLen) {       /* loop for input */
		res = read(serFd,buf,bytesRemaining);   /* returns after 5 chars have been input */
		if (res <= 0) return 0;
		buf+=res; bytesRemaining-=res; bytesReceived+=res;
		//printf("received %d bytes\n",res);
	}
	return maxLen;
}


#else   // ESP_PLATFORM
int initSerial () {
	esp_err_t err;
	const uart_config_t uart_config = {
        .baud_rate = EM_BAUD,
        .data_bits = UART_DATA_8_BITS,
        .parity = UART_PARITY_DISABLE,
        .stop_bits = UART_STOP_BITS_1,
        .flow_ctrl = UART_HW_FLOWCTRL_DISABLE,
		.rx_flow_ctrl_thresh = 0,
		.use_ref_tick = 0

    };
    uart_param_config(EM_UART_NUM, &uart_config);
    uart_set_pin(EM_UART_NUM, EM_TXD_PIN, EM_RXD_PIN, UART_PIN_NO_CHANGE, UART_PIN_NO_CHANGE);
    // We won't use a buffer for sending data.
	err = uart_driver_install(UART_NUM_1, EM_RXD_BUFSIZE, EM_TXD_BUFSIZE, 0, NULL, 0);
	if (err != ESP_OK) ESP_LOGE(TAG,"uart_driver_install failed with code %d",err);
	uart_flush(EM_UART_NUM);
	ignoreErrCount = 2;
	return err;
}
#endif

void uartFlush() {
#ifdef ESP32
				uart_flush(EM_UART_NUM);
#else
				fsync(serFd);
#endif
}



float reg32tofloat (uint16_t r1, uint16_t r2) {
	uint16_t tab_reg[2];
	float f;
	
	tab_reg[0] = r2;
	tab_reg[1] = r1;
	
	memmove(&f,&tab_reg[0],sizeof(f));
	
	//printf (" [%04x %04x %4.2f]  ", r1,r2,f);
	
	return f;
}

	
// SDM630-V2 Power for L1,L2,L3
//#define PWR_REG_START 0x000c
//#define PWR_REG_COUNT 6
//#define CURR_REG_START

#define REG_START 0
#define REG_COUNT 18

#define KWH_REG_START 0x015a
#define KWH_REG_COUNT 6			/* L1,L2,L3 import kw/h */

//SDM630 has starting with 0, UL1,2,3  IL1,2,3   and  PL1,2,3
// so read 9 registers in one go

#define SLAVE_ID 1

int queryRegisters (ModbusMaster *mma, int slaveId, int startReg, int regCount) {
	const char * ME = __func__;
	int res, recLen;
	char buf[128];

	uartFlush();
	modbusBuildRequest04 (mma, slaveId, startReg, regCount);
	//res = write(fd, mma->request.frame, mma->request.length);
	res = uart_write_bytes(EM_UART_NUM, (const char *)mma->request.frame, mma->request.length);
	dumpBuffer((char *)"send", (char *)mma->request.frame,mma->request.length);

	if (res != mma->request.length) {
		DMSG (ME,"uart_write_bytes returned %d bytes instead of expected %d bytes (slaveId: %d, startReg: 0x%04x regCount: %d)",res,mma->request.length, slaveId, startReg, regCount);
		vTaskDelay(1000 / portTICK_PERIOD_MS);
		return 1;
	}
	recLen = uart_read_bytes(EM_UART_NUM, (unsigned char *)&buf[0], mma->predictedResponseLength, pdMS_TO_TICKS(150));
	dumpBuffer((char *)"recv", &buf[0],recLen);
	if (recLen != mma->predictedResponseLength) {
		if (! ignoreErrCount) {
			DMSG (ME,"uart_read_bytes returned %d instead of expected %d",recLen,mma->predictedResponseLength);
		}
		return 1;
	}
	// set received buffer for analyze
	mma->response.frame = (uint8_t *)&buf;
	mma->response.length = recLen;
	res = modbusParseResponse (mma);
	if (res != 0) {
		if (! ignoreErrCount) DMSG (ME,"modbusParseResponse returned err %d",res);
		return res;
	}
	return 0;
}


int queryEnergyMeter (int slaveId, ModbusMaster *mma, em_data_t * d) {
	//const char * ME = __func__;
	int res,reg;

	res = queryRegisters (mma, slaveId, REG_START, REG_COUNT);   // Volt L1,L2,L2, Amps L1,L2,L2  Power L1,L2,L3
	
	if ((res == 0) & (mma->data.regs != NULL)) {
		reg = 0;

		d->voltageL1 = reg32tofloat(mma->data.regs[reg], mma->data.regs[reg+1]); reg += 2;
		d->voltageL2 = reg32tofloat(mma->data.regs[reg], mma->data.regs[reg+1]); reg += 2;
		d->voltageL3 = reg32tofloat(mma->data.regs[reg], mma->data.regs[reg+1]); reg += 2;

		d->currL1    = reg32tofloat(mma->data.regs[reg], mma->data.regs[reg+1]); reg += 2;
		d->currL2    = reg32tofloat(mma->data.regs[reg], mma->data.regs[reg+1]); reg += 2;
		d->currL3    = reg32tofloat(mma->data.regs[reg], mma->data.regs[reg+1]); reg += 2;

		d->pwrL1 = (int)reg32tofloat(mma->data.regs[reg], mma->data.regs[reg+1]); reg += 2;
		d->pwrL2 = (int)reg32tofloat(mma->data.regs[reg], mma->data.regs[reg+1]); reg += 2;
		d->pwrL3 = (int)reg32tofloat(mma->data.regs[reg], mma->data.regs[reg+1]); reg += 2;

		return (0);
	} else
		return (1);
}

int queryEnergyMeterKWH (int slaveId, ModbusMaster *mma, em_data_t * d) {
	//const char * ME = __func__;
	int res,reg;

	res = queryRegisters (mma, slaveId, KWH_REG_START, KWH_REG_COUNT);   // L1,L2,L2 import KWH

	if ((res == 0) & (mma->data.regs != NULL)) {
		reg = 0;

		d->kwhL1 = (int)reg32tofloat(mma->data.regs[reg], mma->data.regs[reg+1]); reg += 2;
		d->kwhL2 = (int)reg32tofloat(mma->data.regs[reg], mma->data.regs[reg+1]); reg += 2;
		d->kwhL3 = (int)reg32tofloat(mma->data.regs[reg], mma->data.regs[reg+1]); reg += 2;
		d->kwh = d->kwhL1 + d->kwhL2 + d->kwhL3;

		return (0);
	} else
		return (1);
}


em_data_t em_data;

em_data_t * energyMeter_getCurrentData() {
	return &em_data;
}

int energyMeter_getPower (int * numPhases, int * powerConsumption) {
	*numPhases=0;
	if (em_data.pwrL1>0) (*numPhases)++;
	if (em_data.pwrL2>0) (*numPhases)++;
	if (em_data.pwrL3>0) (*numPhases)++;
	*powerConsumption = em_data.pwrL1+em_data.pwrL2+em_data.pwrL3;
    return 0;
}

int energyMeter_getPower3 (	int * powerConsumptionL1,
							int * powerConsumptionL2,
							int * powerConsumptionL3) {
	*powerConsumptionL1 = em_data.pwrL1;
	*powerConsumptionL2 = em_data.pwrL2;
	*powerConsumptionL3 = em_data.pwrL3;

	return em_data.pwrL1+em_data.pwrL2+em_data.pwrL3;
}

int energyMeter_getCurrent3(int * currentL1,
							int * currentL2,
							int * currentL3) {
	*currentL1 = (int)em_data.currL1;
	*currentL2 = (int)em_data.currL3;
	*currentL3 = (int)em_data.currL2;

	return (int)(em_data.currL1+em_data.currL2+em_data.currL3);
}


void energyMeter_setPower (int newL1,int newL2,int newL3) {
	const char * ME = __func__;
	if ((em_data.pwrL1 != newL1) | (em_data.pwrL2 != newL2) | (em_data.pwrL3 != newL3)) {
		em_data.pwrL1 = newL1; em_data.pwrL2 = newL2; em_data.pwrL3 = newL3;
		DMSG(ME,"sending EV_CONSUMED_PWR_CHANGE (%d %d %d)",newL1,newL2,newL3);
		ev_send (EV_TASK_UI+EV_TASK_MQTT+EV_TASK_J1772, EV_CONSUMED_PWR_CHANGE, em_data.pwrL1+em_data.pwrL2+em_data.pwrL3);
	}
	//ev_send (EV_TASK_J1772, EV_CONSUMED_PWR_CHANGE, pwrL1+pwrL2+pwrL3);	// send it always to j1772
}

ModbusMaster mma;
TaskHandle_t pvQueryEnergyMeterTaskHandle = NULL;
int initSerialDone = 0;

extern "C" void queryEnergyMeterTask(void * UNUSED(param)) {
	const char * ME = __func__;
	int res;
	int sleepTime = EM_QUERY_EVERY_MSECS;
	//int tmpPwrL1,tmpPwrL2,tmpPwrL3;
	int terminated = 0;
	ev_event_t event;
    ev_data_t data;
    em_data_t em_lastData;
    int queryKwhCountdown = 0;
	
    em_lastData.currL1 = 999;
	ESP_LOGI(TAG,"em task started");
	modbusMasterInit(&mma);
	do {
		res = ev_receive (EV_TASK_METER, &event, &data, pdMS_TO_TICKS(sleepTime));
		if (res == EV_RC_OK) {
    		switch(event) {
    			case EV_END_TASK:
    				terminated = 1;
    				break;
    		}
		}
		if (terminated == 0) {
			res = queryEnergyMeter (SLAVE_ID, &mma,&em_data);
			if (res != 0) {
				if (! ignoreErrCount) DMSG(ME,"queryEnergyMeter failed, pausing for %d seconds",EM_QUERY_ERRPAUSE_MSECS/1000);
				if (ignoreErrCount) ignoreErrCount--;
				sleepTime=EM_QUERY_ERRPAUSE_MSECS; // / portTICK_PERIOD_MS;
				uartFlush();
				em_lastData.currL1 = 999;
			} else {
				sleepTime=EM_QUERY_EVERY_MSECS;
				ignoreErrCount=0;
				energyMeter_setPower (em_data.pwrL1, em_data.pwrL2, em_data.pwrL3);

				if (queryKwhCountdown) {
					queryKwhCountdown--;
				} else {
					// read KW/h
					queryKwhCountdown = EM_QUERY_KWH_INTERVAL;
					res = queryEnergyMeterKWH (SLAVE_ID, &mma,&em_data);
					if (res != 0) {
						DMSG(ME,"queryEnergyMeterKWH failed");
					}
				}
				if (memcmp(&em_lastData,&em_data,sizeof(em_data)) != 0) {
					 memmove(&em_lastData,&em_data,sizeof(em_data));
					 ev_send(EV_TASK_UI,EV_EM_DATA_CHANGED,EV_DATA_NONE);	// for display energy data @ evse screen
				}
			}
        }
/*
        res = xTaskNotifyWait (pdFALSE,      // Don't clear bits on entry.
							   ULONG_MAX,    // Clear all bits on exit.
                               &terminated,  // Stores the notified value.
                               pdMS_TO_TICKS(sleepTime)); */
	} while (!(terminated));
	modbusMasterEnd(&mma);
	pvQueryEnergyMeterTaskHandle = NULL;
    ESP_LOGI(TAG,"queryEnergyMeterTask terminated");
#ifndef ESP_PLATFORM
    close(serFd);
    DMSG(TAG,"serFd closed");
    initSerialDone=0;
#endif  // ESP_PLATFORM
    vTaskDelete(NULL);  // without this -> crash
}




int queryEnergyMeter_endTask(FILE *fd) {
	if (pvQueryEnergyMeterTaskHandle != NULL) {
		ev_send (EV_TASK_METER, EV_END_TASK, EV_DATA_NONE);
		int i = 0;
		while ((i<50) && (pvQueryEnergyMeterTaskHandle != NULL)) {
			YIELD();
			i++;
		}
		if (pvQueryEnergyMeterTaskHandle == NULL) {
			fprintf(fd,"em task terminated\r\n");
			return 0;
		} else {
			fprintf(fd,"timeout waiting for em task to terminate\r\n");
			return 1;
		}
	} else fprintf(fd, CONS_COLOR_STR_W_N("unable to stop queryEnergyMeterTask - task not started"));
	return 1;
}



int queryEnergyMeter_startTask(FILE *fd) {
	esp_err_t err;
#ifndef ESP_PLATFORM
	em_data.voltageL1 = 231.42;
	em_data.voltageL2 = 230.31;
	em_data.voltageL3 = 236.921;
	em_data.kwh = 129749;
	em_data.kwhL1 = em_data.kwh;
#endif  // ESP_PLATFORM

	if (initSerialDone==0) {
		err = initSerial();
	
		if (err != ESP_OK) {
			fprintf(fd,CONS_COLOR_STR_E_N("\r\nstartQueryEnergyMeterTask: initSerial() failed"));
			return (err);
		}
		initSerialDone=1;
	}

	if (pvQueryEnergyMeterTaskHandle == NULL) {
		xTaskCreatePinnedToCore(queryEnergyMeterTask, "EnergyMeter", 4096, NULL, EVSE_STD_TASK_PRIO, &pvQueryEnergyMeterTaskHandle,APP_CPU);
		if (pvQueryEnergyMeterTaskHandle == NULL) {
			fprintf(fd,CONS_COLOR_STR_E_N("\r\nxTaskCreate did not return a handle for the queryEnergyMeterTask"));
			return 1;
		}
	} else
		fprintf(fd,CONS_COLOR_STR_E_N("\rqueryEnergyMeterTask already running"));
	return 0;
}

/*******************************************************************************/

#define ECHO_COUNT 10
#define ECHO_LENGTH 20

int echotest(FILE *out) {
	char sendBuf[ECHO_LENGTH];
	char recBuf[ECHO_LENGTH];
	int count,i;
	unsigned char currCh=1;
	esp_err_t err;
	int res;

	if (initSerialDone==0) {
		err = initSerial();

		if (err != ESP_OK) {
			fprintf(out,CONS_COLOR_STR_E_N("initSerial() failed"));
			return (err);
		}
		initSerialDone=1;
	}

	for (count=0;count<ECHO_COUNT;count++) {
		for (i=0;i<ECHO_LENGTH;i++) { sendBuf[i]=currCh; currCh++; }
		res = uart_write_bytes(EM_UART_NUM, (const char *)&sendBuf, ECHO_LENGTH);
		if (res != ECHO_LENGTH) {
			fprintf(out,CONS_COLOR_STR_E_N("uart_write_bytes returned %d, expected %d"),res,ECHO_LENGTH);
			return 1;
		}
		res = uart_read_bytes(EM_UART_NUM, (unsigned char *)&recBuf, ECHO_LENGTH, pdMS_TO_TICKS(150));
		if (res != ECHO_LENGTH) {
			fprintf(out,CONS_COLOR_STR_E_N("uart_read_bytes returned %d, expected %d"),res,ECHO_LENGTH);
			return 1;
		}
		for (i=0;i<ECHO_LENGTH;i++) {
			if (sendBuf[i] != recBuf[i]) {
				fprintf(out,CONS_COLOR_STR_E_N("verify error, send: 0x%02x, recived: 0x%02x"),(unsigned char)sendBuf[i],(unsigned char)recBuf[i]);
				return 1;
			}
		}
		fprintf(out,"%d: ok\r\n",count);
	}
	return 0;
}

int receiver(FILE *out) {
	char endBuf[4] = "   ";
	char recBuf[ECHO_LENGTH];
	int res,i;

	fprintf(out,"showing received serial data, send end to quit\r\n");

	do {
		res = uart_read_bytes(EM_UART_NUM, (unsigned char *)&recBuf, ECHO_LENGTH, pdMS_TO_TICKS(150));
		for (i=0; i<res; i++) {
			fprintf(out,"%02x ",(unsigned char)recBuf[i]);
			fflush(out);
			endBuf[0] = endBuf[1];
			endBuf[1] = endBuf[2];
			endBuf[2] = recBuf[i];
			if (strcmp(endBuf,"end") == 0) {
				fprintf(out,"\r\n\n");
				return 0;
			}
		}

	} while (1);

	return 0;
}

/*******************************************************************************/


static struct {
	struct arg_lit *start = arg_lit0("s", "start", "Start energy meter task");
	struct arg_lit *stop  = arg_lit0("t", "stop", "Stop energy meter task");
	struct arg_int *debug = arg_int0("d","debug", NULL, "0=off, 1=energy meter debug messages on");
#ifndef ESP_PLATFORM
	struct arg_str *port  = arg_str0("p", "device", NULL, "port/device for rs485 -> energy meter");
#endif  // ESP_PLATFORM
#ifdef INC_ECHO_TEST
	struct arg_lit *echo  = arg_lit0("e", "echotest", "serial echo test, send, receive and verify data");
	struct arg_lit *print = arg_lit0("p", "print", "serial test, print received data");
#endif
	struct arg_lit *info  = arg_lit0("i", "info", "Show current state");
	struct arg_int *L1 = arg_int0 (NULL,NULL,"<watt>","set consumed power L1 (valid only if em task not running)");
	struct arg_int *L2 = arg_int0 (NULL,NULL,"<watt>","set consumed power L2");
	struct arg_int *L3 = arg_int0 (NULL,NULL,"<watt>","set consumed power L3");
    struct arg_end *end = arg_end(2);
} emCommandArgs;



static int em_command(FILE *fd, FILE *, int, char**) {
	int atLeastOneParam = 0;

#ifndef ESP_PLATFORM
	if (emCommandArgs.port->count > 0) {
		configfile.set_str("em_serialPort",emCommandArgs.port->sval[0]);
		configfile.save();
		fprintf(fd,"port/device: %s\r\n",emCommandArgs.port->sval[0]);
		atLeastOneParam++;
	}
#endif  // ESP_PLATFORM
	if (emCommandArgs.debug->count > 0) {
		atLeastOneParam++;
		debug = emCommandArgs.debug->ival[0];
		if (debug == 0) {
			if (debug_stderr != NULL) {
				//fclose(debug_stderr);		// close debug file stream
				debug_stderr = NULL;
			}
		} else {
			//if (debug_stderr != NULL) fclose(debug_stderr);
			debug_stderr = fdopen (fileno(fd), "w");	// open it, we may be on another terminal
		}
		atLeastOneParam++;
		fprintf(fd,CONS_COLOR_STR_I_N("meter debug level: %d"),debug);
	}
#ifdef INC_ECHO_TEST
	if (emCommandArgs.echo->count) {
		if (pvQueryEnergyMeterTaskHandle == NULL) {
			echotest(fd);
		} else {
			fprintf(fd, CONS_COLOR_STR_E_N("em task needs to be stopped before running echo test"));
		}
		return 0;
	}
	if (emCommandArgs.print->count) {
		if (pvQueryEnergyMeterTaskHandle == NULL) {
			receiver(fd);
		} else {
			fprintf(fd, CONS_COLOR_STR_E_N("em task needs to be stopped before running echo test"));
		}
		return 0;
	}
#endif
	if ((emCommandArgs.start->count) && (emCommandArgs.stop->count)) {
		fprintf(fd,"what ? start or stop please\r\n");
		return 1;
	}
	if (emCommandArgs.start->count) {
		queryEnergyMeter_startTask(fd);
		atLeastOneParam++;
	}
	if (emCommandArgs.stop->count) {
		queryEnergyMeter_endTask(fd);
		atLeastOneParam++;
	}
	int L1;
	int L2=0;
	int L3=0;

	if (emCommandArgs.L1->count > 0) {
		atLeastOneParam++;
		L1 = emCommandArgs.L1->ival[0];
		if (emCommandArgs.L2->count > 0) L2 = emCommandArgs.L2->ival[0];
		if (emCommandArgs.L3->count > 0) L3 = emCommandArgs.L3->ival[0];
		energyMeter_setPower(L1,L2,L3);
	};
	if (emCommandArgs.info->count > 0) {
		if (pvQueryEnergyMeterTaskHandle == NULL) {
			fprintf(fd, CONS_COLOR_STR_W_N("queryEnergyMeterTask not running"));
		}
		int phases = (em_data.pwrL1>0) + (em_data.pwrL2>0) + (em_data.pwrL3>0);
		fprintf(fd, CONS_COLOR_I "Power L1:%d L2:%d L3:%d, active phases: %d, total power:%d" CONS_RESET_COLOR "\r\n",
		  em_data.pwrL1,em_data.pwrL2,em_data.pwrL3,phases,em_data.pwrL1+em_data.pwrL2+em_data.pwrL3);
		atLeastOneParam++;
	}
	if (atLeastOneParam == 0) {
		arg_print_glossary(fd, (void**) &emCommandArgs, NULL);
	}
	return 0;
}


void register_energyMeterCommands() {  /* register console commands */
	CMD_REG(em, "start, stop or query energyMeter task", NULL, &em_command, &emCommandArgs);
}
