/*
 * evse_commands.cpp
 *
  * handle evse specific console commands
 *
 *  Created on: Nov 17, 2018
 *      Author: Armin Diehl <ad@ardiehl.de>
 *
 */

#include "evse_commands.h"
#include <stdbool.h>
#include "cons.h"
#include "argtable3_ad.h"
#include "network.h"

#ifdef ESP32
#include "esp_wifi.h"
#include "tcpip_adapter.h"
//#include "esp_event_loop.h"
#include "esp_system.h"
#include "commands.h"
#endif

#include "evse_global.h"
#include "globalLock.h"

//****************************************************************************************

static struct {
    //struct arg_int *timeout = arg_int0(NULL, "timeout", "<t>", "Connection timeout, ms, default=5000");
	struct arg_int *L1     = arg_int0("1", "L1", "0|1", "L1 supported by car");
	struct arg_int *L2     = arg_int0("2", "L2", "0|1", "L2 supported by car");
	struct arg_int *L3     = arg_int0("3", "L3", "0|1", "L3 supported by car, set L1,2,3=0 for auto detection");
	struct arg_int *minAmps= arg_int0("a", "minamps", NULL, "set minimum Amps supported by car, default=6");
	struct arg_int *maxAmps= arg_int0("x", "maxamps", NULL, "set maximum Amps supported by charger, default=32");
	struct arg_lit *list   = arg_lit0("l", "list",      "list config");
	struct arg_str *ntpsrv = arg_str0("n", "ntp", NULL, "name or ip of ntp server");
	struct arg_lit *nontp  = arg_lit0("o", "nontp",     "remove ntp server from config (try dhcp)");
	struct arg_str *tz     = arg_str0("z", "tz" , NULL, "set timezone, defaults to " TZ_DEFAULT);
    struct arg_end *end = arg_end(10);
} config_commandArgs;


#ifdef ESP32
void showIPInfo(FILE *fd) {
	tcpip_adapter_ip_info_t ip_info;
	typedef struct {
			uint8_t a0,a1,a2,a3;
	} ipv4addr_t;

	if (tcpip_adapter_get_ip_info(TCPIP_ADAPTER_IF_ETH, &ip_info) == 0) {
		fprintf(fd," eth: IP: " IPSTR "  GW: " IPSTR " Netmask: " IPSTR "\r\n",IP2STR(&ip_info.ip),IP2STR(&ip_info.gw),IP2STR(&ip_info.netmask));
	}
	if (tcpip_adapter_get_ip_info(TCPIP_ADAPTER_IF_STA, &ip_info) == 0) {
		fprintf(fd,"Wlan: IP: " IPSTR "  GW: " IPSTR " Netmask: " IPSTR "\r\n",IP2STR(&ip_info.ip),IP2STR(&ip_info.gw),IP2STR(&ip_info.netmask));
	}
}
#else
void showIPInfo(FILE *) {}
#endif


static int config_command(FILE *fd, FILE *, int, char**) {
	int L1,L2,L3;
	int atLeastOneParam = 0;
	int amps;


	L1=0; L2=0; L3=0;
	if (config_commandArgs.L1->count) { L1 = config_commandArgs.L1->ival[0]; atLeastOneParam++; }
	if (config_commandArgs.L2->count) { L2 = config_commandArgs.L2->ival[0]; atLeastOneParam++; }
	if (config_commandArgs.L3->count) { L3 = config_commandArgs.L3->ival[0]; atLeastOneParam++; }
	if (atLeastOneParam) {
		configfile.set_u32("PHASE_L1", L1); configfile.set_u32("PHASE_L2", L2); configfile.set_u32("PHASE_L3", L3);
		configfile.set_u32("PHASES",L1+L2+L3);
		configfile.save();
		fprintf (fd,CONS_COLOR_STR_I("supported phases saved as L1=%d, L2=%d, L3=%d\r\n"),L1,L2,L3);
	}

	if (config_commandArgs.minAmps->count) {
		amps = config_commandArgs.minAmps->ival[0]; atLeastOneParam++;
		configfile.set_u32("MINAMPS", amps);
		configfile.save();
		ev_send(EV_TASK_UI+EV_TASK_J1772,EV_MINAMPS_CHANGE,amps);
	}

	if (config_commandArgs.maxAmps->count) {
		amps = config_commandArgs.maxAmps->ival[0]; atLeastOneParam++;
		configfile.set_u32("MAXAMPS", amps);
		configfile.save();
		ev_send(EV_TASK_UI+EV_TASK_J1772,EV_MAXAMPS_CHANGE,amps);
	}

	if (config_commandArgs.tz->count) {
		configfile.set_str("TZ", config_commandArgs.tz->sval[0]);
		configfile.save();
		setTimezone();
	}

	if (config_commandArgs.nontp->count) {
		configfile.set_str("NTP_SERVER", NULL);
		configfile.save();
		atLeastOneParam++;
		initialize_sntp();
	}

	if (config_commandArgs.ntpsrv->count) {
		configfile.set_str("NTP_SERVER", config_commandArgs.ntpsrv->sval[0]);
		configfile.save();
		atLeastOneParam++;
		initialize_sntp();
	}

	if (config_commandArgs.list->count > 0) {
		fprintf(fd,CONS_COLOR_STR_I("----<config file>---------------------\r\n"));
		catFile(fd,configfile.filename());
		fprintf(fd,CONS_COLOR_STR_I("----<eof>-----------------------------\r\n"));
		atLeastOneParam++;
		showIPInfo(fd);
	}

	if (atLeastOneParam == 0) {
		arg_print_glossary(fd, (void**) &config_commandArgs, NULL);
	}
	return 0;
}

/*******************************************************************************/

//****************************************************************************************




/*******************************************************************************/

void register_evseCommands() {  /* register console commands */
	CMD_REG2(config		,"config settings for evse", NULL);
}

void evseCommands_initialize(FILE *) {
	register_evseCommands();
}

