/*
 * evse_keys.cpp
 *
 * handle keys
 * the keys i used are china sensor-touchkeys so they will not bounce. However, the code implements
 * some simple debaunce in case other keys will be used in the future
 *
 *  Created on: Feb 3, 2018
 *      Author: Armin Diehl <ad@ardiehl.de>
 *
 *  June 24, 2018 AD: input changed to mcp23017 (i2c)
 *  June 25, 2018 AD: allow start/stop/debug
 */

#include "evse_keys.h"
#ifdef USE_PIO
#include "mcp23017_portio.h"
#endif
#include <stdbool.h>
#include "cons.h"
#include "argtable3_ad.h"
#include <stdarg.h>

static int debug = 0;
static FILE * debug_stderr = NULL;
static const char * MODULE = "meter ";

static void DMSG(
        const char* tag,
        const char* format, ...) __attribute__ ((format (printf, 2, 3)));

static void DMSG(
        const char* tag,
        const char* format, ...) {
	if (debug) {
		if (debug_stderr != NULL) {
			va_list list;
			va_start(list, format);
			fprintf(debug_stderr,CONS_COLOR_STR_I("\r%s"),MODULE);
			fprintf(debug_stderr,CONS_COLOR_STR_T("%-20s"), tag);
			vfprintf(debug_stderr,format, list);
			va_end(list);
			fprintf(debug_stderr,"\r\n");
			fflush(debug_stderr);
		}
		if (ferror(debug_stderr) != 0) {
			//fclose(debug_stderr);
			debug_stderr = NULL;
			debug = 0;
			//ESP_LOGE(TAG,"debug stdout no longer valid, debug mode disabled");
		}
	}
}




typedef enum {
	KEY_WAIT_PRESS = 1,			/* wait for keypress */
	KEY_WAIT_AUTOREPEAT,		/* first wait after pressing before starting autorepeat */
	KEY_WAIT_RELEASE 			/* wait for release if not autorepeat is enabled for this key */
} keystate_t;



struct keydata_t {
	int			enableAutorepeat;
#ifdef USE_PIO
	uint8_t     pioBit;
#else
	gpio_num_t	gpio;
#endif
	int			delayCount;
	keystate_t	state;
	int			keyEventData;
	char        name[5];
};

#ifdef USE_PIO
keydata_t keys[KEY_MAX] = {
		{ true,  KEY_BIT_UP,   0, KEY_WAIT_PRESS, EV_KEY_UP,  "UP" },
		{ true,  KEY_BIT_DN,   0, KEY_WAIT_PRESS, EV_KEY_DN,  "DOWN" },
		{ false, KEY_BIT_OK,   0, KEY_WAIT_PRESS, EV_KEY_OK,  "OK" },
		{ false, KEY_BIT_BACK, 0, KEY_WAIT_PRESS, EV_KEY_BACK,"BACK" }
};

#else
keydata_t keys[KEY_MAX] = {
		{ true,  KEY_PORT_UP, 0, KEY_WAIT_PRESS, EV_KEY_UP, "UP" },
		{ true,  KEY_PORT_DN, 0, KEY_WAIT_PRESS, EV_KEY_DN, "DOWN" },
		{ false, KEY_PORT_OK, 0, KEY_WAIT_PRESS, EV_KEY_OK, "OK" },
		{ false, KEY_PORT_BACK, 0, KEY_WAIT_PRESS, EV_KEY_BACK, "BACK" }
};

#define READ_KEY_VAL(n) gpio_get_level(keys[n].gpio)
#endif

TaskHandle_t keysTaskHandle = NULL;

#ifdef USE_PIO
uint8_t currPortVal;

int readKeyVal (int currKey, int hardRead) {
const char * ME = __func__;
	if (hardRead) {
		uint8_t lastVal = currPortVal;
		pio_getByte (KEY_PIO_IDX, KEY_PIO_BYTE, &currPortVal);
		if (currPortVal != lastVal)
			DMSG(ME,"read 0x%02x, lastVal: 0x%02x",currPortVal,lastVal);
	}
	if (currPortVal & (1 << keys[currKey].pioBit)) return 1;
	return 0;
}
#else
int readKeyVal (int currKey, int) {
	return gpio_get_level(keys[currKey].gpio);
}
#endif


volatile int keysTaskQuit;




extern "C" void keysTask(void * UNUSED(param)) {
	const char * ME = __func__;
	int currKey=0;
	keysTaskQuit = 0;
	DMSG(ME,"task started");
	do {
		switch (keys[currKey].state) {
			case KEY_WAIT_PRESS:
				if (readKeyVal(currKey,(currKey == 0))) {
					DMSG(ME,"keypress %d (%s)",currKey,keys[currKey].name);
					ev_send (EV_TASK_KEYS,EV_KEY_PRESS, keys[currKey].keyEventData);	// send the key event
					if (keys[currKey].enableAutorepeat) {
						keys[currKey].state = KEY_WAIT_AUTOREPEAT;
						keys[currKey].delayCount = KEY_AUTOREPEAT_WAIT_FIRST;
					} else {
						keys[currKey].state = KEY_WAIT_AUTOREPEAT;
					}
					vTaskDelay(pdMS_TO_TICKS(KEY_DEBOUNCE_MS));
				} else {
					// set next key to check
					currKey++;
					if (currKey == KEY_MAX) {
						currKey=0;
						vTaskDelay(pdMS_TO_TICKS(KEY_WAIT_IF_NO_KEY));
					}
				}
				break;
			case KEY_WAIT_AUTOREPEAT:
				if (readKeyVal(currKey,true)) {
					keys[currKey].delayCount--;
					if (keys[currKey].delayCount == 0) {
						DMSG(ME,"keyrepeat %d (%s)",currKey,keys[currKey].name);
						ev_send (EV_TASK_KEYS,EV_KEY_PRESS, keys[currKey].keyEventData);	// send the key event
						keys[currKey].delayCount = KEY_AUTOREPEAT_WAIT;
					}
				} else {
					DMSG(ME,"keyrelease %d (%s)",currKey,keys[currKey].name);
					keys[currKey].state = KEY_WAIT_PRESS;
#ifdef KEY_ENABLE_RELEASE_EVENT
					ev_send (EV_TASK_KEYS,EV_KEY_RELEASE, keys[currKey].keyEventData);	// send the key event
#endif
				}
				vTaskDelay(pdMS_TO_TICKS(100));
				break;
			case KEY_WAIT_RELEASE:
				if (readKeyVal(currKey,true) == 0) {
					DMSG(ME,"release %d",currKey);
					keys[currKey].state = KEY_WAIT_PRESS;
#ifdef KEY_ENABLE_RELEASE_EVENT
					ev_send (EV_TASK_KEYS,EV_KEY_RELEASE, keys[currKey].keyEventData);	// send the key event
#endif
				}
				vTaskDelay(pdMS_TO_TICKS(100));
		}
	} while (keysTaskQuit == 0);
	DMSG(ME,"task terminated");
	keysTaskHandle = NULL;
	vTaskDelete(NULL);
}



int keys_startTask(FILE *fp) {
	if (keysTaskHandle == NULL) {
		xTaskCreatePinnedToCore(keysTask, "Keyboard", 4096, NULL, EVSE_STD_TASK_PRIO, &keysTaskHandle,APP_CPU);
		if (keysTaskHandle == NULL) {
			fprintf(fp,CONS_COLOR_STR_E_N("\r\nxTaskCreate did not return a handle for the keyboard task"));
			return 1;
		}
	} else
		fprintf(fp,CONS_COLOR_STR_E_N("\rkeyboard task already running"));
	return 0;
}


int keys_endTask(FILE *fd) {
	if (keysTaskHandle != NULL) {
		keysTaskQuit = 1;
		int i = 0;
		while ((i<50) && (keysTaskHandle != NULL)) {
			YIELD();
			i++;
		}
		if (keysTaskHandle == NULL) {
			fprintf(fd,"keys task terminated\r\n");
			return 0;
		} else {
			fprintf(fd,"timeout waiting for keys task to terminate\r\n");
			return 1;
		}
	} else fprintf(fd, CONS_COLOR_STR_W_N("unable to stop keyboard task - task not started"));
	return 1;
}



void keys_initialize(FILE *fd) {

	if (keysTaskHandle == NULL) {
#ifdef USE_PIO
		uint16_t inputBits = (1 << KEY_BIT_UP) + (1 << KEY_BIT_DN) + (1 << KEY_BIT_BACK) + (1 << KEY_BIT_OK);
		pio_setPinsInput (KEY_PIO_IDX, inputBits);

#else
		for (int i=0; i<KEY_MAX; i++) {
			gpio_set_direction(keys[i].gpio,GPIO_MODE_INPUT);		// input
			gpio_set_pull_mode(keys[i].gpio, GPIO_PULLDOWN_ONLY);	// and add a pulldown in case it is not connected
		}
#endif
		keys_startTask(fd);
	}
	return;
}

/*******************************************************************************/

static struct {
	struct arg_lit *start = arg_lit0("s", "start", "Start keyboard task");
	struct arg_lit *stop  = arg_lit0("t", "stop", "Stop keyboard task");
	struct arg_int *debug = arg_int0("d","debug", NULL, "0=off, 1=keyboard debug messages on");
    struct arg_end *end = arg_end(2);
} keys_commandArgs;



static int keys_command(FILE *fd, FILE *, int, char**) {
	int atLeastOneParam = 0;

	if (keys_commandArgs.debug->count > 0) {
		atLeastOneParam++;
		debug = keys_commandArgs.debug->ival[0];
		if (debug == 0) {
			if (debug_stderr != NULL) {
				//fclose(debug_stderr);		// close debug file stream
				debug_stderr = NULL;
			}
		} else {
			//if (debug_stderr != NULL) fclose(debug_stderr);
			debug_stderr = fdopen (fileno(fd), "w");	// open it, we may be on another terminal
		}
		atLeastOneParam++;
		fprintf(fd,CONS_COLOR_STR_I_N("keys debug level: %d"),debug);
	}

	if ((keys_commandArgs.start->count) && (keys_commandArgs.stop->count)) {
		fprintf(fd,"what ? start or stop please\r\n");
		return 1;
	}
	if (keys_commandArgs.start->count) {
		keys_startTask(fd);
		atLeastOneParam++;
	}
	if (keys_commandArgs.stop->count) {
		keys_endTask(fd);
		atLeastOneParam++;
	}
	if (atLeastOneParam == 0) {
		arg_print_glossary(fd, (void**) &keys_commandArgs, NULL);
	}
	return 0;
}


void register_keysCommands() {  /* register console commands */
	CMD_REG(keys, "start / stop keyboard task", NULL, &keys_command, &keys_commandArgs);
}

