/* EVSE_AD - j1772
   Handles communication with car
     - detect cable amps
     - set available amps to car
     - detect supported phases
     - adjust charge current based ony available solar power

   Armin Diehl <ad@ardiehl.de> December 2017

   Jun 10, 2018
    - first tests, duty cycle was a little bit low
   Jun 12, 2018
    - fixed pwm at 100=+12V we had short pulses and duty was a little bit to low
   Unless required by applicable law or agreed to in writing, this
   software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
   CONDITIONS OF ANY KIND, either express or implied.
*/


#include "evse_global.h"
#include "j1772pilot.h"
#include "evse_event.h"
#include "cons.h"
#include "argtable3_ad.h"

#ifdef ESP_PLATFORM
#include "driver/ledc.h"
#include "driver/gpio.h"
#include "driver/adc.h"
#include "esp_adc_cal.h"
#else
#include "esp_compat.h"
#endif

#include "energyMeter.h"
#include <stdarg.h>
#include <stdlib.h>
#include <assert.h>
#include "queryInverter.h"

static const char *TAG = "J1772pilot";


// if a cable is connected, this will be the maximal possible current, otherwise 0
int cableAmps = 0;		// current detected cable maximum Amps
int chargeAmps;			// currently active pwm
j1772_status_t status;

int solarAvailableAmps = 0;			// from inverter (excluding the power we use, so available is this one + what we are currently using)
int emChargingAmps = 0;				// from energy meter
int emChargingWatts = 0;				// from energy meter
int maxPhasesSupportedByCar = 3;		// phases supported by car
int maxPhasesDetected = false;		// phases will be detected via energy Meter
int currNumPhases = 0;			// number of phases currently switched on
int relaisOffCounter = 0;
int phasesDetectCountdown = 0;
int supportedPhasesByCar[3];	// array of supported phases by car
int j1772_amps_min;				// minimum amps, defaults to 6
int j1772_amps_max;				//m maximum amps, defaults to 16
static j1772_manual_settings_t manSettings;

// for setting adc voltages in emulator
#ifndef ESP_PLATFORM
uint32_t adc1_ch6 = 3221;   // PP >3220=no cable
uint32_t adc1_ch7 = 2500;   // CP <2450=car connected
#endif


static FILE * debug_stderr = NULL;
static const char * MODULE = "j1772 ";
static int debug = 0;


static void DMSG(
        const char* tag,
        const char* format, ...) __attribute__ ((format (printf, 2, 3)));

const char * STATUS_TXT[] =
{ 	"J1772_CABLE_DISCONNECTED",
	"J1772_CABLE_CONNECTED",
	"J1772_CAR_READY",
	"J1772_CAR_CHARGE",
	"J1772_CAR_CHARGE_VENT",
	"J1772_CAR_ERROR"
};

const char * nullStr = "";

int j1772_getPwmAmps () {
	return chargeAmps;
}

const char * statusTxt(j1772_status_t sta) {
	if ( ((int)sta<0) || ((int)sta>=(int)J1772_LAST) ) {
		return nullStr;
	}
	return STATUS_TXT[(int)sta];
}


static void DMSG(
        const char* tag,
        const char* format, ...) {
	if (debug) {
		if (debug_stderr != NULL) {
			va_list list;
			va_start(list, format);
			fprintf(debug_stderr,CONS_COLOR_STR_I("\r%s"),MODULE);
			fprintf(debug_stderr,CONS_COLOR_STR_T("%-20s"), tag);
			//fprintf(debug_stderr,": ");
			vfprintf(debug_stderr,format, list);
			va_end(list);
			fprintf(debug_stderr,CONS_COLOR_STR_W(" status:%s(%d)"),statusTxt(status),(int) status);
			fprintf(debug_stderr,"\r\n");
			fflush(debug_stderr);
		}
		if (ferror(debug_stderr) != 0) {
			fclose(debug_stderr);
			debug_stderr = NULL;
			debug = 0;
			ESP_LOGE(TAG,"debug stdout no longer valid, debug mode disabled");
		}
	}
}


static int powerRelaisLastMask = 0xff;

void resetSupportedPhases () {
	supportedPhasesByCar[0] = BIT0;
	supportedPhasesByCar[1] = BIT0 + BIT1;
	supportedPhasesByCar[2] = BIT0 + BIT1 + BIT2;
	maxPhasesDetected = false;
	phasesDetectCountdown = 0;
	maxPhasesSupportedByCar = 3;
}

void setSupportedPhases (int emL1W, int emL2W, int emL3W) {
	const char * ME = __func__;

	int phase = 0;
	int last = 0;
	maxPhasesSupportedByCar = 0;
	supportedPhasesByCar[0] = 0;
	supportedPhasesByCar[1] = 0;
	supportedPhasesByCar[2] = 0;


	if (emL1W >= j1772_DETECT_MIN_WATTS_PHASE)maxPhasesSupportedByCar++;
	if (emL2W >= j1772_DETECT_MIN_WATTS_PHASE)maxPhasesSupportedByCar++;
	if (emL3W >= j1772_DETECT_MIN_WATTS_PHASE)maxPhasesSupportedByCar++;

	if (emL1W >= j1772_DETECT_MIN_WATTS_PHASE) {
		supportedPhasesByCar[phase] |= BIT0;
		last = BIT0;
		phase++;
	}

	if (emL2W >= j1772_DETECT_MIN_WATTS_PHASE) {
		supportedPhasesByCar[phase] |= BIT1;
		supportedPhasesByCar[phase] |= last;
		last |= BIT1;
		phase++;
	}

	if (emL3W >= j1772_DETECT_MIN_WATTS_PHASE) {
		supportedPhasesByCar[phase] |= BIT2;
		supportedPhasesByCar[phase] |= last;
		last |= BIT2;
		phase++;
	}

	DMSG(ME,"L1:%d, L2:%d, L3:%d, maxPhasesSupportedByCar: %d, [0x%02x,0x%02x,0x%02x]",emL1W,emL2W,emL3W,maxPhasesSupportedByCar,
			supportedPhasesByCar[0],supportedPhasesByCar[1],supportedPhasesByCar[2]);
	powerRelaisLastMask = 0xff;
}


void setSupportedPhasesViaMask (int pMask) {
	if (pMask == 0) {
		resetSupportedPhases();
		return;
	}
	setSupportedPhases (
			(pMask & 1)*j1772_DETECT_MIN_WATTS_PHASE,
			(pMask & 2)*j1772_DETECT_MIN_WATTS_PHASE,
			(pMask & 4)*j1772_DETECT_MIN_WATTS_PHASE);
}

// L1: GPIO13
// L2: GPIO2
// L3: GPIO15


void setPowerRelais (int numPhases) {
	const char * ME = __func__;

	int phasesMask;
	char L1str[4] = "off";
	char L2str[4] = "off";
	char L3str[4] = "off";

	if (numPhases > 3) numPhases = 3;
	if (numPhases == 0) {
		phasesMask = 0;		// all off
	} else
	if (maxPhasesDetected) {
		phasesMask = supportedPhasesByCar[numPhases-1];	// bitmask of supported phases by car
	} else {
		phasesMask = 0xff;	// all phases on if we have not detected the capabilities of the car
	}

#ifdef USE_PIO
	uint8_t bitsToClear;
	uint8_t bitsToSet;
	bitsToClear = 0; bitsToSet = 0;
	if (phasesMask & BIT0) { bitsToSet |= (1 << j1772_REL_L1_BIT); strcpy(L1str,"on"); } else bitsToClear |= (1 << j1772_REL_L1_BIT);
	if (phasesMask & BIT1) { bitsToSet |= (1 << j1772_REL_L2_BIT); strcpy(L2str,"on"); } else bitsToClear |= (1 << j1772_REL_L2_BIT);
	if (phasesMask & BIT2) { bitsToSet |= (1 << j1772_REL_L3_BIT); strcpy(L3str,"on"); } else bitsToClear |= (1 << j1772_REL_L3_BIT);
	pio_clearPins (j1772_PIO_IDX, bitsToClear);
	pio_setPins (j1772_PIO_IDX, bitsToSet);
#else
	if (phasesMask & BIT0) { gpio_set_level(j1772_L1_GPIO,0); strcpy(L1str,"on"); } else gpio_set_level(j1772_L1_GPIO,1);
	if (phasesMask & BIT1) { gpio_set_level(j1772_L2_GPIO,0); strcpy(L2str,"on"); } else gpio_set_level(j1772_L2_GPIO,1);
	if (phasesMask & BIT2) { gpio_set_level(j1772_L3_GPIO,0); strcpy(L3str,"on"); } else gpio_set_level(j1772_L3_GPIO,1);
#endif
	currNumPhases = numPhases;
	if (powerRelaisLastMask != phasesMask) {
		DMSG(ME,"Phases: %d, L1:%s, L2:%s, L3:%s Mask:0x%02x",numPhases,L1str,L2str,L3str,phasesMask);
		powerRelaisLastMask = phasesMask;
		ev_send (EV_TASK_UI+EV_TASK_MQTT, EV_CAR_POWER_PHASES_CHANGE, phasesMask);
	}
}


int availableAmps() {
	if (manSettings.enabled) {
		if (manSettings.endTime != -1)
			if (manSettings.endTime < time(NULL)) {
				manSettings.enabled = 0;
			}
	}
    if (manSettings.enabled) {
		if ((manSettings.startTime == -1) || (manSettings.startTime >= time(NULL))) {
			if ((manSettings.endTime == -1) || (manSettings.endTime <= time(NULL))) {
				return manSettings.man_amps;
			}
		}
	}
	return solarAvailableAmps + emChargingAmps;
}


void setPinOut (gpio_num_t gpioNum) {
	gpio_config_t c;

	c.pin_bit_mask = (uint64_t)1 << (int)gpioNum;		// GPIO pin: set with bit mask, each bit maps to a GPIO
	c.mode = GPIO_MODE_OUTPUT;				// GPIO mode: set input/output mode
	c.pull_up_en = GPIO_PULLUP_DISABLE;		// GPIO pull-up
	c.pull_down_en = GPIO_PULLDOWN_DISABLE;	// GPIO pull-down
	c.intr_type = GPIO_INTR_DISABLE;

	esp_err_t err = gpio_config(&c);
	if (err) ESP_LOGE(TAG, "gpio_config returned %d",err);
}

void j1772pwm_portsInitialize() {
#ifdef USE_PIO
	uint16_t outputBits = (1 << j1772_REL_L1_BIT) + (1 << j1772_REL_L2_BIT) + (1 << j1772_REL_L3_BIT);
	pio_setPinsOutput (j1772_PIO_IDX, outputBits);
#else
	setPinOut(j1772_L1_GPIO); gpio_set_level(j1772_L1_GPIO,1);
	setPinOut(j1772_L2_GPIO); gpio_set_level(j1772_L2_GPIO,1);
	setPinOut(j1772_L3_GPIO); gpio_set_level(j1772_L2_GPIO,1);
	//setPinOut((gpio_num_t)j1772_pilot_GPIO); gpio_set_level((gpio_num_t)j1772_pilot_GPIO,1);
#endif
}




/* ad 11 Jun 2018
 * ledc does not work on 100% pwm, we will always get one clock (depending on resolution) pulse even of
 * we set duty to max (i can see that on my new scope)
 * works with ledc_stop. Tries to configure the port as gpio but after that ledc did not work
 */
#define J1772_LEDC_TIMER      LEDC_TIMER_0
#define J1772_LEDC_CHANNEL    LEDC_CHANNEL_0
#define J1772_LEDC_TIMER_RES  LEDC_TIMER_10_BIT
// duty: Set the LEDC duty, the duty range is [0, (2**bit_num) - 1]
#define J1772_DUTY_MAX        1023


void ledc_init() {
	const char * ME = __func__;
	ESP_LOGD(ME,"lecd_init");
#ifdef ESP_PLATFORM
	esp_err_t err;
	/* Prepare and set configuration of timers
	 * that will be used by LED Controller */
	ledc_timer_config_t ledc_timer;
	memset((void *)&ledc_timer,0,sizeof(ledc_timer));
	ledc_timer.speed_mode = LEDC_HIGH_SPEED_MODE;		// timer mode
	ledc_timer.duty_resolution = J1772_LEDC_TIMER_RES;	// resolution of PWM duty
	ledc_timer.timer_num = J1772_LEDC_TIMER;			// timer index
	ledc_timer.freq_hz = 1000;							// frequency of PWM signal
	//ledc_timer_resume(LEDC_HIGH_SPEED_MODE, J1772_LEDC_TIMER);

	// Set configuration of timer0 for high speed channels
	err = ledc_timer_config(&ledc_timer);
	DMSG(ME,"ledc_timer_config returned %d",err);

	ledc_channel_config_t ledc_channel = {
			.gpio_num   = j1772_pilot_GPIO,
			.speed_mode = LEDC_HIGH_SPEED_MODE,
			.channel    = J1772_LEDC_CHANNEL,
			.intr_type  = LEDC_INTR_DISABLE,
			.timer_sel  = J1772_LEDC_TIMER,
			.duty       = J1772_DUTY_MAX,
			.hpoint     = 0					// TODO: AD 10.11.2018: new, does 0 work (0xfffff does not work)
	};
	err = ledc_channel_config(&ledc_channel);
	DMSG(ME,"ledc_channel_config returned %d",err);

	// on off signal for pilot
	setPinOut (j1772_pilotOff_GPIO);
	gpio_set_level(j1772_pilotOff_GPIO,0);
#endif
}



void j1772pwm_onoff (int on) {
	gpio_set_level(j1772_pilotOff_GPIO,on);
}



void j1772_ledc_config(int pwmPercent, int on) {
	const char * ME = __func__;

	if ((pwmPercent == 0) | (pwmPercent == 100)) {
#ifdef ESP_PLATFORM
		// set the port to 0 or 1 without pwm
		ledc_stop(LEDC_HIGH_SPEED_MODE, J1772_LEDC_CHANNEL, (pwmPercent == 100) );
#else
		DMSG(TAG, "ledc_stop (%d) on: %d", (pwmPercent == 100), on);
#endif
	} else {
		uint32_t duty;
		// due to limited speed of the LM358, we need to add 2%
		//if (pwmPercent < 90) pwmPercent += 2;  // we now have LF353/LF412

		duty = (pwmPercent * J1772_DUTY_MAX) / 100;	// duty for timer (0..1023)
		DMSG(ME,"ledc_setduty/update_duty %d, percent: %d, on: %d",duty,pwmPercent,on);
#ifdef ESP_PLATFORM

		esp_err_t err;
		if (duty < J1772_DUTY_MAX-1) duty +=1;  // we are always a little bit to low (at the esp32 pin)

		err = ledc_set_duty(LEDC_HIGH_SPEED_MODE, J1772_LEDC_CHANNEL, duty);
		if (err != ESP_OK) ESP_LOGE(TAG, "ledc_set_duty returned %d", err);
		ledc_update_duty(LEDC_HIGH_SPEED_MODE, J1772_LEDC_CHANNEL);
#endif  // ESP_PLATFORM
		j1772pwm_onoff (on);	// enable or disable pilot output (to be able to set 0 Volt, needed at least for Volt/Ampera gen1, these cars will ignore pwm after a too long pause
	}
}

void j1772pwm_initialize(FILE *fd) {
	j1772pwm_portsInitialize();
	ledc_init();

	j1772_ledc_config(100,0);  // disable pilot out (set to 0 volt)
    j1772_startTask(fd);
}



void j1772pwm_setAmps(int Amps, int ignoreCableAmps) {
	int percentDuty;
	int pilotOn;

	pilotOn = 1;  // pilot signal enabled
	if (Amps == j1772_PILOT_OFF) {
		percentDuty = 100;  // +12
		pilotOn = 0;				// positive pilot signal disabled
	} else
	if (Amps == j1772_PILOT_P12) {
		percentDuty = 100;  // +12V
	} else
	if (Amps == j1772_PILOT_M12) {
		percentDuty = 0;		  // -12V
	} else {
		if (ignoreCableAmps == 0)
		  if (Amps > cableAmps) Amps = cableAmps;
	    if (Amps > j1772_amps_max) Amps = j1772_amps_max;  // Limit to max supported by charger
		if (Amps < j1772_amps_min) {
			percentDuty = 0;
			Amps = 0;
		} else {
			if (Amps < 51) percentDuty = Amps *100 / 60;
			else percentDuty = ((Amps * 10) / 25) + 64;
		}
	}
	j1772_ledc_config(percentDuty, pilotOn);		// set lecd duty (or pwm on/off)
    
    if (chargeAmps != Amps) {
    	chargeAmps = Amps;
    	if (percentDuty == 0)             DMSG(TAG, "j1772pwm_setAmps(-12V)");
    	else if (Amps == j1772_PILOT_P12) DMSG(TAG, "j1772pwm_setAmps(+12V)");
    	else if (Amps == j1772_PILOT_OFF) DMSG(TAG, "j1772pwm_setAmps(off)");
    	else                              DMSG(TAG, "j1772pwm_setAmps(%d): duty: %d %%", Amps, percentDuty);
    	ev_send (EV_TASK_UI+EV_TASK_MQTT, EV_PWMAMPS_CHANGE, chargeAmps);
    }
}


static struct {
	struct arg_lit *zero = arg_lit0("0", "off", "switch off pilot (0 Volt)");
	struct arg_lit *perc = arg_lit0("p", "percent", "parameter is %, not amps, 0=-12V, 100=+12V");
    struct arg_int *Amps = arg_int0(NULL, NULL, NULL, "Maximum available current (or % if -p is given)");
    struct arg_end *end = arg_end(5);
} pilot_commandArgs;



int pilot_command(FILE *, FILE *, int, char**)
{
	int percent;

	if (pilot_commandArgs.zero->count) {
		j1772pwm_setAmps(j1772_PILOT_OFF, 1);
	} else
	if (pilot_commandArgs.perc->count) {
		percent = pilot_commandArgs.Amps->ival[0];
		if (percent > 100) percent = 100;
		j1772_ledc_config(percent,1);
	} else
      j1772pwm_setAmps(pilot_commandArgs.Amps->ival[0], 1);
    return 0;
}


TaskHandle_t pvj1772TaskHandle = NULL;

esp_adc_cal_characteristics_t *adc_chars;

#define DEFAULT_VREF   1100

void adcInit() {
	adc1_config_width(ADC_WIDTH_BIT_10);
	adc1_config_channel_atten(ADC_CH_CABLEDETECT, ADC_ATTEN_DB_11);
	adc1_config_channel_atten(ADC_CH_PILOT, ADC_ATTEN_DB_11);
	if (adc_chars == NULL) {
		//Characterize ADC
		adc_chars = (esp_adc_cal_characteristics_t *)calloc(1, sizeof(esp_adc_cal_characteristics_t));
		esp_adc_cal_characterize(ADC_UNIT_1, ADC_ATTEN_DB_11, ADC_WIDTH_BIT_10, DEFAULT_VREF, adc_chars);

	}
}


void adcDeInit() {
	if (adc_chars != NULL) {
		free(adc_chars);
		adc_chars = NULL;
	}
}

#ifdef ESP_PLATFORM
int adcReadVoltage (adc1_channel_t ch, int samples) {
	uint32_t adc_reading;

	if (samples < 2) {
		adc_reading = adc1_get_raw(ch);
	} else {
		adc_reading = 0;
		for (int i = 0; i < samples; i++) {
			adc_reading += adc1_get_raw(ch);
        }
        adc_reading /= samples;
	}
	//Convert adc_reading to voltage in mV
	return esp_adc_cal_raw_to_voltage(adc_reading, adc_chars);
}
#else
int adcReadVoltage (int ch,int) {
	if (ch == ADC_CH_CABLEDETECT) return adc1_ch6;
	return adc1_ch7;
}
#endif /* ESP_PLATFORM */


// GPIO34, ADC1 CH6, returns voltage, sets cableAmps
#define PP_READ_COUNT 3
int adcReadCableCurrent () {
	uint32_t voltage;
	int cableAmpsOld;

	cableAmpsOld = cableAmps;

    voltage = adcReadVoltage(ADC_CH_CABLEDETECT,PP_READ_COUNT);

	if (voltage > 2900) cableAmps = 0;
	else if (voltage > 2250) cableAmps = 13;
	else if (voltage > 1150) cableAmps = 20;
	else if (voltage > 570) cableAmps = 32;
	else if (voltage > 240) cableAmps = 64;
	else cableAmps = 0;


	if (cableAmps != cableAmpsOld) {
		DMSG(TAG,"adcReadCableCurrent %d mV, %d Amps, sending EV_CABLE_CHANGE",voltage,cableAmps);
		ev_send (EV_TASK_UI+EV_TASK_J1772+EV_TASK_MQTT, EV_CABLE_CHANGE, cableAmps);
	}
	return voltage;
}



static int adcReadCPlast = -1;
// GPIO34, ADC1 CH7, returns voltage, sends event on change
int adcReadCP (j1772_status_t * statusNew) {
	int voltage, diffVoltage;

    voltage = adcReadVoltage (ADC_CH_CABLEDETECT, 1);
    voltage = voltage / 10; voltage = voltage * 10;

    if (statusNew) *statusNew = status;

    if (status == J1772_CABLE_DISCONNECTED) return voltage;

    diffVoltage = abs(adcReadCPlast - voltage);
	if (diffVoltage < 150) {		// to reduce number of messages only
		return voltage;
	}

	if (voltage < 1000) {
		if (status != J1772_CAR_CHARGE_VENT) ev_send (EV_TASK_UI+EV_TASK_J1772, EV_CAR_CHARGE_VENT, EV_DATA_NONE);
		if (statusNew != NULL) *statusNew = J1772_CAR_CHARGE_VENT;
	}
	else if (voltage < 1850) {
		if (status != J1772_CAR_CHARGE) ev_send (EV_TASK_UI+EV_TASK_J1772, EV_CAR_CHARGE, EV_DATA_NONE);
		if (statusNew != NULL) *statusNew = J1772_CAR_CHARGE;
	}
	else if (voltage < 2450) {
		if (status != J1772_CAR_READY) ev_send (EV_TASK_UI+EV_TASK_J1772, EV_CAR_READY, EV_DATA_NONE);
		if (statusNew) *statusNew = J1772_CAR_READY;
	} else {
		if (status != J1772_CABLE_CONNECTED) ev_send (EV_TASK_UI+EV_TASK_J1772, EV_CABLE_CHANGE, -1);
		if (statusNew) *statusNew = J1772_CABLE_CONNECTED;
	}

	DMSG(TAG,"adcReadCP %d mV",voltage);
	adcReadCPlast = voltage;
	return voltage;
}

void setDefaultPhases () {
	const char * ME = __func__;

	int L1,L2,L3;
	int i = configfile.get_u32("PHASES");
	if (i) {
		L1 = configfile.get_u32("PHASE_L1");
		L2 = configfile.get_u32("PHASE_L2");
		L3 = configfile.get_u32("PHASE_L3");
		setSupportedPhases (	L1*j1772_DETECT_MIN_WATTS_PHASE,
								L2*j1772_DETECT_MIN_WATTS_PHASE,
								L3*j1772_DETECT_MIN_WATTS_PHASE);
		if (maxPhasesSupportedByCar > 0) {  // calced by setSupportedPhases
			maxPhasesDetected = true;
			//setPowerRelais(maxPhasesSupportedByCar);
			DMSG(ME,"Autodetect phases disabled by config, using %d phases, L1:%d, L2:%d, L3:%d",maxPhasesSupportedByCar,L1,L2,L3);
		}
	}
}

void adjustChargeCurrent() {
	const char * ME = __func__;

	int availableAmpsPerPhase, availA;
	int oldNumPhases;
	int oldemChargingAmps;

	oldNumPhases = currNumPhases;
	oldemChargingAmps = emChargingAmps;

	// we can only set something if the car is ready or is charging
	if ((status != J1772_CAR_READY) && (status != J1772_CAR_CHARGE) && (status != J1772_CAR_CHARGE_VENT)) return;

	availA = availableAmps();
	//DMSG(ME,"availableAmps: %d",availA);

	if (availA < j1772_amps_min) {
		if (currNumPhases>0)
			if (relaisOffCounter == 0) relaisOffCounter = j1772_RELAY_DELAY;
		if (chargeAmps != j1772_PILOT_P12) {
			PWM_12V();	// signal ready but no PWM = no Amps available
			DMSG(ME,"availableAmps: %d, stopped charging",availA);
		}
		return;
	}
	relaisOffCounter = 0;  // we have at least the minimal power, dont turn relais off

	if (manSettings.enabled) {
		if (manSettings.man_phases) {
			setSupportedPhasesViaMask (manSettings.man_phases);
		} else {
			setDefaultPhases();
		}
		if (manSettings.man_amps) availA = manSettings.man_amps;
		setPowerRelais(maxPhasesSupportedByCar);

	} else {
		// if we have not yet detected the number of phases supported by car, start with 3
		// phases and minimum current until we get new numbers from energy meter or reached timeout
		if (maxPhasesDetected == false) {

			// check if we have a fixed configuration for num phases and phases assignment
			setDefaultPhases();

			if ((phasesDetectCountdown == 0) & (maxPhasesDetected == false)) {
				if (maxPhasesDetected == false) {
					DMSG(ME,"detecting phases used by car, 3 phases are on, minimum charge current signaled");
					resetSupportedPhases ();
					phasesDetectCountdown = j1772_PHASEDETECT_DELAY;	// timeout timer
					setPowerRelais(3);									// turn on all 3 phases
					j1772pwm_setAmps(j1772_amps_min,0);					// 6 AMPS per phase
				} else {
					DMSG(ME,"using configured number of phases (%d) and phase assignment (0x%02x 0x%02x 0x%02x)",maxPhasesSupportedByCar,supportedPhasesByCar[0],supportedPhasesByCar[1],supportedPhasesByCar[2]);
				}
				return;		// wait until timeout or phases detected
			}
		}

		// always try to use as max phases as possible
		if (maxPhasesSupportedByCar == 3) {
			availableAmpsPerPhase = availA / 3;

			if (currNumPhases == 3) {	// are we already on 3 phases ?
				// then we can leave 3 phases on if we are above the minimum current for each phase
				if (availableAmpsPerPhase >= j1772_amps_min) {
					DMSG(ME,"adjusting 3 phase current from %d to %d (amps available: %d)",emChargingAmps,availableAmpsPerPhase,availA);
					j1772pwm_setAmps(availableAmpsPerPhase,0);	// adjust pwm to the current for each phase
					return;
				}
			}
			if (availableAmpsPerPhase >= j1772_AMPS_MIN_SWITCH_TO_INC_PHASES) {
				// we have a little bit more than the minimum current for 3 phases, switch to 3 phases
				PWM_12V();				// means pause
				setPowerRelais(3);		// turn on all 3 phases
				vTaskDelay(pdMS_TO_TICKS(j1772_CHANGE_PHASE_DELAY_MS));
				DMSG(ME,"changing from %d phases %d A to 3 phases %d A (amps available: %d)",oldNumPhases,oldemChargingAmps,availableAmpsPerPhase,availA);
				j1772pwm_setAmps(availableAmpsPerPhase,0);	// adjust pwm to the current for each phase
				return;
			}
		}

		// try 2 phases
		if (maxPhasesSupportedByCar == 2) {
			availableAmpsPerPhase = availA / 2;

			if (currNumPhases == 2) {	// are we already on 2 phases ?
				// then we can leave 2 phases on if we are above the minimum current for each phase
				if (availableAmpsPerPhase >= j1772_amps_min) {
					j1772pwm_setAmps(availableAmpsPerPhase,0);	// adjust pwm to the current for each phase
					return;
				}
			}
			if (availableAmpsPerPhase >= j1772_AMPS_MIN_SWITCH_TO_INC_PHASES) {
				// we have a little bit more than the minimum current for 3 phases, switch to 3 phases
				PWM_12V();				// means pause
				setPowerRelais(2);		// turn on 2 phases
				vTaskDelay(pdMS_TO_TICKS(j1772_CHANGE_PHASE_DELAY_MS));
				j1772pwm_setAmps(availableAmpsPerPhase,0);	// adjust pwm to the current for each phase
				return;
			}
		}

		// we have only one phase
		if (currNumPhases != 1) {	// are we already on 1 phase ?
			PWM_12V();				// means pause
			setPowerRelais(1);		// turn on 1 phase
			vTaskDelay(pdMS_TO_TICKS(j1772_CHANGE_PHASE_DELAY_MS));
		}
	}
	j1772pwm_setAmps(availA,0);	// adjust pwm to the current for each phase
}

// returns 1 is car is connected
int j1772_carConnected() {

	int retVal = 0;
	switch(status) {
    	case J1772_CAR_READY:
    	case J1772_CAR_CHARGE:
    	case J1772_CAR_CHARGE_VENT:
    		retVal = 1;
	}
	return retVal;
}

int j1772_carCharging() {  // // returns 1 if car is charging
	int retVal = 0;

	if ((status == J1772_CAR_CHARGE) | (status == J1772_CAR_CHARGE_VENT)) retVal++;
	return retVal;
}

void setState (j1772_status_t newState) {
	const char * ME = __func__;

	if (status == newState) return;
	status = newState;

	switch(newState) {
    	case J1772_CABLE_DISCONNECTED:
    		DMSG(ME,"J1772_CABLE_DISCONNECTED");
    		setPowerRelais(0);
    		PWM_MINUS12V();
    		resetSupportedPhases ();
       		break;
    	case J1772_CABLE_CONNECTED:
    		DMSG(ME,"J1772_CABLE_CONNECTED");
    		setPowerRelais(0);
    		PWM_12V();
    		resetSupportedPhases ();
       		break;
    	case J1772_CAR_READY:
    	case J1772_CAR_CHARGE:
    	case J1772_CAR_CHARGE_VENT:
    		// if we have some energy available, enable relays and tell car
    		DMSG(ME,"%s (%d)",statusTxt(newState),(int)newState);
    		adjustChargeCurrent();
    		break;
    	default:
    		DMSG(ME,"unknown state %s (%d)", statusTxt(newState),(int)newState);
	}
}


extern "C" void j1772Task(void * UNUSED(param)) {
	const char * ME = __func__;
	int res;
	ev_event_t event;
    ev_data_t data;
    int emNumPhases;
    int L1W,L2W,L3W;	// consumption from power meter in watt
    int solarAvailableAmpsOld;

	ESP_LOGI(TAG,"j1772 task started");
	status = J1772_CABLE_DISCONNECTED;
	adcInit();
	j1772pwm_setAmps(j1772_PILOT_P12,0); // and set +12V to the pilot line

	j1772_amps_min = configfile.get_u32 ("MINAMPS");
	if (j1772_amps_min < j1772_PWM_AMPS_MIN) {
		j1772_amps_min = j1772_PWM_AMPS_MIN;
		configfile.set_u32 ("MINAMPS", j1772_amps_min);
		configfile.save();
	}

	j1772_amps_max = configfile.get_u32 ("MAXAMPS");
	if (j1772_amps_max == 0) {
		j1772_amps_max = j1772_AMPS_MAX_DEFAULT;
		configfile.set_u32 ("MAXAMPS", j1772_amps_max);
		configfile.save();
	}


	do {
		res = ev_receive (EV_TASK_J1772, &event, &data, pdMS_TO_TICKS(250));
		if (res == EV_RC_OK) {
    		switch(event) {
    			case EV_MINAMPS_CHANGE:
    				j1772_amps_min = data;
    				adjustChargeCurrent();
    				break;
    			case EV_MAXAMPS_CHANGE:
    				j1772_amps_max = data;
    				adjustChargeCurrent();
    				break;
    			case EV_END_TASK:
    				pvj1772TaskHandle = NULL;
    				ESP_LOGI(TAG,"j1772Task terminated");
    				adcInit();
    				vTaskDelete(NULL);
    				break;
    			case EV_CABLE_CHANGE:
    				DMSG(ME,"EV_CABLE_CHANGE (%d)",data);
    				if (data == 0) {	// no cable detected
    					setState (J1772_CABLE_DISCONNECTED);
    				} else if (status == J1772_CABLE_DISCONNECTED) {
    					setState (J1772_CABLE_CONNECTED);	// new cable detected
    				} else if (data < 0) {
    					setState (J1772_CABLE_CONNECTED);	// car no longer connected but cable is connected
    				}
    				break;
    			case EV_AVAILABLE_PWR_CHANGE:
    				solarAvailableAmpsOld = solarAvailableAmps;
    				solarAvailableAmps = data / GRID_VOLTAGE;

    				if (solarAvailableAmps > cableAmps) {
    					if (solarAvailableAmpsOld != solarAvailableAmps)
    						DMSG(ME,"EV_AVAILABLE_PWR_CHANGE (%d) solarAvailableAmps: %d (%dW), limited to cableAmps (%d)",data,solarAvailableAmps,solarAvailableAmps * GRID_VOLTAGE,cableAmps);
    					solarAvailableAmps = cableAmps;
    				} else {
    					if (solarAvailableAmpsOld != solarAvailableAmps)
    						DMSG(ME,"EV_AVAILABLE_PWR_CHANGE (%d) solarAvailableAmps: %d (%dW)",data,solarAvailableAmps,solarAvailableAmps * GRID_VOLTAGE);
    				}
    				adjustChargeCurrent();
    				break;
    			case EV_CONSUMED_PWR_CHANGE:
    				energyMeter_getPower3 (&L1W,&L2W,&L3W);
    				energyMeter_getPower (&emNumPhases,&emChargingWatts);
    				emChargingAmps = emChargingWatts / GRID_VOLTAGE;
    				DMSG(ME,"EV_CONSUMED_PWR_CHANGE emPhases:%d, emWatts: %d, emAmps: %d (%d)",emNumPhases, emChargingWatts, emChargingAmps, data);
    				if (maxPhasesDetected == false) {
    					if (phasesDetectCountdown) {
    						if (emChargingWatts > 500) {
    							DMSG(TAG,"detected %d phases and %d watt (%d A) consumption",emNumPhases,emChargingWatts,emChargingAmps);
    							maxPhasesDetected = true;
    							setSupportedPhases (L1W, L2W, L3W);
    						}
    					}
    				}
    				adjustChargeCurrent();
    				break;
    			case EV_CAR_READY:
    				setState(J1772_CAR_READY);
    				break;
    			case EV_CAR_CHARGE:
    				setState(J1772_CAR_CHARGE);
    				break;
    			case EV_CAR_CHARGE_VENT:
    				setState(J1772_CAR_CHARGE_VENT);
    				break;
    			default:
    				DMSG(ME,"**** UNSUPPORTED EVENT RECEIVED **** %s (%d)",eventTxt(event),(int)event);
    				//assert(0);
    		}
    		YIELD();
		} else {
			adcReadCableCurrent ();
			if (status != J1772_CABLE_DISCONNECTED)	adcReadCP (NULL);	// car status
			if (currNumPhases > 0)
				if (relaisOffCounter > 0) {
					relaisOffCounter--;
					if (relaisOffCounter == 0) {
						DMSG(ME,"relaisOffCounter reached zero, turning off power");
						setPowerRelais(0);
					}
				}
			if (maxPhasesDetected == false) {
				if (phasesDetectCountdown) {
					phasesDetectCountdown--;
					if (phasesDetectCountdown == 0) {
						// unable to detect phases, assume 1 phase only
						DMSG(TAG,"unable to detect number of phases supported by car, assuming 1 and L1");
						resetSupportedPhases ();
						maxPhasesDetected = true;
						maxPhasesSupportedByCar = 1;
						adjustChargeCurrent();
					}
				}
			}
			YIELD();
		}
	} while (1);
	assert(0);
}



void j1772_getManualSettings (j1772_manual_settings_t * settings) {
	*settings = manSettings;
}


void j1772_updateManualSettings (j1772_manual_settings_t * settings) {
	manSettings = *settings;
	adjustChargeCurrent();
}


int j1772_manualModeEnabled () {
	return manSettings.enabled;
}


int j1772_minAmps() {
	return j1772_amps_min;
}

int j1772_maxAmps () {
	if (cableAmps < j1772_amps_max) return cableAmps;
	return j1772_amps_max;
}


int j1772_endTask(FILE *fd) {
	if (pvj1772TaskHandle != NULL) {
		ev_send (EV_TASK_J1772, EV_END_TASK, EV_DATA_NONE);
		int i = 0;
		while ((i<50) && (pvj1772TaskHandle != NULL)) {
			YIELD();
			i++;
		}
		if (pvj1772TaskHandle == NULL) {
			fprintf(fd,"j1772 task terminated\r\n");
			return 0;
		} else {
			fprintf(fd,"timeout waiting for j1772 task to terminate\r\n");
			return 1;
		}
	} else fprintf(fd, CONS_COLOR_STR_W_N("unable to stop j1772Task - task not started"));
	return 1;
}



int j1772_startTask(FILE *fd) {

	xTaskCreatePinnedToCore(j1772Task, "j1772", 4096, NULL, EVSE_STD_TASK_PRIO, &pvj1772TaskHandle, APP_CPU);
	if (pvj1772TaskHandle == NULL) {
		fprintf(fd,CONS_COLOR_STR_E_N("\r\nxTaskCreate did not return a handle for the j1772Task"));
		return 1;
	}
	return 0;
}



static struct {
	struct arg_lit *start = arg_lit0("s", "start", "Start j1772 task");
	struct arg_lit *stop  = arg_lit0("t", "stop", "Stop task");
	struct arg_int *debug = arg_int0("d","debug", NULL, "0=off, 1=debug messages on");
	struct arg_lit *info  = arg_lit0("i", "info", "Show current state");
    struct arg_end *end   = arg_end(5);
} j_commandArgs;



static int j_command(FILE *fd, FILE *, int, char**) {
	int atLeastOneParam = 0;
	if (j_commandArgs.debug->count > 0) {
		atLeastOneParam++;
		debug = j_commandArgs.debug->ival[0];
		if (debug == 0) {
			if (debug_stderr != NULL) {
				//fclose(debug_stderr);		// close debug file stream
				debug_stderr = NULL;
			}
		} else {
			//if (debug_stderr != NULL) fclose(debug_stderr);
			debug_stderr = fdopen (fileno(fd), "w");	// open it, we may be on another terminal
		}
		atLeastOneParam++;
		fprintf(fd,CONS_COLOR_STR_I_N("j1772 debug level: %d"),debug);
	}
	if ((j_commandArgs.start->count) && (j_commandArgs.stop->count)) {
		fprintf(fd,"what ? start or stop please\r\n");
		return 1;
	}
	if (j_commandArgs.start->count) {
		j1772_startTask(fd);
		atLeastOneParam++;
	}
	if (j_commandArgs.stop->count) {
		j1772_endTask(fd);
		atLeastOneParam++;
	}
	if (j_commandArgs.info->count) {
		fprintf(fd,CONS_COLOR_STR_I_N("cableAmps               : %d"),cableAmps);		// current detected cable maximum Amps
		fprintf(fd,CONS_COLOR_STR_I_N("chargeAmps              : %d"),chargeAmps);
		fprintf(fd,CONS_COLOR_STR_I_N("minAmps                 : %d"),j1772_amps_min);

		fprintf(fd,CONS_COLOR_STR_I_N("status                  : %s"),statusTxt(status));
		fprintf(fd,CONS_COLOR_STR_I_N("solarAvailableAmps      : %d (%d watt)"),solarAvailableAmps,solarAvailableAmps * GRID_VOLTAGE);
		fprintf(fd,CONS_COLOR_STR_I_N("availableAmps for charge: %d (%d watt)"),availableAmps(),availableAmps() * GRID_VOLTAGE);
		fprintf(fd,CONS_COLOR_STR_I_N("emChargingAmps          : %d (%d watt)"),emChargingAmps, emChargingWatts);

		fprintf(fd,CONS_COLOR_STR_I_N("maxPhasesSupportedByCar : %d"),maxPhasesSupportedByCar);
		fprintf(fd,CONS_COLOR_STR_I_N("supportedPhasesByCar    : 0x%02x 0x%02x 0x%02x"),supportedPhasesByCar[0],supportedPhasesByCar[1],supportedPhasesByCar[2]);
		fprintf(fd,CONS_COLOR_STR_I_N("maxPhasesDetected       : %s"),maxPhasesDetected ? "true" : "false");
		fprintf(fd,CONS_COLOR_STR_I_N("currNumPhases           : %d"),currNumPhases);
		fprintf(fd,CONS_COLOR_STR_I_N("relaisOffCounter        : %d"),relaisOffCounter);
		fprintf(fd,CONS_COLOR_STR_I_N("phasesDetectCountdown   : %d"),phasesDetectCountdown);

		atLeastOneParam++;
	}
	if (atLeastOneParam == 0) arg_print_glossary(fd, (void**) &j_commandArgs, NULL);
	return 0;
}

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------

static struct {
	struct arg_int *startA= arg_int0("s", "start", "Amp", "Start amps, defaults to 6 or configured MINAMPS");
	struct arg_int *endA  = arg_int0("e", "end",   "Amp", "End anps, defaults to detect cable Amps");
	struct arg_int *delay = arg_int0("d", "delay", "Sec", "Delay before setting the next current available, def=15, min=5");
    struct arg_end *end   = arg_end(5);
} pwmtest_commandArgs;


// return 1 to end test
int pwmtest (FILE *fd, FILE *fp_stdin, int amps, int timeoutSec){
	j1772_manual_settings_t settings;
	int i, timeRemaining;
	em_data_t *emData;

	emData = energyMeter_getCurrentData();  // ptr to engery meter data

	settings.enabled = true;
	settings.endTime = -1;
	settings.startTime = -1;
	settings.man_phases = 0x07;  // always 3 phases
	settings.man_amps = amps;
	j1772_updateManualSettings (&settings);  // set fixed amps and manual charge

	if (! j1772_carCharging()) {
		do {
			fprintf(fd,CONS_COLOR_STR_I("pwmtest") ": waiting for car to be connected and start charging, press any key to abort\r\n"); FFLUSH(fd);
			i = console_getChar (fp_stdin, 15000);
			if (i >= 0) return 1;
		} while (! j1772_carCharging());
	}
	timeRemaining = timeoutSec;
	while (timeRemaining > 0) {
		i = console_getChar (fp_stdin, 1000); timeRemaining--;
		if (i>=0) return 1;
		fprintf(fd,"pwmtest %d pwm: %d meter: %2.1f %2.1f %2.1f %3.1f %3.1f %3.1f %d %d %d\r\n",
				timeRemaining,amps,
				emData->currL1,emData->currL2,emData->currL2,
				emData->voltageL1,emData->voltageL2,emData->voltageL3,
				emData->pwrL1,emData->pwrL2,emData->pwrL3);
		FFLUSH(fd);
	}
	fprintf(fd,"pwmtest %d end\r\n",amps);

	return 0;

}


static int pwmtest_command(FILE *fd, FILE *fp_stdin, int, char**) {
	int startAmps;
	int endAmps;
	int delaySecs;
	int i,phasesSave,phasesL1Save,phasesL2Save,phasesL3Save;
	j1772_manual_settings_t settingsSave;
	int invTaskWasStarted;

	j1772_getManualSettings (&settingsSave);

	if (cableAmps < j1772_PWM_AMPS_MIN) {
		fprintf(fd,CONS_COLOR_STR_W_N("pwmtest: Warning: no cable detected, cable current may be out of range"));
		//return 1;
	}

	if (pwmtest_commandArgs.startA->count) {
		startAmps = pwmtest_commandArgs.startA->ival[0];
	} else {
		startAmps = configfile.get_u32 ("MINAMPS");
	}
	if (startAmps < j1772_PWM_AMPS_MIN) startAmps = j1772_PWM_AMPS_MIN;
	if (cableAmps > 0)
	  if (startAmps > cableAmps) startAmps = cableAmps;
	if (startAmps > j1772_amps_max) startAmps = j1772_amps_max;

	if (pwmtest_commandArgs.endA->count) {
		endAmps = pwmtest_commandArgs.endA->ival[0];
	} else {
		endAmps = cableAmps;
	}
	if (endAmps < j1772_PWM_AMPS_MIN) endAmps = j1772_PWM_AMPS_MIN;
	if (cableAmps > 0)
		if (endAmps > cableAmps) endAmps = cableAmps;
	if (endAmps > j1772_amps_max) endAmps = j1772_amps_max;


/*
	if (startAmps > endAmps) {
		i = startAmps;
		startAmps = endAmps;
		endAmps = i;
	}
*/
	delaySecs = 15;
	if (pwmtest_commandArgs.delay->count) delaySecs = pwmtest_commandArgs.delay->ival[0];
	if (delaySecs < 5) delaySecs = 5;

	if (j1772_carConnected()) {
		fprintf(fd,CONS_COLOR_STR_I("pwmtest") ": please disconnect car before starting this test\r\n");
		return 1;
	}
/*
	fprintf(fd,CONS_COLOR_STR_I("keytest") "\r\n");
	do {
		i = console_getChar (fp_stdin, 2000);
		fprintf(fd,"got %d %c\r\n",i,(char)i);
	} while (i != 'x');
*/
	fprintf(fd,CONS_COLOR_STR_I("pwmtest") ": starting pwm amps test from %d A, to %d A with delay %d seconds\r\n",startAmps,endAmps,delaySecs);
	fprintf(fd,"press y to start test ");
	fflush(fd);



	i = console_getChar (fp_stdin, 60000);
	if (i == 'y') {
		fprintf(fd,"%c\r\nStarting\r\n",(char)i); fflush(fd);
		phasesSave = configfile.get_u32("PHASES");
		phasesL1Save = configfile.get_u32("PHASE_L1");
		phasesL2Save = configfile.get_u32("PHASE_L2");
		phasesL3Save = configfile.get_u32("PHASE_L3");

		// enable all phases and disable autodetect
		configfile.set_u32("PHASES",3);
		configfile.set_u32("PHASE_L1",1);
		configfile.set_u32("PHASE_L2",1);
		configfile.set_u32("PHASE_L3",1);

		invTaskWasStarted = queryInverter_taskStarted();
		if (invTaskWasStarted) queryInverter_endTask(fd);

		if (startAmps > endAmps) {
			for (i=startAmps; i>=endAmps; i--) if (pwmtest(fd,fp_stdin,i,delaySecs)) break;
		} else
		{
			for (i=startAmps; i<=endAmps; i++) if (pwmtest(fd,fp_stdin,i,delaySecs)) break;
		}

		configfile.set_u32("PHASES",phasesSave);
		configfile.set_u32("PHASE_L1",phasesL1Save);
		configfile.set_u32("PHASE_L2",phasesL2Save);
		configfile.set_u32("PHASE_L3",phasesL3Save);
		j1772_updateManualSettings (&settingsSave);

		if (invTaskWasStarted) queryInverter_startTask(fd);

	} else fprintf(fd,"n\r\n");

	//if (atLeastOneParam == 0) arg_print_glossary(fd, (void**) &j_commandArgs, NULL);
	return 0;
}

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------


#ifndef ESP_PLATFORM
static struct {
	struct arg_int *voltage = arg_int0 (NULL,NULL,"<mV>","voltage for pp if running on emulator");
	struct arg_int *current = arg_int0 ("a","amps",NULL,"set amps supported by cable");
    struct arg_end *end = arg_end(2);
} pp_commandArgs;
#endif


static int pp_command(FILE *fd, FILE *, int, char**) {
#ifndef ESP_PLATFORM
	int a;
	if (pp_commandArgs.voltage->count > 0) {
		adc1_ch6 = pp_commandArgs.voltage->ival[0];
	}
	if (pp_commandArgs.current->count > 0) {
		a = pp_commandArgs.current->ival[0];
		switch (a) {
			case 13:	adc1_ch6 = 3000; break;
			case 20:	adc1_ch6 = 1800; break;
			case 32:	adc1_ch6 = 1000; break;
			case 63:	adc1_ch6 = 500; break;
			default:
				fprintf(fd,CONS_COLOR_STR_E_N("invalid current(%d), valid would be 13,20,32 and 63"),a);
				return 1;
		}
	}
#endif
	int voltage = adcReadCableCurrent ();
	fprintf(fd,"PP: %dmV %dA\r\n",voltage,cableAmps);
	return 0;
}


static struct {
#ifndef ESP_PLATFORM
	struct arg_lit *carNone   = arg_lit0("n", "none", "set car not connected");
	struct arg_lit *carReady  = arg_lit0("r", "ready", "set car ready");
	struct arg_lit *carCharge = arg_lit0("c", "charge", "set car charge");
	struct arg_lit *carVent   = arg_lit0("v", "vent", "set car charge+vent");
	struct arg_int *voltage  = arg_int0 (NULL,NULL,"<mV>","voltage for cp if running on emulator");
#endif
    struct arg_end *end = arg_end(2);
} cp_commandArgs;



static int cp_command(FILE *fd, FILE *, int, char**) {
#ifndef ESP_PLATFORM
	if (cp_commandArgs.voltage->count > 0)	adc1_ch7 = cp_commandArgs.voltage->ival[0];
	if (cp_commandArgs.carNone->count > 0)	adc1_ch7 = 3200;
	if (cp_commandArgs.carReady->count > 0)	adc1_ch7 = 2200;
	if (cp_commandArgs.carCharge->count > 0)	adc1_ch7 = 1500;
	if (cp_commandArgs.carVent->count > 0)	adc1_ch7 = 1000;
#endif

	j1772_status_t statusNew;
	j1772_status_t statusOld;

	statusOld = status;
	int voltage = adcReadCP (&statusNew);
	fprintf(fd,"CP: %dmV newStatus: %s (%d), prevStatus: %s (%d)\r\n",voltage,statusTxt(statusNew),int(statusNew),statusTxt(statusOld),(int(statusOld)));
	return 0;
}


void register_j1772commands() {
    CMD_REG(pilot, "Set pilot PWM", NULL, &pilot_command, &pilot_commandArgs);

#ifndef ESP_PLATFORM
    CMD_REG(pp,"show/set pp voltage for cable detection",NULL,&pp_command, &pp_commandArgs);
    CMD_REG(cp,"show/set cp voltage for car status detection",NULL,&cp_command, &cp_commandArgs);
#else
    CMD_REG(pp,"show pp voltage for cable detection"     ,NULL,&pp_command, NULL);
    CMD_REG(cp,"show control pilot voltage for car status detection",NULL,&cp_command, NULL);
#endif
    CMD_REG(pwmtest,"test car power consumption for various charge amps",NULL,&pwmtest_command, &pwmtest_commandArgs);

    CMD_REG(j1772,"j1772 (start,stop,info,debug)",NULL,&j_command, &j_commandArgs);
}
